/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.hibernate.SessionFactory;
import org.hibernate.cache.Cache;
import org.hibernate.cfg.Configuration;
import org.hibernate.engine.SessionFactoryImplementor;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.pvm.internal.jobexecutor.JobExecutor;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.CommonClassLoader;
import org.ow2.bonita.runtime.PackageClassLoader;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.DbTool;
import org.ow2.bonita.util.EnvConstants;


/**
 * @author Guillaume Porcher
 *
 */
public final class LocalTestSetup {

  private LocalTestSetup() { }

  @SuppressWarnings("unchecked")
  public static void setUp(boolean recreateDb) {
    if (!GlobalEnvironmentFactory.isInitialized()) {
      recreateDb = true;
    }
    final JobExecutor jobExecutor = GlobalEnvironmentFactory.getEnvironmentFactory().get(JobExecutor.class);
    if (recreateDb) {
      if (jobExecutor != null) {
        jobExecutor.stop();
      }
      final Map<String, String> hibernateConfigs = new HashMap<String, String>();
      hibernateConfigs.put(EnvConstants.CORE_HB_CONF_NAME,
          EnvConstants.CORE_HB_SESSION_FACTORY_NAME);

      hibernateConfigs.put(EnvConstants.HISTORY_HB_CONF_NAME,
          EnvConstants.HISTORY_HB_SESSION_FACTORY_NAME);
      for (final Map.Entry<String, String> e : hibernateConfigs.entrySet()) {
        final Configuration config = (Configuration) GlobalEnvironmentFactory.getEnvironmentFactory().get(e.getKey());
        if (config != null) {
          // recreate db between tests
          DbTool.recreateDb(e.getKey());
          // clean 2nd level cache:
          final SessionFactory sessionFactory =
            (SessionFactory) GlobalEnvironmentFactory.getEnvironmentFactory().get(e.getValue());
          if (sessionFactory != null && sessionFactory instanceof SessionFactoryImplementor) {
            // clear query results
            sessionFactory.evictQueries();
            // clear 2nd level cache
            final Map<String, Cache> caches = ((SessionFactoryImplementor)sessionFactory).getAllSecondLevelCacheRegions();
            for (final Cache cache : caches.values()) {
              cache.clear();
            }
          }
        }
      }
    }
    if (jobExecutor != null) {
      jobExecutor.start();
    }
  }

  public static void tearDown(final boolean testFail) throws BonitaException {
    final JobExecutor jobExecutor = GlobalEnvironmentFactory.getEnvironmentFactory().get(JobExecutor.class);
    if (jobExecutor != null && jobExecutor.isActive()) {
      jobExecutor.stop();
    }
    checkEverythingCleaned(!testFail);
  }

  private static void checkEverythingCleaned(final boolean failIfNotClean) throws BonitaException {
    String failMsg = "";
    final Set<String> classes = new HashSet<String>();

    final Map<PackageDefinitionUUID, PackageClassLoader> packageClassLoaders = ClassDataLoader.getPackageClassLoaders();
    if (packageClassLoaders != null && !packageClassLoaders.isEmpty()) {
      failMsg += "Some classes are still deployed at package level. You must write a test which "
        + "cleans everything. Please, remove them. Class list : " + classes;
      ClassDataLoader.clear();
    }
    final CommonClassLoader commonClassLoader = ClassDataLoader.getCommonClassLoader();
    if (!commonClassLoader.isEmpty()) {
      ClassDataLoader.clear();
      failMsg += "Some classes are still deployed at common level. You must write a test which "
          + "cleans everything. Please, remove them. Class list : " + classes;

    }

    final Set<ProcessInstance> instances = AccessorUtil.getQueryRuntimeAPI().getProcessInstances();
    if (instances != null && !instances.isEmpty()) {
      final Set<String> instanceUUIDs = new HashSet<String>();
      for (final ProcessInstance processInst : instances) {
        instanceUUIDs.add(processInst.getUUID().toString());
        if (processInst.getParentInstanceUUID() == null) {
          AccessorUtil.getRuntimeAPI().deleteProcessInstance(processInst.getUUID());
        }
      }
      failMsg += "Some instances are still found. You must write a test which cleans everything. Instance UUIDs = "
                 + instanceUUIDs;
    }

    final Set<PackageDefinition> remainingPackages = AccessorUtil.getQueryDefinitionAPI().getPackages();
    if (remainingPackages != null && !remainingPackages.isEmpty()) {
      final Set<String> packageIds = new HashSet<String>();
      for (final PackageDefinition packageDef : remainingPackages) {
        packageIds.add(packageDef.getPackageId());
      }
      failMsg += "Some packages are still found. You must write a test which cleans everything. Package ids = " + packageIds;

      // handle dependencies when removing packages.
      Set<PackageDefinition> packages = AccessorUtil.getQueryDefinitionAPI().getPackages();
      while (packages != null && !packages.isEmpty()) {
        // create list of all available packages
        final Set<PackageDefinitionUUID> loopPackageIds = new HashSet<PackageDefinitionUUID>();
        for (final PackageDefinition packageDef : packages) {
          loopPackageIds.add(packageDef.getPackageDefinitionUUID());
        }
        // remove from list package that are dependencies of other packages
        for (final PackageDefinition packageDef : packages) {
          for (final String dep : packageDef.getPackageDependencies()) {
            loopPackageIds.remove(new PackageDefinitionUUID(dep));
          }
        }
        // remove packages which do not have dependencies
        for (final PackageDefinitionUUID packageDefinitionUUID : loopPackageIds) {
          AccessorUtil.getManagementAPI().deletePackage(packageDefinitionUUID);
        }
        // do the same with remaining packages
        packages = AccessorUtil.getQueryDefinitionAPI().getPackages();
      }
    }


    final Set<ProcessDefinition> processes = AccessorUtil.getQueryDefinitionAPI().getProcesses();
    if (processes != null && !processes.isEmpty()) {
      final Set<String> processIds = new HashSet<String>();
      for (final ProcessDefinition processDef : processes) {
        processIds.add(processDef.getProcessId());
        AccessorUtil.getManagementAPI().deletePackage(processDef.getPackageDefinitionUUID());
      }
      failMsg += "Some processes are still found. You must write a test which cleans everything. Process ids = " + processIds;
    }

    if (failIfNotClean && !failMsg.equals("")) {
      throw new BonitaRuntimeException(failMsg);
    }
  }
}
