package org.ow2.bonita.activity.instantiation;

import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import junit.framework.Assert;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.activity.instantiation.instantiator.BadJoinNActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.BadVarActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.HookActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.JoinNumberConditionFail;
import org.ow2.bonita.activity.instantiation.instantiator.JoinNumberConditionPass;
import org.ow2.bonita.activity.instantiation.instantiator.NullActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.TestActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.TestOneJoinActivityInstantiator;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.ExpressionEvaluationException;
import org.ow2.bonita.facade.exception.MultiInstantiatorInvocationException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;

public class MultiInstantiationTest extends APITestCase {

  public void testMultiInstantiation() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("multiInstantiation.xpdl");
    final String processId = "multiInstantiation";

    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
      TestActivityInstantiator.class));
    final ProcessDefinition process = processMap.get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    this.checkExecutedManyTimes(instanceUUID, new String[]{"r1"}, 2);

    final Set<ActivityInstance<ActivityBody>> actIntances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");

    final Set<Object> varValues = new HashSet<Object>();

    for (final ActivityInstance<ActivityBody> activityInstance : actIntances) {
      final Map<String, Object> variables = activityInstance.getLastKnownVariableValues();
      Assert.assertTrue(variables.containsKey("testVar"));
      varValues.add(variables.get("testVar"));
    }

    final Set<Object> expectedValues = new HashSet<Object>();
    expectedValues.add("val1");
    expectedValues.add("val2");

    Assert.assertEquals(expectedValues, varValues);

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }


  public void testDeployWrongInstantiator() {
    //Test deployment with multiInstantiatorClass class not found
    final URL xpdlUrl = this.getClass().getResource("incorrectMultiInstantiation.xpdl");
    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl));
      Assert.fail("Deployment shouldn't passed");
    } catch (final DeploymentException e) {
      Assert.assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a MultiInstantiator class name"));
      Assert.assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a variable id"));

    }
  }

  public void testTransitionsConditions() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("transitionConditionMultiInstantiation.xpdl");
    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(
      xpdlUrl, JoinNumberConditionPass.class, JoinNumberConditionFail.class));
    //Test join number with transition condition on activity variable
    // this is not allowed
    final String processId = "transitionCondition";
    final ProcessDefinitionUUID processUUID = processMap.get(processId).getUUID();
    try {
      this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.fail("Unreachable statement !");
    } catch (final ExpressionEvaluationException e) {
      final String expected = "conditionVar";
      Assert.assertTrue("Expected: " + expected + ", result: " + e.getMessage(), e.getMessage().contains(expected));
    }
    final Set<ProcessInstance> processInstances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertTrue(processInstances == null || processInstances.isEmpty());

    final PackageDefinitionUUID packageUUID = processMap.get(processId).getPackageDefinitionUUID();
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testHookMultiInstantiation() throws BonitaException {
    //Test hook execution with multi instantiation
    final URL xpdlUrl = this.getClass().getResource("hookMultiInstantiation.xpdl");
    final String processId = "hookMultiInstantiation";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
      HookActivityInstantiator.class, IncrInstanceVariableHook.class)).get(processId);

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(process.getUUID());
    this.checkExecutedManyTimes(instanceUUID, new String[]{"r1"}, 3);

    final Set<ActivityInstance<ActivityBody>> actIntances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");

    final Set<Object> varValues = new HashSet<Object>();
    final Set<Object> expectedValues = new HashSet<Object>();

    for (final ActivityInstance<ActivityBody> activity : actIntances) {
      final Map<String, Object> variables = activity.getLastKnownVariableValues();
      Assert.assertTrue(variables.containsKey("counter"));
      varValues.add(variables.get("counter"));
    }

    expectedValues.add(4L);
    expectedValues.add(5L);
    expectedValues.add(6L);

    Assert.assertEquals(expectedValues, varValues);

    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Multiinstantiation element with a variable not declared in the XPDL
   */
  public void testInvalidVarName() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    final URL xpdlUrl = this.getClass().getResource("invalidVarNameMultiInstantiation.xpdl");
    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, BadVarActivityInstantiator.class));
      Assert.fail("exception expected");
    } catch (final DeploymentException e) {
      Assert.assertTrue(e.getMessage().contains("MultiInstantiation variable testVar must be a local variable of activity r1"));
    }
  }

  /*
   * Multiinstantiation element without variable
   */
  public void testNoVariable() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    final URL xpdlUrl = this.getClass().getResource("noVarMultiInstantiation.xpdl");
    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, BadVarActivityInstantiator.class));
      Assert.fail("exception expected");
    } catch (final DeploymentException e) {
      Assert.assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a variable id (in a nested Variable element)"));
    }
  }

  public void testFailInstantiator() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    final URL xpdlUrl = this.getClass().getResource("failMultiInstantiation.xpdl");
    String processId = "multiInstantiation";
    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
      NullActivityInstantiator.class, BadJoinNActivityInstantiator.class));
    ProcessDefinition process = processMap.get(processId);
    try {
      this.getRuntimeAPI().instantiateProcess(process.getUUID());
      Assert.fail("Failing multiInstantiatorClass should throw an exception");
    } catch (final MultiInstantiatorInvocationException e) {
      Assert.assertTrue(this.causeContains("MultiInstantiator execution returned null in activity r1", e));
    }
    this.getRuntimeAPI().deleteAllProcessInstances(process.getUUID());

    processId = "invalidJoinNumber";
    process = processMap.get(processId);
    try {
      this.getRuntimeAPI().instantiateProcess(process.getUUID());
      Assert.fail("Process instantiation shouldn't passed because of invalid join number");
    } catch (final MultiInstantiatorInvocationException e) {
      Assert.assertTrue(this.causeContains("The join number must be greater than 0", e));
    }

    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  // multi instantiation element referring to global variable
  public void testGlobalVariableMultiInstantiation() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("globalVarMultiInstantiation.xpdl");

    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
          TestActivityInstantiator.class));
      Assert.fail("Deployment should fail");
    } catch (final DeploymentException e) {
      Assert.assertTrue(e.getMessage().contains("MultiInstantiation variable testVar must be a local variable of activity r1"));
    }
  }

  public void testSubflowMultiInstantiation() throws BonitaException {
    //Test subflow with multi instantiation
    final URL xpdlUrl = this.getClass().getResource("subflowMultiInstantiation.xpdl");
    String processId = "mainCondPass";
    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(
        xpdlUrl, JoinNumberConditionPass.class, JoinNumberConditionFail.class, CheckExecutionHook.class));

    ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processMap.get(processId).getUUID());
    this.checkExecutedManyTimes(instanceUUID, new String[]{"a2"}, 4);
    this.checkExecutedOnce(instanceUUID, new String[]{"a3", "a4"});

    this.getRuntimeAPI().deleteAllProcessInstances(processMap.get(processId).getUUID());

    processId = "mainCondFail";
    instanceUUID = this.getRuntimeAPI().instantiateProcess(processMap.get(processId).getUUID());
    this.checkExecutedManyTimes(instanceUUID, new String[]{"a2"}, 3);
    this.checkStopped(instanceUUID, new String[]{"a3"});

    try {
      this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4");
      Assert.fail("Activity : a3 must not be executed in this instance!");
    } catch (final ActivityNotFoundException e) {
      //nothing, we expect an exception
    }
    this.getRuntimeAPI().deleteAllProcessInstances(processMap.get(processId).getUUID());

    this.getManagementAPI().undeploy(processMap.get(processId).getPackageDefinitionUUID());
    this.getManagementAPI().deletePackage(processMap.get(processId).getPackageDefinitionUUID());
  }


  // This test creates two instance of task r, and wait for only one to be done. The other task is aborted.
  public void testPartialJoinTaskMultiInstantiation() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("joinMultiInstantiation.xpdl");
    final String processId = "multiInstantiation";

    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
      TestOneJoinActivityInstantiator.class));
    final ProcessDefinition process = processMap.get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    final Set<ActivityInstance<ActivityBody>> actIntances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
    Assert.assertEquals(2, actIntances.size());

    final Set<Object> varValues = new HashSet<Object>();

    // check two instance of activity r1 have been created
    for (final ActivityInstance<ActivityBody> activityInstance : actIntances) {
      final Map<String, Object> variables = activityInstance.getLastKnownVariableValues();
      Assert.assertTrue(variables.containsKey("testVar"));
      varValues.add(variables.get("testVar"));
    }

    final Set<Object> expectedValues = new HashSet<Object>();
    expectedValues.add("val1");
    expectedValues.add("val2");

    Assert.assertEquals(expectedValues, varValues);

    // execute one task
    final ActivityInstance<ActivityBody> activityInstance = actIntances.iterator().next();
    final ActivityBody body = activityInstance.getBody();
    Assert.assertTrue(body instanceof TaskInstance);
    if (body instanceof TaskInstance) {
      final TaskUUID taskUUID = ((TaskInstance) body).getUUID();
      this.getRuntimeAPI().startTask(taskUUID, false);
      this.getRuntimeAPI().finishTask(taskUUID, false);
    }

    // check we are now on activity r2
    final Set<ActivityInstance<ActivityBody>> r2instances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r2");
    Assert.assertEquals(1, r2instances.size());

    // check the second task r1 has been aborted
    final Collection<ActivityInstance<TaskInstance>> abortedTasks =
      this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.ABORTED);
    Assert.assertEquals(1, abortedTasks.size());
    Assert.assertEquals("r1", abortedTasks.iterator().next().getActivityId());

    // execute task r2 to finish the instance
    final Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    Assert.assertEquals("r2", task.getActivityId());
    Assert.assertEquals(ActivityState.READY, task.getBody().getState());
    final TaskUUID taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, false);
    this.getRuntimeAPI().finishTask(taskUUID, false);

    this.checkExecutedOnce(instanceUUID, new String[]{"r2", "BonitaEnd"});

    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * Automatic process that creates 2 subflows with a Manual activity.
   *
   */
   public void testPartialJoinSubFlowMultiInstantiation() throws BonitaException {
     final URL xpdlUrl = this.getClass().getResource("joinSubflowMultiInstantiation.xpdl");
     final String processId = "main";
     final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
         TestOneJoinActivityInstantiator.class)).get(processId);
     final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
     final ProcessDefinitionUUID processUUID = process.getUUID();
     final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
     this.checkStopped(instanceUUID, new String[]{"a1"});

     // check two tasks have been created in two subprocess
     final Collection<ActivityInstance< TaskInstance >> taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
     Assert.assertEquals(2, taskActivities.size());
     final Iterator<ActivityInstance< TaskInstance >> iterator = taskActivities.iterator();
     final ActivityInstance< TaskInstance > taskActivity = iterator.next();
     final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(taskActivity.getProcessInstanceUUID());
     Assert.assertEquals(instanceUUID, processInstance.getParentInstanceUUID());
     // check other task is not in the same sub process
     final ActivityInstance< TaskInstance > otherTaskActivity = iterator.next();
     Assert.assertFalse(otherTaskActivity.getProcessInstanceUUID().equals(taskActivity.getProcessInstanceUUID()));

     // execute one task => the subprocess will be finished
     final TaskUUID taskUUID = taskActivity.getBody().getUUID();
     this.getRuntimeAPI().startTask(taskUUID, true);
     this.getRuntimeAPI().finishTask(taskUUID, true);

     Assert.assertEquals(InstanceState.FINISHED, this.getQueryRuntimeAPI().getProcessInstance(taskActivity.getProcessInstanceUUID()).getInstanceState());

     // check other subprocess instance has been aborted
     Assert.assertEquals(InstanceState.ABORTED, this.getQueryRuntimeAPI().getProcessInstance(otherTaskActivity.getProcessInstanceUUID()).getInstanceState());
     Assert.assertEquals(ActivityState.ABORTED, this.getQueryRuntimeAPI().getActivityInstance(otherTaskActivity.getUUID()).getBody().getState());

     // check end of main execution
     this.checkExecutedOnce(instanceUUID, new String[]{"a1", "a3"});

     this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
     this.getManagementAPI().undeploy(packageUUID);
     this.getManagementAPI().deletePackage(packageUUID);
   }

   // This test creates two instance of task r1.
   // The task r3 is executed, and as r2 as a join XOR, r1 tasks are aborted.
   public void testJoinXORAbortMultiInstantiation() throws BonitaException {
     final URL xpdlUrl = this.getClass().getResource("joinXORMultiInstantiation.xpdl");
     final String processId = "multiInstantiation";

     final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
       TestOneJoinActivityInstantiator.class));
     final ProcessDefinition process = processMap.get(processId);
     final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
     final ProcessDefinitionUUID processUUID = process.getUUID();
     final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

     final Set<ActivityInstance<ActivityBody>> actIntances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
     Assert.assertEquals(2, actIntances.size());

     final Set<Object> varValues = new HashSet<Object>();

     // check two instance of activity r1 have been created
     for (final ActivityInstance<ActivityBody> activityInstance : actIntances) {
       final Map<String, Object> variables = activityInstance.getLastKnownVariableValues();
       Assert.assertTrue(variables.containsKey("testVar"));
       varValues.add(variables.get("testVar"));
     }

     final Set<Object> expectedValues = new HashSet<Object>();
     expectedValues.add("val1");
     expectedValues.add("val2");

     Assert.assertEquals(expectedValues, varValues);

     // execute task r3
     final Set<ActivityInstance<ActivityBody>> r3Intances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r3");
     Assert.assertEquals(1, r3Intances.size());
     final ActivityInstance<ActivityBody> activityInstance = r3Intances.iterator().next();
     final ActivityBody body = activityInstance.getBody();
     Assert.assertTrue(body instanceof TaskInstance);
     if (body instanceof TaskInstance) {
       final TaskUUID taskUUID = ((TaskInstance) body).getUUID();
       this.getRuntimeAPI().startTask(taskUUID, false);
       this.getRuntimeAPI().finishTask(taskUUID, false);
     }

     // check we are now on activity r2
     final Set<ActivityInstance<ActivityBody>> r2instances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r2");
     Assert.assertEquals(1, r2instances.size());

     // check the tasks r1 has been aborted
     final Collection<ActivityInstance<TaskInstance>> abortedTasks =
       this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.ABORTED);
     Assert.assertEquals(2, abortedTasks.size());
     for (final ActivityInstance<TaskInstance> abortedTask : abortedTasks) {
       Assert.assertEquals("r1", abortedTask.getActivityId());
       Assert.assertEquals(ActivityState.ABORTED, abortedTask.getBody().getState());
     }

     // execute task r2 to finish the instance
     final Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
     Assert.assertEquals(1, tasks.size());
     final ActivityInstance<TaskInstance> task = tasks.iterator().next();
     Assert.assertEquals("r2", task.getActivityId());
     Assert.assertEquals(ActivityState.READY, task.getBody().getState());
     final TaskUUID taskUUID = task.getBody().getUUID();
     this.getRuntimeAPI().startTask(taskUUID, false);
     this.getRuntimeAPI().finishTask(taskUUID, false);

     this.checkExecutedOnce(instanceUUID, new String[]{"r2", "BonitaEnd"});

     this.getManagementAPI().undeploy(packageUUID);
     this.getManagementAPI().deletePackage(packageUUID);
   }
}
