package org.ow2.bonita.async;

import java.io.IOException;
import java.net.URL;
import java.util.Set;

import junit.framework.Assert;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;

public class AsyncTest extends APITestCase {

  private final long basicMaxWaitTime = 120 * 1000;
  private final long basicSleepTime = 100;
  
  public void testAsyncInstance() throws BonitaException {
    //A (auto/sync) -> B(auto/async) -> C(auto/async)
    //check the instance is finished. Check the var was set (hook was executed)
    //this test check that a process having asynchronous activities works as if it hasn't.
    final URL xpdlUrl = this.getClass().getResource("asyncInstance.xpdl");
    final String processId = "asyncInstance";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, NoOpHook.class, SetVarHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);

    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.checkExecutedOnce(instance, new String[]{"A", "B", "C"});

    final String var = (String) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "var");
    Assert.assertEquals("var was not set from hook", "setFromHook", var);

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncBfail() throws BonitaException, IOException {
    //A (auto/sync) -> B(auto/async) -> C(auto/sync)
    //check if B has a hook which throws an exception, A is executed, B and C not.
    //This test checks transaction has been committed before executing B
    final URL xpdlUrl = this.getClass().getResource("asyncBfail.xpdl");
    final String processId = "asyncBfail";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, NoOpHook.class, SetVarHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);

    final Set<ActivityInstance<ActivityBody>> activityInsts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "A");
    Assert.assertEquals("Activity A executed more than once.", 1, activityInsts.size());
    final ActivityInstance<ActivityBody> activityInst = activityInsts.iterator().next();
    Assert.assertNotNull("Activity A not executed", activityInst);
    Assert.assertNotNull("Bad state for activity A", activityInst.getBody().getEndedDate());

    Assert.assertEquals("instance is finished", InstanceState.STARTED, instance.getInstanceState());
    final String var = (String) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "var");
    Assert.assertEquals("var was set from hook", "novalue", var);

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncTasks() throws BonitaException, IOException {
    // A (manu/sync) -> B (manu/async) -> C (manu/async) -> D (manu/sync)
    // this test check that a process having asynchronous activities works as if it hasn't.
    final URL xpdlUrl = this.getClass().getResource("asyncTasks.xpdl");
    final String processId = "asyncTasks";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, SetVarHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    this.waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "A");
    this.executeTask(instanceUUID, "A");

    this.waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "B");
    this.executeTask(instanceUUID, "B");

    String var = (String) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "var");
    Assert.assertEquals("var was not set from hook", "novalue", var);

    this.waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "C");
    this.executeTask(instanceUUID, "C");

    var = (String) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "var");
    Assert.assertEquals("var was not set from hook", "setFromHook", var);

    this.waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "D");
    this.executeTask(instanceUUID, "D");

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncCycle() throws BonitaException, IOException {
    // A (manu/sync) -> B (manu/async) -> C (manu/async) -> D (manu/sync)
    // this test check that a process having asynchronous activities works as if it hasn't.
    final URL xpdlUrl = this.getClass().getResource("asyncCycle.xpdl");
    final String processId = "asyncCycle";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, NoOpHook.class, IncrementHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID);

    final Long counter = (Long) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    Assert.assertEquals("counter is not equal to 5", 5, counter.intValue());

    this.checkExecutedManyTimes(instanceUUID, new String[]{"A", "B", "C"}, 5);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncMultiInstantiator() throws BonitaException, IOException {
    // this tests checks that a multi instantiated activity async works
    final URL xpdlUrl = this.getClass().getResource("asyncMulti.xpdl");
    final String processId = "asyncMulti";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class, MultiIncrementHook.class, MultiInstantiator.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID);

    final Long counter = (Long) this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    final int expected = 11 + 17 + 29;
    Assert.assertEquals("counter is not equal to " + expected, expected, counter.intValue());

    this.checkExecutedOnce(instanceUUID, new String[]{"A", "C"});
    this.checkExecutedManyTimes(instanceUUID, new String[]{"B"}, 3);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncSubflow() throws BonitaException, IOException {
    // this tests checks that a subprocess activity async works
    // and that an async activity in a subprocess works
    final URL parentXpdlUrl = this.getClass().getResource("asyncSubParent.xpdl");
    final String parentProcessId = "asyncSubParent";
    final URL childXpdlUrl = this.getClass().getResource("asyncSubChild.xpdl");
    final String childProcessId = "asyncSubChild";

    // deploy child
    final ProcessDefinition childProcess = this.getManagementAPI().deployBar(this.getBarContent(childXpdlUrl,
        NoOpHook.class)).get(childProcessId);
    final PackageDefinitionUUID childPackageUUID = childProcess.getPackageDefinitionUUID();

    // deploy parent
    final ProcessDefinition parentProcess = this.getManagementAPI().deployBar(this.getBarContent(parentXpdlUrl,
        NoOpHook.class)).get(parentProcessId);
    final PackageDefinitionUUID parentPackageUUID = parentProcess.getPackageDefinitionUUID();
    final ProcessDefinitionUUID parentProcessUUID = parentProcess.getUUID();

    final ProcessInstanceUUID parentInstanceUUID = this.getRuntimeAPI().instantiateProcess(parentProcessUUID);

    this.waitForChildInstantiated(this.basicMaxWaitTime, this.basicSleepTime, parentInstanceUUID);
    ProcessInstance parentInstance = this.getQueryRuntimeAPI().getProcessInstance(parentInstanceUUID);
    final ProcessInstanceUUID childInstanceUUID = parentInstance.getChildrenInstanceUUID().iterator().next();

    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, childInstanceUUID);

    final ProcessInstance childInstance = this.getQueryRuntimeAPI().getProcessInstance(childInstanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, childInstance.getInstanceState());

    this.checkExecutedOnce(childInstanceUUID, new String[]{"A"});

    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, parentInstanceUUID);

    parentInstance = this.getQueryRuntimeAPI().getProcessInstance(parentInstanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, parentInstance.getInstanceState());

    this.checkExecutedOnce(parentInstanceUUID, new String[]{"A", "B"});

    this.getRuntimeAPI().deleteProcessInstance(parentInstanceUUID);

    this.getManagementAPI().undeploy(parentPackageUUID);
    this.getManagementAPI().deletePackage(parentPackageUUID);

    this.getManagementAPI().undeploy(childPackageUUID);
    this.getManagementAPI().deletePackage(childPackageUUID);
  }

  public void testAsyncManySplitsCycleJoinAndSimple() throws BonitaException, IOException {
    final URL xpdlUrl = this.getClass().getResource("asyncSplitCycleJoinAndSimple.xpdl");
    final String processId = "asyncSplitCycle";
    //A, C and D are sync, others are async
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class, IncrementHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    final int loopNb = 3;

    //instantiate call is not blocking. Wait for instanceEnd
   this.waitForInstanceEnd(1000* 5000, this.basicSleepTime, instanceUUID);

    this.checkExecutedOnce(instanceUUID, new String[]{"D"});
    final int nodeNb = 2;
    final String[] nodes = new String[nodeNb + 2];
    for (int i = 0; i < nodeNb; i++) {
      nodes[i] = "B" + (i+1);
    }
    nodes[nodeNb] = "A";
    nodes[nodeNb + 1] = "C";

    this.checkExecutedManyTimes(instanceUUID, nodes, loopNb);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testAsyncManySplitsCycleJoinAnd() throws BonitaException, IOException {
    final URL xpdlUrl = this.getClass().getResource("asyncSplitCycleJoinAnd.xpdl");
    final String processId = "asyncSplitCycle";
    //A, C and D are sync, others are async
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class, IncrementHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    final int loopNb = 20;

    //instantiate call is not blocking. Wait for instanceEnd
   this.waitForInstanceEnd(this.basicMaxWaitTime * (1 + loopNb / 2), this.basicSleepTime, instanceUUID);

    this.checkExecutedOnce(instanceUUID, new String[]{"D"});
    final int nodeNb = 10;
    final String[] nodes = new String[nodeNb + 2];
    for (int i = 0; i < nodeNb; i++) {
      nodes[i] = "B" + (i+1);
    }
    nodes[nodeNb] = "A";
    nodes[nodeNb + 1] = "C";

    this.checkExecutedManyTimes(instanceUUID, nodes, loopNb);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }
  public void testAsyncManySplitsCycleJoinXor() throws BonitaException, IOException {
    final URL xpdlUrl = this.getClass().getResource("asyncSplitCycleJoinXor.xpdl");
    final String processId = "asyncSplitCycle";
    //A, C and D are sync, others are async
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class, IncrementHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    final int loopNb = 20;

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime * (1 + loopNb / 2), this.basicSleepTime, instanceUUID);

    this.checkExecutedOnce(instanceUUID, new String[]{"D"});

    this.checkExecutedManyTimes(instanceUUID, new String[]{"A", "C"}, loopNb);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testAsyncJoinXor() throws BonitaException, IOException {
    final URL xpdlUrl = this.getClass().getResource("asyncJoinXor.xpdl");
    final String processId = "asyncJoinXor";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    this.waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "D");

    try {
      //important so that second job on C executes
      Thread.sleep(2000);
    } catch (final InterruptedException e) {

    }

    this.executeTask(instanceUUID, "D");

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID);

    this.checkExecutedOnce(instanceUUID, new String[]{"A", "C", "D"});

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  
  public void testAsyncJoinXorCycle() throws BonitaException, IOException {
    final URL xpdlUrl = this.getClass().getResource("asyncJoinXorCycle.xpdl");
    final String processId = "asyncJoinXorCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        NoOpHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    waitForTask(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID, "D");

    try {
      //important so that second job on C executes
      Thread.sleep(2000);
    } catch (InterruptedException e) {

    }

    executeTask(instanceUUID, "D");

    //instantiate call is not blocking. Wait for instanceEnd
    this.waitForInstanceEnd(this.basicMaxWaitTime, this.basicSleepTime, instanceUUID);

    this.checkExecutedOnce(instanceUUID, new String[]{"D"});

    int loopNb = 10;
    this.checkExecutedManyTimes(instanceUUID, new String[]{"A", "C"}, loopNb);

    final ProcessInstance instance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals("instance is not finished", InstanceState.FINISHED, instance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }
  
}
