/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.deadline;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.impl.DeadlineDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.util.Misc;

/**
 * @author Pascal Verdage
 */
public class DeadlineParsingTest extends APITestCase {

  public void testParseInvalidExecution() {
    Class< ? > hook = NullHook.class;
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("INVALID", "1000", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    try {
      getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId).getPackageDefinitionUUID();

      fail();
    } catch (DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("Invalid enum value"));
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("Invalid enum value"));
    }
  }

  public void testParseNoExecution() throws DeploymentException, ProcessNotFoundException, 
  ActivityNotFoundException, PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    Class< ? > hook = NullHook.class;
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, "1000", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    Set<DeadlineDefinition> deadlines = null;
    try {
      deadlines = getQueryDefinitionAPI().getProcessActivity(processUUID, "a").getDeadlines();
    } finally {
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }
    Set<DeadlineDefinition> expected = new HashSet<DeadlineDefinition>();
    expected.add(new DeadlineDefinitionImpl("1000", hook.getName(), DeadlineDefinition.Execution.SYNCHR));

    assertSameContent(expected, deadlines);
  }

  public void testParseValidExecutions() throws DeploymentException, ProcessNotFoundException, 
  ActivityNotFoundException, PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    Class< ? > hook = NullHook.class;
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("SYNCHR", "1000", hook.getName())
      .addDeadline("ASYNCHR", "1000", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    Set<DeadlineDefinition> deadlines = null;
    try {
      deadlines = getQueryDefinitionAPI().getProcessActivity(processUUID, "a").getDeadlines();
    } finally {
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }

    Set<DeadlineDefinition> expected = new HashSet<DeadlineDefinition>();
    expected.add(new DeadlineDefinitionImpl("1000", hook.getName(), DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("1000", hook.getName(), DeadlineDefinition.Execution.ASYNCHR));
    assertSameContent(expected, deadlines);
  }

  public void testParseInvalidCondition() {
    Class< ? > hook = NullHook.class;
    String processId = "main";
    String packageId = "invalidCondition";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline condition")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("ASYNCHR", "2008/05/31/23/59/", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    try {
      ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
      PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
      getManagementAPI().undeploy(packageUUID);

      fail();
    } catch (DeploymentException e) {
      assertTrue(e.getMessage().contains("deadline condition"));
    }
  }

  public void testParseNoCondition() {
    Class< ? > hook = NullHook.class;
    String processId = "main";
    String packageId = "noCondition";
    String document = new DeadlineDocument(packageId, "Test of no deadline condition")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("ASYNCHR", null, hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    try {
      ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
      PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
      getManagementAPI().undeploy(packageUUID);

      fail();
    } catch (DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("DeadlineCondition element is not specified"));
    }
  }

  public void testParseValidConditions() throws DeploymentException, 
  ProcessNotFoundException, ActivityNotFoundException, 
  PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    Class< ? > hook = NullHook.class;
    String exception = hook.getName();
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("SYNCHR", "10000", exception)
      .addDeadline("SYNCHR", "2008/05/31/23/59/59", exception)
      .addDeadline("SYNCHR", "2008/05/31/23/59/59/999", exception)
      // the parsing stops as soon as a valid format is detected
      .addDeadline("SYNCHR", "2008/05/31/23/59/59/999aaa", exception)
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    Set<DeadlineDefinition> deadlines = null;
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      deadlines = getQueryDefinitionAPI().getProcessActivity(processUUID, "a").getDeadlines();
    } finally {
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }

    Set<DeadlineDefinition> expected = new HashSet<DeadlineDefinition>();
    expected.add(new DeadlineDefinitionImpl("10000", exception, DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("2008/05/31/23/59/59", exception, DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("2008/05/31/23/59/59/999", exception, DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("2008/05/31/23/59/59/999aaa", exception, DeadlineDefinition.Execution.SYNCHR));

    assertSameContent(expected, deadlines);
  }

  public void testParseInvalidException() {
    String processId = "main";
    String packageId = "invalidCondition";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline condition")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("ASYNCHR", "2008/05/31/23/59/59", "UnexistingHook")
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    try {
      getManagementAPI().deployBar(getBarContent(document.getBytes())).get(processId);
      fail();
    } catch (DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("No Class available"));
    }
  }

  public void testParseNoException() {
    String processId = "main";
    String packageId = "invalidCondition";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline condition")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("ASYNCHR", "2008/05/31/23/59/", null)
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    try {
      ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes())).get(processId);
      PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
      getManagementAPI().undeploy(packageUUID);
      fail();
    } catch (DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("ExceptionName element is not specified"));
    }
  }

  public void testParseValidExceptions() throws DeploymentException, 
  ProcessNotFoundException, ActivityNotFoundException, 
  PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("SYNCHR", "10000", NotAHook.class.getName())
      .addDeadline("SYNCHR", "10000", SetTimeHook.class.getName())
      .addDeadline("SYNCHR", "10000", SetTimeTxHook.class.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    Set<DeadlineDefinition> deadlines = null;
    ProcessDefinition process = getManagementAPI().deployBar(
        getBarContent(document.getBytes(), NotAHook.class, SetTimeHook.class, SetTimeTxHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      deadlines = getQueryDefinitionAPI().getProcessActivity(processUUID, "a").getDeadlines();
    } finally {
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }

    Set<DeadlineDefinition> expected = new HashSet<DeadlineDefinition>();
    expected.add(new DeadlineDefinitionImpl("10000", NotAHook.class.getName(), DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("10000", SetTimeHook.class.getName(), DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("10000", SetTimeTxHook.class.getName(), DeadlineDefinition.Execution.SYNCHR));
    assertSameContent(expected, deadlines);
  }

  public void testSameDeadline() throws DeploymentException, 
  ProcessNotFoundException, ActivityNotFoundException, 
  PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    Class< ? > hook = NullHook.class;
    String exception = hook.getName();
    String processId = "main";
    String packageId = "invalidExecution";
    String document = new DeadlineDocument(packageId, "Test of an invalid deadline execution")
      .startProcess(processId, "main")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline("SYNCHR", "1000", exception)
      .addDeadline("SYNCHR", "1000", exception)
      .addDeadline("SYNCHR", "1000", exception)
      .endActivity().endActivities().endProcess().endDocument().getDocument();
    Set<DeadlineDefinition> deadlines = null;

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      deadlines = getQueryDefinitionAPI().getProcessActivity(processUUID, "a").getDeadlines();
    } finally {
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }

    assertEquals(3, deadlines.size());
    Set<DeadlineDefinition> expected = new HashSet<DeadlineDefinition>();
    expected.add(new DeadlineDefinitionImpl("1000", exception, DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("1000", exception, DeadlineDefinition.Execution.SYNCHR));
    expected.add(new DeadlineDefinitionImpl("1000", exception, DeadlineDefinition.Execution.SYNCHR));
    assertSameContent(expected, deadlines);
  }

  public static void assertSameContent(Set<DeadlineDefinition> expected, Set<DeadlineDefinition> actual) {
    if (actual == null) {
      assertNull("actual is null, but not expected", expected);
    } else {
      assertNotNull("expected is null, but not actual", expected);
      assertEquals(expected.size(), actual.size());
      List<DeadlineDefinition> availableValues = new ArrayList<DeadlineDefinition>(expected);
      for (DeadlineDefinition actualDef : actual) {
        for (DeadlineDefinition availableDef : availableValues) {
          if (actualDef.getExecution().equals(availableDef.getExecution())
              && actualDef.getExceptionName().equals(availableDef.getExceptionName())
              && actualDef.getDeadlineCondition().equals(availableDef.getDeadlineCondition())) {
            availableValues.remove(availableDef);
            break;
          }
        }
      }
      assertEquals(availableValues.size() + " expected elements were not found", 0, availableValues.size());
    }
  }
}
