package org.ow2.bonita.facade;

import java.net.URL;
import java.util.Arrays;
import java.util.Collection;

import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.SimpleCallbackHandler;

public class AuthorizedStartFinishTest extends APITestCase {



  /*
   *  Test startTask(String taskUUID) from the runtimeAPI.
   *
   *  test:
   *    - initial state need to be READY (IllegalTaskStateException raised if state = SUSPENDED)
   *    - if successful state become: EXECUTING.
   *    - if task assigned to the authenticated user check he can start the task.
   *
   *    - check that an other user into the candidates list can't start the task.
   *    - check also that an other user not into the candidates list can't start the task.
   *
   *    - if the task is not assigned (role mapper without performer assign)
   *      check that an other users of the candidates list can start the task.
   *
   * tests following exceptions are raised:
   *  @throws TaskNotFoundException if the task has not been found.
   *  @throws IllegalTaskStateException if the state of the task has not READY state.
   *  @throws BonitaInternalException if an exception occurs (ex. : if taskUUID == null)
   */
  public void testStartTask() throws BonitaException, LoginException {
    String processId = "testGetTask";
    final String loginAdmin = "admin";
    final String passwordAdmin = "bpm";
    final String loginJohn = "john";
    final String passwordJohn = "bpm";
    final String loginJack = "jack";
    final String passwordJack = "bpm";
    URL xpdlUrl = this.getClass().getResource("testGetTask.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    TaskUUID taskUUID = null;
    ProcessInstanceUUID instanceUUID = null;

    //instantiate the process => create the task a (with candidates = admin; userId = admin)
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!, exception: " + Misc.getStackTraceFrom(e));
    }
    //check throws BonitaInternalException
    try {
      loginAs(loginAdmin, passwordAdmin);
      getRuntimeAPI().startTask(null, true);
      fail("Check null argument");
    } catch (IllegalArgumentException be) {
      // ok
    }
    //check throws TaskNotFoundException
    try {
      getRuntimeAPI().startTask(IdFactory.getNewTaskUUID(), true);
      fail("Check non existent task");
    } catch (TaskNotFoundException be) {
      //OK
    }
    checkStopped(instanceUUID, new String[]{});
    Collection<ActivityInstance< TaskInstance >> taskActivities = 
      getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    assertEquals(1, taskActivities.size());
    ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();
    taskUUID = taskActivity.getBody().getUUID();
    assertEquals(ActivityState.READY, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());

    try {
      //start task as the assigned user for this task.
      getRuntimeAPI().startTask(taskUUID, true);
      assertEquals(ActivityState.EXECUTING, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());
      getRuntimeAPI().suspendTask(taskUUID, true);
      assertEquals(ActivityState.SUSPENDED, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!: " + Misc.getStackTraceFrom(e));
    }

    //suspend the task and check IllegalTaskStateException is raised
    try {
      getRuntimeAPI().startTask(taskUUID, true);
      fail("This task has bad state : SUSPENDED instead of READY");
    } catch (IllegalTaskStateException be) {
      //OK
    }

    //resume and try to startTask. =>  check IllegalTaskStateException is raised
    try {
      getRuntimeAPI().resumeTask(taskUUID, false);
      getRuntimeAPI().startTask(taskUUID, true);
      fail("This task has bad state : EXECUTING instead of READY");
    } catch (IllegalTaskStateException be) {
      //OK
    }

    //finish the task => create task b (with candidates= john, admin;   userId = null)
    try {
      getRuntimeAPI().finishTask(taskUUID, true);
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!: " + Misc.getStackTraceFrom(e));
    }

    //LOG as Jack and check this user can't start the task
    try {
      taskActivities = getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
      assertEquals(1, taskActivities.size());
      taskActivity = taskActivities.iterator().next();
      taskUUID = taskActivity.getBody().getUUID();
      assertEquals(ActivityState.READY, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());
      loginAs(loginJack, passwordJack);
      getRuntimeAPI().startTask(taskUUID, true);
      fail("This user [Jack] should not be authorized to start this task ! candidates = "
          + taskActivity.getBody().getTaskCandidates());
    } catch (BonitaException be) {
      //OK
    }

    //LOG as john and check this user can start this task (no user assigned to this task
    //and candidates = john, admin
    //TODO
    try {
      loginAs(loginJohn, passwordJohn);
      getRuntimeAPI().startTask(taskUUID, true);
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!: " + Misc.getStackTraceFrom(e));
    }

    getManagementAPI().undeploy(packageUUID);
  }

  /*
   * Test finishTask(String taskUUID) from the runtimeAPI.
   *
   * test:
   *    - initial state need to be EXECUTING (IllegalTaskStateException raised if state = SUSPENDED)
   *    - if successful task is no more in the task list.
   *    - check that only the user that start the task can finish it
   *      even this user is in the candidates list.
   *    - @throws TaskNotFoundException if the task has not been found.
   *    - @throws IllegalTaskStateException if the state of the task has not EXECUTING state.
   *    - @throws BonitaInternalException if an exception occurs (for ex. with illegal taskUUID parameter).
   *
   */
  public void testFinishTask() throws BonitaException, LoginException {
    String processId = "testGetTask";
    final String loginAdmin = "admin";
    final String passwordAdmin = "bpm";
    final String loginJohn = "john";
    final String passwordJohn = "bpm";
    URL xpdlUrl = this.getClass().getResource("testGetTask.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = null;
    TaskUUID taskUUID = null;
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!");
    }
    //check BonitaInternalException
    try {
      getRuntimeAPI().finishTask(null, true);
      fail("Check null argument");
    } catch (IllegalArgumentException be) {
      // ok
    }
    try {
      getRuntimeAPI().finishTask(IdFactory.getNewTaskUUID(), true);
      fail("Check non existent instance");
    } catch (TaskNotFoundException be) {
      //OK
    }
    checkStopped(instanceUUID, new String[]{});
    Collection<ActivityInstance< TaskInstance >> taskActivities = 
      getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    assertEquals(1, taskActivities.size());
    ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();
    taskUUID = taskActivity.getBody().getUUID();
    assertEquals(ActivityState.READY, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());

    //check IllegalTaskStateException
    //try to finish the task when state = READY
    try {
      getRuntimeAPI().finishTask(taskUUID, true);
      fail("Try to finish this task that has bad state : READY instead of EXECUTING");
    } catch (IllegalTaskStateException be) {
      //OK
    }

    try {
      getRuntimeAPI().suspendTask(taskUUID, true);
      assertEquals(ActivityState.SUSPENDED, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());

    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!");
    }

    //Try to finish it as this task has state = SUSPENDED
    try {
      getRuntimeAPI().finishTask(taskUUID, true);
      fail("Try to finish this task that has bad state : SUSPENDED instead of EXECUTING");
    } catch (IllegalTaskStateException be) {
      //OK
    }

    //resume the task; finish the task with the same user having started the task;
    //check this task is no more in the tasks list.
    //=> create task b (candidates list = admin, john, userId = null)
    try {
      getRuntimeAPI().resumeTask(taskUUID, false);
      getRuntimeAPI().startTask(taskUUID, true);
      getRuntimeAPI().finishTask(taskUUID, true);

      taskActivities = getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
      assertEquals(1, taskActivities.size());
      //there's only the new task.
      assertEquals(1, taskActivities.size());
      taskActivity = taskActivities.iterator().next();
      assertEquals("b", taskActivity.getActivityId());
      assertNotSame(taskUUID, taskActivity.getBody().getUUID());
      taskUUID = taskActivity.getBody().getUUID();
      assertEquals(ActivityState.READY, getQueryRuntimeAPI().getTask(taskUUID).getBody().getState());
      //check the task is not assigned before start
      assertTrue(taskActivity.getBody().getTaskCandidates().containsAll(Arrays.asList(loginAdmin, loginJohn)));
      assertTrue(taskActivity.getBody().getTaskUser() == null);
      //start task b
      getRuntimeAPI().startTask(taskUUID, true);
      taskActivities = getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.EXECUTING);
      assertEquals(1, taskActivities.size());
      taskActivity = taskActivities.iterator().next();
      //task is not assigned after start
      assertTrue(taskActivity.getBody().getTaskCandidates().containsAll(Arrays.asList("admin", "john")));
      assertEquals("admin", taskActivity.getBody().getStartedBy());
    } catch (BonitaException be) {
      fail("Ouch! How can it be?! Should never reach this statement!");
    }

    //task has been started by admin.
    //check a user even in the candidates list cannot finish the task.
    try {
      loginAs(loginJohn, passwordJohn);
      getRuntimeAPI().finishTask(taskUUID, true);
      fail("This user [" + getLogin() + "]should not be authorized to finish this task ! candidates = "
          + taskActivity.getBody().getTaskCandidates());
    } catch (BonitaException be) {
      //OK
    }

    //finish the test.
    loginAs(loginAdmin, passwordAdmin);
    getManagementAPI().undeploy(packageUUID);
  }


  private void loginAs(String loginName, String password) throws LoginException {
    loginContext.logout();
    try {
      loginContext = new LoginContext("Bonita", new SimpleCallbackHandler(loginName, password));
      loginContext.login();
    } catch (LoginException e) {
      throw new RuntimeException("Please, configure a JAAS test user with login: "
          + loginName + " and password: " + password, e);
    }
  }


}
