/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.net.URL;

import org.hibernate.Query;
import org.hibernate.Session;
import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.PackageDefinitionImpl;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.impl.ProcessInstanceImpl;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.pvm.env.Environment;
import org.ow2.bonita.util.ClassDataTool;
import org.ow2.bonita.util.Command;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.Misc;

/**
 * @author Guillaume Porcher
 */


public class CommandAPITest extends APITestCase {

  public void testGetProcessInstance() throws Exception {
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    getManagementAPI().deployClass(ClassDataTool.getClassData(GetProcessInstance.class));
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = null;
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      ProcessInstance processInstance = getCommandAPI().execute(new GetProcessInstance(instanceUUID));
      assertNotNull(processInstance);
      assertEquals(InstanceState.STARTED, processInstance.getInstanceState());
    } finally {
      
      getRuntimeAPI().deleteProcessInstance(instanceUUID);
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
      getManagementAPI().removeClass(GetProcessInstance.class.getName());
    }
  }

  public void testGetProcessInstanceBadArgs() throws Exception {
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();

    getManagementAPI().deployClass(ClassDataTool.getClassData(GetProcessInstance.class));
    try {
      getCommandAPI().execute(new GetProcessInstance(instanceUUID));
      fail("Expected exception");
    } catch (InstanceNotFoundException e) {
      assertEquals(instanceUUID, e.getInstanceUUID());
    }

    try {
      getCommandAPI().execute(new GetProcessInstance(null));
      fail("Expected exception");
    } catch (IllegalArgumentException e) {
      // ok
    }
    getManagementAPI().removeClass(GetProcessInstance.class.getName());
  }
  
  public void testDirectQuery() throws Exception {
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    getManagementAPI().deployClass(ClassDataTool.getClassData(GetPackageDefinition.class));
    PackageDefinition pack = getCommandAPI().execute(new GetPackageDefinition(packageUUID));
    assertNotNull(pack);
    assertEquals(packageUUID, pack.getUUID());
    assertEquals("usertest1_Project", pack.getPackageId());
    
    getManagementAPI().deletePackage(packageUUID);

  }

  private static class GetProcessInstance implements Command<ProcessInstance> {

    /**
     * 
     */
    private static final long serialVersionUID = -336648734170008833L;
    private final ProcessInstanceUUID instanceUUID;

    public GetProcessInstance(final ProcessInstanceUUID instanceUUID) {
      Misc.checkArgsNotNull(instanceUUID);
      this.instanceUUID = instanceUUID;
    }

    public ProcessInstance execute(final Environment env) throws Exception {
      final ProcessInstance result = EngineEnvTool.getAllQueriers().getProcessInstance(this.instanceUUID);
      if (result == null) {
        throw new InstanceNotFoundException(this.instanceUUID);
      }
      return new ProcessInstanceImpl(result);
    }

  }
  
  private static class GetPackageDefinition implements Command<PackageDefinition> {

    /**
     * 
     */
    private static final long serialVersionUID = 7324849491306521922L;
    private final PackageDefinitionUUID packageUUID;

    public GetPackageDefinition(final PackageDefinitionUUID packageUUID) {
      Misc.checkArgsNotNull(packageUUID);
      this.packageUUID = packageUUID;
    }

    public PackageDefinition execute(final Environment env) throws Exception {
      String queryString = "";
      queryString += "select package ";
      queryString += "from org.ow2.bonita.facade.def.majorElement.impl.PackageFullDefinitionImpl as package ";
      queryString += "where package.uuid.value = :packageUUID";

      Session session = env.get(Session.class);
      Query query = session.createQuery(queryString);
      query.setCacheable(true);
      query.setParameter("packageUUID", packageUUID.toString());
      return new PackageDefinitionImpl((PackageDefinition) query.uniqueResult());
    }

  }
  
}
