/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.net.URL;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.dataType.BasicTypeDefinition;
import org.ow2.bonita.facade.def.dataType.DataTypeDefinition;
import org.ow2.bonita.facade.def.dataType.EnumerationTypeDefinition;
import org.ow2.bonita.facade.def.element.FormalParameterDefinition;
import org.ow2.bonita.facade.def.element.SimulationInformationDefinition;
import org.ow2.bonita.facade.def.element.TimeEstimationDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition.PackageState;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.exception.ActivityDefNotFoundException;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DataFieldNotFoundException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ParticipantNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ParticipantDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.services.record.DefHook1;
import org.ow2.bonita.services.record.DefHook2;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.Misc;


/**
 * @author Pierre Vigneras
 * @date Jan 31, 2008
 */
public class DefinitionAPITest extends APITestCase {

  private static final String  PACKAGE_ID = "definitionTest_Project";
  private static final String  PROCESS1_ID = "definitionTest";
  private static final String  PROCESS1_NAME = "definitionTestName";
  private static final String  PROCESS2_ID = "definitionTest2";


  public DefinitionAPITest() {
    super();
  }

  public void testDefinitionAPI() throws BonitaException {
    final String resourceName = "definitionTest.xpdl";
    //final String[] processDefinitionUUID = new String[] {"processA", "processB", "processC"};
    final URL xpdlUrl = this.getClass().getResource(resourceName);
    Misc.badStateIfNull(xpdlUrl, "Can't find URL for resource name: " + resourceName);
    getManagementAPI().deployBar(getBarContent(xpdlUrl, DefHook1.class, DefHook2.class, AdminsRoleMapper.class));

    final QueryDefinitionAPI api = getQueryDefinitionAPI();

    Set<ProcessDefinition> deployedProcesses = api.getProcesses(ProcessDefinition.ProcessState.DEPLOYED);
    assertNotNull(deployedProcesses);
    assertEquals(2, deployedProcesses.size());
    ProcessDefinition process1 = api.getProcesses(PROCESS1_ID, ProcessDefinition.ProcessState.DEPLOYED).iterator().next();
    assertNotNull(process1);
    ProcessDefinitionUUID processDefinitionUUID = process1.getUUID();
    assertNotNull(processDefinitionUUID);
    Set<PackageDefinition> deployedPackages = api.getPackages(PackageDefinition.PackageState.DEPLOYED);
    assertNotNull(deployedPackages);
    assertEquals(1, deployedPackages.size());
    PackageDefinition deployedPackage = api.getPackages(PACKAGE_ID, PackageDefinition.PackageState.DEPLOYED).iterator().next();
    assertNotNull(deployedPackage);

    PackageDefinitionUUID packageUUID = deployedPackage.getUUID();

    //assert package when deployed
    assertPackage(api, packageUUID, false);

    //assert participants
    assertParticipants(api, processDefinitionUUID);

    //Undeploy
    getManagementAPI().undeploy(packageUUID);

    deployedProcesses = api.getProcesses(ProcessDefinition.ProcessState.DEPLOYED);
    assertNotNull(deployedProcesses);
    assertEquals(0, deployedProcesses.size());
    assertEquals(0, api.getProcesses(PROCESS1_ID, ProcessDefinition.ProcessState.DEPLOYED).size());

    deployedPackages = api.getPackages(PackageDefinition.PackageState.DEPLOYED);
    assertNotNull(deployedPackages);
    assertEquals(0, deployedPackages.size());
    assertEquals(0, api.getPackages(PACKAGE_ID, PackageDefinition.PackageState.DEPLOYED).size());

    //assert package when deployed
    assertPackage(api, packageUUID, true);

    getManagementAPI().deletePackage(packageUUID);
  }

  private void assertPackage(QueryDefinitionAPI api, PackageDefinitionUUID packageUUID, boolean undeployed)
    throws PackageNotFoundException, ProcessNotFoundException, ActivityNotFoundException {
    Set<PackageDefinition> set = api.getPackages();
    assertNotNull(set);
    assertEquals(set.toString(), 1, set.size());
    set = api.getPackages(PACKAGE_ID);
    assertNotNull(set);
    assertEquals(set.toString(), 1, set.size());
    PackageDefinition def = null;
    if (!undeployed) {
      def = api.getLastPackage(PACKAGE_ID);
      assertEquals(PackageState.DEPLOYED, def.getState());
    } else {
      Set<PackageDefinition> packages = api.getPackages(PACKAGE_ID, PackageDefinition.PackageState.UNDEPLOYED);
      assertNotNull(packages);
      assertFalse(packages.isEmpty());
      assertEquals(1, packages.size());
      def = packages.iterator().next();
      assertEquals(PackageState.UNDEPLOYED, def.getState());
    }
    //test invalid
    try {
      api.getLastPackage("bidon");
    } catch (PackageNotFoundException pe) {
      //OK
    }

    assertNotNull(def);
    PackageDefinition def2 = api.getPackage(packageUUID);
    assertNotNull(def2);
    assertEquals("Wrong packageDefinitionUUID", packageUUID, def2.getPackageDefinitionUUID());
    assertEquals("Author", "Romulus", def.getAuthor());
    assertEquals("Description" , "My Package Description", def.getDescription());
    assertEquals("CodePage", "UTF8", def.getCodepage());
    assertEquals("Conformance class", PackageDefinition.GraphConformance.NON_BLOCKED, def.getConformanceClass());
    //hooks
    Set<String> classDeps = def.getClassDependencies();
    assertNotNull("Class dependencies", classDeps);
    assertTrue(classDeps.toString(), def.getClassDependencies().size() == 3);
    assertTrue("DefHook1 is in class dependencies", classDeps.contains(DefHook1.class.getName()));
    assertTrue("DefHook2 is in class dependencies", classDeps.contains(DefHook2.class.getName()));
    assertTrue("AdminsRoleMapper is in class dependencies", classDeps.contains(AdminsRoleMapper.class.getName()));
    assertEquals("Cost unit", "My Cost Unit", def.getCostUnit());
    assertEquals("Country key", "fr", def.getCountrykey());
    assertEquals("Created", "My Package Created", def.getCreated());
    assertEquals("Deployed by", "admin", def.getDeployedBy());
    assertNotNull("Deployed Date", def.getDeployedDate());
    assertEquals("Documentation", "My Package Documentation", def.getDocumentation());
//    assertNotNull("External packages", def.getExternalPackages());
    assertEquals("Id", PACKAGE_ID, def.getPackageId());
    assertEquals("Name", "definitionTest", def.getName());
//    assertNotNull("Dependencies", def.getPackageDependencies());
    assertEquals("Priority Unit", "My Priority Unit", def.getPriorityUnit());
    assertEquals("Publication status", PackageDefinition.PublicationStatus.UNDER_TEST, def.getPublicationStatus());
    assertNotNull("Responsibles", def.getResponsibles());
    assertTrue("Responsibles", def.getResponsibles().size() == 2);
//    assertEquals("Script Grammar", "My Script Grammar", def.getScriptGrammar());
    assertEquals("Script Type", "bsh", def.getScriptType());
    assertEquals("Script Version", "2.0b1", def.getScriptVersion());
//    assertNotNull("Type Declarations", def.getTypeDeclarations());
    if (!undeployed) {
      assertNull("Undeployed By", def.getUndeployedBy());
      assertNull("Undeployed Date", def.getUndeployedDate());
    } else {
      assertNotNull("Undeployed By", def.getUndeployedBy());
      assertNotNull("Undeployed Date", def.getUndeployedDate());
    }
    assertEquals("Vendor", "Bonita Project Team", def.getVendor());
    assertEquals("Version", "0.1", def.getVersion());
    assertEquals("XPDL Version", "1.0", def.getXPDLVersion());

    //not valid parameter
    PackageDefinitionUUID badPackageDefinitionUUID = IdFactory.getNewPackageUUID();
    try {
      api.getPackage(badPackageDefinitionUUID);
    } catch (PackageNotFoundException pe) {
      //OK
    }

    assertProcesses(api, packageUUID, undeployed);
  }

  private void assertProcesses(QueryDefinitionAPI api, PackageDefinitionUUID packageUUID, boolean undeployed)
    throws PackageNotFoundException, ProcessNotFoundException, ActivityNotFoundException {
    PackageDefinitionUUID badPackageDefinitionUUID = IdFactory.getNewPackageUUID();
    Set<ProcessDefinition> processes = api.getPackageProcesses(packageUUID);
    assertNotNull(processes);
    processes = api.getProcesses();

    //not valid parameters
    try {
      processes = api.getPackageProcesses(badPackageDefinitionUUID);
    } catch (PackageNotFoundException pe) {
      //OK
    }


    assertEquals(processes.toString(), 2, processes.size());
    Iterator<ProcessDefinition> it = processes.iterator();
    while (it.hasNext()) {
      ProcessDefinition process = it.next();
      assertTrue("Invalid process processDefinitionUUID",
          process.getProcessId().equals(PROCESS1_ID) || process.getProcessId().equals(PROCESS2_ID));
    }
    processes = api.getProcesses(PROCESS1_ID);
    assertNotNull(processes);
    assertEquals(1, processes.size());
    ProcessDefinition process1 = processes.iterator().next();
    assertNotNull(process1);
    assertEquals("definitionTest", process1.getProcessId());

    // not valid parameter
    ProcessDefinitionUUID badProcessDefinitionUUID = IdFactory.getNewProcessUUID();
    try {
      process1 = api.getProcess(badProcessDefinitionUUID);
    } catch (ProcessNotFoundException pe) {
      //OK
    }

    process1 = api.getProcess(process1.getUUID());
    assertNotNull(process1);
    assertEquals(PROCESS1_ID, process1.getProcessId());

    //invalid parameter
    try {
      api.getPackageProcess(badPackageDefinitionUUID, PROCESS1_ID);
    } catch (PackageNotFoundException pe) {
      //OK
    }
    try {
      api.getPackageProcess(packageUUID, "bidon");
    } catch (ProcessNotFoundException pe) {
      //OK
    }

    assertNotNull(api.getPackageProcess(packageUUID, PROCESS1_ID));

    
    if (!undeployed) {
      process1 = api.getLastProcess(PACKAGE_ID, PROCESS1_ID);
      assertEquals(ProcessState.DEPLOYED, process1.getState());
    } else {
      Set<ProcessDefinition> list = api.getProcesses(PROCESS1_ID, ProcessState.UNDEPLOYED);
      assertNotNull(list);
      assertFalse(list.isEmpty());
      assertEquals(1, list.size());
      process1 = list.iterator().next();
      assertEquals(ProcessState.UNDEPLOYED, process1.getState());
    }
    
    assertNotNull(process1);
    assertEquals(PROCESS1_ID, process1.getProcessId());

    processes = api.getProcesses(PROCESS2_ID);
    assertNotNull(processes);
    assertEquals(1, processes.size());
    ProcessDefinition process2 = processes.iterator().next();
    assertNotNull(process2);

    processes = api.getProcesses();
    assertNotNull(processes);
    assertEquals(processes.toString(), 2, processes.size());

    processes = api.getProcesses(PROCESS1_ID);
    assertNotNull(processes);
    assertEquals(processes.toString(), 1, processes.size());

    processes = api.getProcesses(PROCESS2_ID);
    assertNotNull(processes);
    assertEquals(processes.toString(), 1, processes.size());


    assertNotNull(process1);
    assertEquals(PROCESS1_ID, process1.getProcessId());
    assertEquals(PROCESS1_NAME, process1.getName());

    assertNotNull(process2);
    assertEquals(PROCESS2_ID, process2.getProcessId());
    //by default, if no name is specified, name is filled with processDefinitionUUID
    assertEquals(PROCESS2_ID, process2.getName());
    assertEquals("AccessLevel", ProcessDefinition.AccessLevel.PRIVATE, process2.getAccessLevel());


    assertProcess1(api, packageUUID);
  }

  private void assertProcess1(QueryDefinitionAPI api, PackageDefinitionUUID packageUUID)
    throws ProcessNotFoundException, ActivityNotFoundException {
    Set<ProcessDefinition> processes = api.getProcesses(PROCESS1_ID);
    assertNotNull(processes);
    assertEquals(1, processes.size());
    ProcessDefinition process1 = processes.iterator().next();
    assertEquals("processDefinitionUUID", PROCESS1_ID, process1.getProcessId());
    assertEquals("name", PROCESS1_NAME, process1.getName());
    assertEquals("packageDefinitionUUID", packageUUID, process1.getPackageDefinitionUUID());

    assertEquals("AccessLevel", ProcessDefinition.AccessLevel.PUBLIC, process1.getAccessLevel());
    assertEquals("Author", "P_Author", process1.getAuthor());
    assertEquals("Codepage", "P_Codepage", process1.getCodepage());
    assertEquals("Countrykey", "P_Countrykey", process1.getCountrykey());
    assertEquals("Created", "P_Created", process1.getCreated());
    assertEquals("Description", "P_Description", process1.getDescription());
    assertEquals("DurationUnit", ProcessDefinition.DurationUnit.Y, process1.getDurationUnit());
    assertEquals("Limit", "P_Limit", process1.getLimit());
    assertEquals("Priority", "P_Priority", process1.getPriority());
    assertEquals("PublicationStatus", PackageDefinition.PublicationStatus.UNDER_REVISION, process1.getPublicationStatus());
    assertEquals("ValidFrom", "P_ValidFrom", process1.getValidFrom());
    assertEquals("ValidTo", "P_ValidTo", process1.getValidTo());
    assertEquals("Version", "P_Version", process1.getVersion());
    assertEquals("Priority", "P_Priority", process1.getPriority());
    TimeEstimationDefinition timeEstimation = process1.getTimeEstimation();
    assertNotNull(timeEstimation);
    assertEquals("Duration", "P_Duration", timeEstimation.getDuration());
    assertEquals("WaitingTime", "P_WaitingTime", timeEstimation.getWaitingTime());
    assertEquals("WorkingTime", "P_WorkingTime", timeEstimation.getWorkingTime());
    Set<String> responsibles = process1.getResponsibles();
    assertNotNull(responsibles);
    assertEquals(2, responsibles.size());
    assertTrue(responsibles.contains("P_Responsible1"));
    assertTrue(responsibles.contains("P_Responsible2"));
    List<FormalParameterDefinition> formalParameters = process1.getFormalParameters();
    assertNotNull(formalParameters);
    assertEquals(2, formalParameters.size());

    FormalParameterDefinition formalParameter1 = formalParameters.get(0);
    assertEquals("fp1", formalParameter1.getId());
    assertEquals("fp1_Description", formalParameter1.getDescription());
    assertEquals(DataTypeDefinition.Type.BasicType, formalParameter1.getDataType().getType());
    BasicTypeDefinition basicType = (BasicTypeDefinition) formalParameter1.getDataType().getValue();
    assertEquals(BasicTypeDefinition.Type.STRING, basicType.getType());

    FormalParameterDefinition formalParameter2 = formalParameters.get(1);
    assertEquals("fp2", formalParameter2.getId());
    assertEquals("fp2_Description", formalParameter2.getDescription());
    assertEquals(DataTypeDefinition.Type.EnumerationType, formalParameter2.getDataType().getType());
    EnumerationTypeDefinition enumType = (EnumerationTypeDefinition) formalParameter2.getDataType().getValue();
    assertEquals(1, enumType.getEnumerationValues().size());
    assertEquals("Enum1", enumType.getEnumerationValues().iterator().next());

    assertActivities(api, packageUUID, process1.getUUID());
  }

  private void assertActivities(QueryDefinitionAPI api, PackageDefinitionUUID packageUUID,
      ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException, ActivityNotFoundException {
    //invalid parameter
    ProcessDefinitionUUID badProcessDefinitionUUID = IdFactory.getNewProcessUUID();
    try {
      api.getProcessActivities(badProcessDefinitionUUID);
      fail("This process must not be found");
    } catch (ProcessNotFoundException pe) {
      //OK
    }

    Set<ActivityDefinition> activities = api.getProcessActivities(processDefinitionUUID);
    assertNotNull(activities);
    assertEquals(2, activities.size());

    //invalid parameter
    try {
      api.getProcessActivity(badProcessDefinitionUUID, "Activity_A");
      fail("This process must not be found");
    } catch (ProcessNotFoundException pe) {
      //OK
    }
    try {
      api.getProcessActivity(processDefinitionUUID, "bidon");
      fail("This process must not be found");
    } catch (ActivityNotFoundException ae) {
      //OK
    }

    //invalid parameter
    try {
      api.getProcessActivityId(badProcessDefinitionUUID, "Activity A Name");
    } catch (ProcessNotFoundException pe) {
      //OK
    }

    ActivityDefinition activityA = api.getProcessActivity(processDefinitionUUID, "Activity_A");
    assertNotNull(activityA);
    assertEquals("Activity_A", activityA.getActivityId());
    assertEquals("getProcessActivityId", api.getProcessActivityId(processDefinitionUUID, "Activity A Name"), activityA.getUUID());
    assertEquals("PackageId", packageUUID, activityA.getPackageDefinitionUUID());
    assertEquals("Description", "A_Description", activityA.getDescription());
    assertEquals("Documentation", "A_Documentation", activityA.getDocumentation());
    assertEquals("FinishMode", ActivityDefinition.FinishMode.Manual, activityA.getFinishMode());
    assertEquals("Icon", "A_Icon", activityA.getIcon());
    //assertEquals("Limit", "A_Limit", activityA.getLimit());
    assertEquals("Name", "Activity A Name", activityA.getName());
    assertEquals("Performer", "Caesar", activityA.getPerformer());
    assertEquals("StartMode", ActivityDefinition.StartMode.Manual, activityA.getStartMode());
    //assertEquals("Priority", "A_Priority", activityA.getPriority());
    SimulationInformationDefinition simulationInformation = activityA.getSimulationInformation();
    assertNotNull(simulationInformation);
    assertNotNull(simulationInformation.getInstantiation());
    assertEquals(SimulationInformationDefinition.Instantiation.ONCE, simulationInformation.getInstantiation());
    TimeEstimationDefinition timeEstimation = simulationInformation.getTimeEstimation();
    assertNotNull(timeEstimation);
    assertEquals("Duration", "D", timeEstimation.getDuration());
    assertEquals("WaitingTime", "WAT", timeEstimation.getWaitingTime());
    assertEquals("WorkingTime", "WOT", timeEstimation.getWorkingTime());
    assertEquals("Cost", "Cost", simulationInformation.getCost());
/*
    assertEquals("", "", activityA.get());
    assertEquals("", "", activityA.get());
    assertEquals("", "", activityA.get());


    activityA.getBlockId()
    activityA.getDataFields()
    activityA.getDeadlines()

    activityA.getHooks()
    activityA.getPerformerAssign()

    activityA.getSubFlow()
    activityA.getTools()
    activityA.getTransitionRestrictions()
*/
  }
  private void assertParticipants(QueryDefinitionAPI api,
      ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException, ParticipantNotFoundException {
    String humanParticipantId = "Remus";
    String roleParticipantId = "manager";
    //constructs bad UUIIDs
    ProcessDefinitionUUID badProcessDefinitionUUID = IdFactory.getNewProcessUUID();

    //------test getProcessParticipants()----------
    Set<ParticipantDefinition> participants = api.getProcessParticipants(processDefinitionUUID);
    assertNotNull(participants);
    assertEquals(3, participants.size());

    try {
      participants = api.getProcessParticipants(badProcessDefinitionUUID);
    } catch (ProcessNotFoundException pe) {
      //OK
    }

    //--------- test getProcessParticipant() - Human ---------
    ParticipantDefinition participantHuman1 = api.getProcessParticipant(processDefinitionUUID, humanParticipantId);
    assertNotNull(participantHuman1);
    assertEquals("ParticipantId", humanParticipantId, participantHuman1.getParticipantId());
    assertEquals("ParticipantType", ParticipantDefinition.ParticipantType.HUMAN, participantHuman1.getParticipantType());
    // test getProcessParticipant() - Role with role mapper
    ParticipantDefinition participantRole1 = api.getProcessParticipant(processDefinitionUUID, roleParticipantId);
    assertNotNull(participantRole1);
    assertEquals("ParticipantType", ParticipantDefinition.ParticipantType.ROLE, participantRole1.getParticipantType());
    assertNotNull("DefinitionMapper", participantRole1.getRoleMapper());

    assertNotNull("ClassName in RoleMapperDefinition is not set", participantRole1.getRoleMapper().getClassName());
    assertTrue("className in role mapper", participantRole1.getRoleMapper().getClassName()
        .equals(AdminsRoleMapper.class.getName()));

    try {
      participantHuman1 = api.getProcessParticipant(badProcessDefinitionUUID, humanParticipantId);
    } catch (ProcessNotFoundException pe) {
      //OK
    }
    try {
      participantHuman1 = api.getProcessParticipant(processDefinitionUUID, "toto");
    } catch (ParticipantNotFoundException pe) {
      //OK
    }


    //--------- test getProcessParticipantId() ---------
    ParticipantDefinitionUUID  participantDefinitionUUID = api.getProcessParticipantId(processDefinitionUUID, humanParticipantId);
    assertEquals("ParticipantDefinitionUUID", participantDefinitionUUID, participantHuman1.getUUID());
    try {
      participantHuman1 = api.getProcessParticipant(badProcessDefinitionUUID, humanParticipantId);
    } catch (ProcessNotFoundException pe) {
      //OK
    }
  }

  public void testGetVariableDefinitions() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processDefinitionUUID = process.getUUID();

    ProcessDefinitionUUID badProcessDefinitionUUID = IdFactory.getNewProcessUUID();

    // Check process variable definitions
    Set<DataFieldDefinition> datafields = getQueryDefinitionAPI().getProcessDataFields(processDefinitionUUID);
    assertFalse(datafields.isEmpty());
    assertEquals(2, datafields.size());

    DataFieldDefinition stringProcessDataField = null;
    DataFieldDefinition enumProcessDataField = null;
    DataFieldDefinition stringActivityDataField = null;
    DataFieldDefinition enumActivityDataField = null;

    for (DataFieldDefinition datafield : datafields) {
      assertNotNull(datafield);
      assertNotNull(datafield.getDataFieldId());
      String id = datafield.getDataFieldId();

      if (id.equals("string_process")) {
        stringProcessDataField = datafield;
      } else if (id.equals("enum_process")) {
        enumProcessDataField = datafield;
      } else if (id.equals("string_activity")) {
        stringActivityDataField = datafield;
      } else if (id.equals("enum_activity")) {
        enumActivityDataField = datafield;
      }
    }

    assertNotNull(stringProcessDataField);
    assertNotNull(enumProcessDataField);
    assertNull(stringActivityDataField);
    assertNull(enumActivityDataField);

    assertTrue(stringProcessDataField.getDataFieldId().equals("string_process"));
    assertTrue(stringProcessDataField.getInitialValue().equals("initial string value"));
    assertTrue(enumProcessDataField.getDataFieldId().equals("enum_process"));
    assertTrue(enumProcessDataField.getDataType().getType().name().equals("EnumerationType"));

    //Check ProcessNotFoundException
    try {
      getQueryDefinitionAPI().getProcessDataFields(badProcessDefinitionUUID);
      fail("check ProcessNotFoundException is raised");
    } catch (ProcessNotFoundException e) {
      assertEquals(badProcessDefinitionUUID, e.getProcessUUID());
    }

    //******* test the method getProcessDataField()
    //string type
    DataFieldDefinition stringProcessDef = getQueryDefinitionAPI().getProcessDataField(processDefinitionUUID, "string_process");
    assertNotNull(stringProcessDef.getUUID());
    assertEquals("string_process", stringProcessDef.getDataFieldId());
    assertEquals("string_process", stringProcessDef.getName());
    assertEquals(DataTypeDefinition.Type.BasicType, stringProcessDef.getDataType().getType());
    assertEquals("initial string value", stringProcessDef.getInitialValue());
    assertEquals(null, stringProcessDef.getLength());
    assertEquals("string for process", stringProcessDef.getDescription());

    //enumeration type
    DataFieldDefinition enumerationProcessDef = getQueryDefinitionAPI()
      .getProcessDataField(processDefinitionUUID, "enum_process");
    assertEquals(DataTypeDefinition.Type.EnumerationType, enumerationProcessDef.getDataType().getType());

    //Check ProcessNotFoundException & DataFieldNotFoundException
    try {
      getQueryDefinitionAPI().getProcessDataField(badProcessDefinitionUUID, "string_process");
      fail("check raisinf exception : ProcessNotFoundException");
    } catch (ProcessNotFoundException e) {
      assertEquals(badProcessDefinitionUUID, e.getProcessUUID());
    }
    try {
      getQueryDefinitionAPI().getProcessDataField(processDefinitionUUID, "bad_string_process");
      fail("check raisinf exception : ProcessNotFoundException");
    } catch (DataFieldNotFoundException e) {
      assertEquals("bad_string_process", e.getDataFieldId());
    }

    // activity variables contains: process variables + local variables (act1 defines no local vars)

    //***** test getActivityDataFields & getActivityDataField
    /* Activity 1 */
    ActivityDefinitionUUID activityUUID = getQueryDefinitionAPI().getProcessActivityId(processDefinitionUUID, "act1");
    datafields = getQueryDefinitionAPI().getActivityDataFields(activityUUID);
    assertTrue(datafields.isEmpty());

    /* Activity 2 */
    activityUUID = getQueryDefinitionAPI().getProcessActivityId(processDefinitionUUID, "act2");
    datafields = getQueryDefinitionAPI().getActivityDataFields(activityUUID);
    assertFalse(datafields.isEmpty());
    assertEquals(2, datafields.size());

    stringProcessDataField = null;
    enumProcessDataField = null;
    stringActivityDataField = null;
    enumActivityDataField = null;

    for (DataFieldDefinition datafield : datafields) {
      assertNotNull(datafield);
      assertNotNull(datafield.getDataFieldId());
      String id = datafield.getDataFieldId();
      if (id.equals("string_process")) {
        stringProcessDataField = datafield;
      } else if (id.equals("enum_process")) {
        enumProcessDataField = datafield;
      } else if (id.equals("string_activity")) {
        stringActivityDataField = datafield;
      } else if (id.equals("enum_activity")) {
        enumActivityDataField = datafield;
      }
    }
    assertNull(stringProcessDataField);
    assertNull(enumProcessDataField);
    assertNotNull(stringActivityDataField);
    assertNotNull(enumActivityDataField);

    //Testing getActivityDataField method
    DataFieldDefinition stringActivityDef = getQueryDefinitionAPI().getActivityDataField(activityUUID, "string_activity");
    assertNotNull(stringActivityDef.getUUID());
    assertEquals("string_activity", stringActivityDef.getDataFieldId());
    assertEquals("string_activity", stringActivityDef.getName());
    assertEquals(DataTypeDefinition.Type.BasicType, stringActivityDef.getDataType().getType());
    assertEquals("initial value", stringActivityDef.getInitialValue());
    assertEquals(null, stringActivityDef.getLength());
    assertEquals("string for activity", stringActivityDef.getDescription());

    //enumeration type
    DataFieldDefinition enumerationActivityDef = getQueryDefinitionAPI()
      .getActivityDataField(activityUUID, "enum_activity");
    assertEquals(DataTypeDefinition.Type.EnumerationType, enumerationActivityDef.getDataType().getType());

    //Check exceptions :ActivityDefNotFoundException, DataFieldNotFoundException
    ActivityDefinitionUUID badActivityUUID = IdFactory.getNewActivityDefinitionUUID();
    try {
      getQueryDefinitionAPI().getActivityDataField(badActivityUUID, "string_activity");
      fail("check the exception ActivityDefNotFoundException is raised");
    } catch (ActivityDefNotFoundException e) {
      assertEquals(badActivityUUID, e.getActivityUUID());
    }

    try {
      getQueryDefinitionAPI().getActivityDataField(activityUUID, "string_process");
      fail("check the exception DataFieldNotFoundException is raised");
    } catch (DataFieldNotFoundException e) {
      assertEquals("string_process", e.getDataFieldId());
    }
    /* Activity 3 */
    activityUUID = getQueryDefinitionAPI().getProcessActivityId(processDefinitionUUID, "act3");
    datafields = getQueryDefinitionAPI().getActivityDataFields(activityUUID);
    assertTrue(datafields.isEmpty());

    //check ActivityDefNotFoundException for get getActivityDataFields
    ActivityDefinitionUUID badActivityDefinitionUUID = IdFactory.getNewActivityDefinitionUUID();
    try {
      getQueryDefinitionAPI().getActivityDataFields(badActivityDefinitionUUID);
    } catch (ActivityDefNotFoundException e) {
      assertEquals(badActivityDefinitionUUID, e.getActivityUUID());
    }

    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }


  /*
   still not tested :

  Set<ParticipantDefinition> getProcessParticipants(String processDefinitionUUID, String deploymentId);
  ParticipantDefinition getProcessParticipant(String processDefinitionUUID, String deploymentId, String participantUUID);
  String getProcessParticipantId(String processDefinitionUUID, String deploymentId, String participantName);

  Set<ActivityDefinition> getProcessActivities(String processDefinitionUUID, String deploymentId);
  ActivityDefinition getProcessActivity(String processDefinitionUUID, String deploymentId, String activityId);
  String getProcessActivityId(String processDefinitionUUID, String deploymentId, String activityId);

   */
}
