/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import junit.framework.TestCase;

import org.ow2.bonita.facade.internal.InternalCommandAPI;
import org.ow2.bonita.facade.internal.InternalQueryDefinitionAPI;
import org.ow2.bonita.facade.internal.InternalQueryRuntimeAPI;
import org.ow2.bonita.facade.internal.InternalRuntimeAPI;
import org.ow2.bonita.util.BonitaRuntimeException;


/**
 * @author Guillaume Porcher
 *
 */
public class InternalAPITest extends TestCase {

  protected void checkCompatible(final Class< ? > clientApi,
      final Class< ? > internalApi, final boolean hasQueryList) throws NoSuchMethodException {
    
    //check all methods in client API (RuntimeAPI) has a corresponding method in internal API (InternalRuntimeAPI)
    for (final Method clientApiMethod : clientApi.getMethods()) {
      Method internalApiMethod = null;
      Class< ? >[] internalApiMethodParams;
      if (hasQueryList) {
        final Class< ? >[] clientApiMethodParams = clientApiMethod.getParameterTypes();
        if (clientApiMethodParams != null) {
          internalApiMethodParams = new Class< ? >[clientApiMethodParams.length + 1];
          for (int i = 0; i < clientApiMethodParams.length; i++) {
            internalApiMethodParams[i] = clientApiMethodParams[i];
          }
          internalApiMethodParams[clientApiMethodParams.length] = String.class;
        } else {
          internalApiMethodParams = new Class< ? >[] {String.class};
        }
      } else {
        internalApiMethodParams = clientApiMethod.getParameterTypes();
      }
      try {
        internalApiMethod = internalApi.getMethod(clientApiMethod.getName(), internalApiMethodParams);
      } catch (final NoSuchMethodException e) {
        fail("No internal method matches with client method: " + clientApiMethod);
      }
      assertNotNull(internalApiMethod);
      final Set<Class< ? >> internalApiMethodExceptions = 
        new HashSet<Class< ? >>(Arrays.asList(internalApiMethod.getExceptionTypes()));
      final Set<Class< ? >> clientApiMethodExceptions = 
        new HashSet<Class< ? >>(Arrays.asList(clientApiMethod.getExceptionTypes()));
      clientApiMethodExceptions.add(RemoteException.class);
      assertEquals("Invalid exception in method: " + clientApiMethod, internalApiMethodExceptions, clientApiMethodExceptions);
    }
    
    //check all methods in internal API (InternalRuntimeAPI) has a corresponding method in client API (RuntimeAPI)
    for (final Method internalApiMethod : internalApi.getMethods()) {
      Method clientApiMethod = null;
      Class< ? >[] clientApiMethodParams;
      if (hasQueryList) {
        final Class< ? >[] p = internalApiMethod.getParameterTypes();
        if (p != null) {
          clientApiMethodParams = new Class< ? >[p.length - 1];
          for (int i = 0; i < p.length - 1; i++) {
            clientApiMethodParams[i] = p[i];
          }
        } else {
          throw new BonitaRuntimeException("invalid method " + internalApiMethod);
        }
      } else {
        clientApiMethodParams = internalApiMethod.getParameterTypes();
      }
      try {
        clientApiMethod = clientApi.getMethod(internalApiMethod.getName(), clientApiMethodParams);
      } catch (final NoSuchMethodException e) {
        fail("No client method matches with internal method: " + internalApiMethod);
      }
      assertNotNull(clientApiMethod);
      final Set<Class< ? >> clientApiMethodExceptions = 
        new HashSet<Class< ? >>(Arrays.asList(clientApiMethod.getExceptionTypes()));
      final Set<Class< ? >> internalApiMethodExceptions = 
        new HashSet<Class< ? >>(Arrays.asList(internalApiMethod.getExceptionTypes()));
      clientApiMethodExceptions.add(RemoteException.class);
      assertEquals("Invalid exception in method: " + clientApiMethod, clientApiMethodExceptions, internalApiMethodExceptions);
    }
  }

  public void testRuntimeAPI() throws Exception {
    checkCompatible(RuntimeAPI.class, InternalRuntimeAPI.class, false);
  }

  public void testQueryDefintionAPI() throws Exception {
    checkCompatible(QueryDefinitionAPI.class, InternalQueryDefinitionAPI.class, true);
  }

  public void testQueryRuntimeAPI() throws Exception {
    checkCompatible(QueryRuntimeAPI.class, InternalQueryRuntimeAPI.class, true);
  }

  public void testCommandAPI() throws Exception {
    checkCompatible(CommandAPI.class, InternalCommandAPI.class, false);
  }
}
