package org.ow2.bonita.facade;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.BusinessArchiveFactory;
import org.ow2.bonita.util.ClassDataTool;
import org.ow2.bonita.util.Misc;

public class ManagementAPITest extends APITestCase {

  private static final String PROCESS_ID = "process1";
  private static final String XPDLFILE = "testManagementAPI.xpdl";
  private static final String XPDLFILEV2 = "testManagementAPIV2.0.xpdl";
  private static final String XPDLCORRUPTED = "testManagementAPIWithParsingError.xpdl";

  @SuppressWarnings("unchecked")
  public void testDeployCommandInBar() throws BonitaException, IOException {
    final String processId = "testManagementAPICommand";
    final String xpdlFile = processId + ".xpdl";

    final Class[] classes = {GetProcessIdCommand.class};
    final URL url = this.getClass().getResource(xpdlFile);
    final byte[] xpdlContent = Misc.getAllContentFrom(url);

    //**** Test deployBar(bytesBar) with valid parameter:

    final byte[] barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(xpdlFile, xpdlContent, classes);
    final Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    assertNotNull(mapProcessDefinitions);
    assertEquals(1, mapProcessDefinitions.size());
    final ProcessDefinition process = mapProcessDefinitions.get(processId);
    final PackageDefinitionUUID packageDefinitionUUID = process.getPackageDefinitionUUID();

    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      final String returnedProcessId = getCommandAPI().execute(new GetProcessIdCommand(processUUID), packageDefinitionUUID);
      assertEquals(processId, returnedProcessId);
    } catch (final Exception e) {
      if (!(e instanceof ProcessNotFoundException)) {
        throw new BonitaRuntimeException(e);
      }
      throw (ProcessNotFoundException) e;
    } finally {
      getManagementAPI().undeploy(packageDefinitionUUID);
      getManagementAPI().deletePackage(packageDefinitionUUID);
    }

  }

  private File createTempFile(final String prefix, final String suffix) throws IOException {
    File file = null;
    boolean succeed = false;
    int i = 0;
    do {
        file = new File(System.getProperty("java.io.tmpdir") + File.separator + prefix + i + "." + suffix);
        if (!file.exists()) {
          file.createNewFile();
          file.deleteOnExit();
          succeed = true;
        }
        i++;
    } while (!succeed);
    return file;
  }
  /*
   * testing deployBar(byte[])
   * testing deployBar(Url)
   */
  @SuppressWarnings("unchecked")
  public void testDeployBar() throws BonitaException, IOException {

    //Common utilities for the test.
    final Class[] emptyClasses = {};
    final Class[] classes = {Accept.class, AcceptGlobal.class};
    final URL url = this.getClass().getResource(XPDLFILE);
    final byte[] xpdlContent = Misc.getAllContentFrom(url);

    //**** Test deployBar(bytesBar) with valid parameter:

    byte[] barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILE, xpdlContent, classes);
    Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    //Check deployment is OK
    PackageDefinitionUUID packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    getManagementAPI().undeploy(packageDefinitionUUID);
    getManagementAPI().deletePackage(packageDefinitionUUID);

    //**** Test deployBar(bytesBar) with valid parameter but empty classes table
    barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILE, xpdlContent, emptyClasses);
    mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    //Check deployment is OK
    packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    getManagementAPI().undeploy(packageDefinitionUUID);

    File file = createTempFile(PROCESS_ID, "bar");
    byte[] barContent = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILE, xpdlContent, classes);
    BusinessArchiveFactory.generateBusinessArchiveFile(file, barContent);
    URL urlBar = new File(file.getAbsolutePath()).toURI().toURL();

    // Do not delete package !!!

    //Test DeploymentException: raised if A process with id process1 and version 1.0 has already been deployed.
    // Please change the version.
    try {
      mapProcessDefinitions = getManagementAPI().deployBar(urlBar);
      packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
      fail("Check DeploymentException raised if same version of the process is deployed twice even if undeployed done.");
    } catch (final DeploymentException de) {
      //assertEquals("A process with id process1 and version 1.0 has already been deployed."
      // + "Please change the version.", de.getMessage());
    }

    getManagementAPI().deletePackage(packageDefinitionUUID);

    //(valid parameter) Check package with new version of processes can be deployed (with also new version of the package)
    file = createTempFile(PROCESS_ID, "bar");

    final URL urlV2 = this.getClass().getResource(XPDLFILEV2);
    final byte[] xpdlContentV2 = Misc.getAllContentFrom(urlV2);
    barContent = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILEV2, xpdlContentV2, classes);
    BusinessArchiveFactory.generateBusinessArchiveFile(file, barContent);

    urlBar = new File(file.getAbsolutePath()).toURI().toURL();

    mapProcessDefinitions = getManagementAPI().deployBar(urlBar);
    packageDefinitionUUID = checkDeployment(mapProcessDefinitions);

    getManagementAPI().deletePackage(packageDefinitionUUID);


    //**** giving a null bar (bytes table)
    final byte[] bar = null;
    try {
      getManagementAPI().deployBar(bar);
      fail("Check null bar");
    } catch (final IllegalArgumentException e) {
      // ok
      assertTrue(e.getMessage().contains("Some parameters are null in org.ow2.bonita.facade.impl.ManagementAPIImpl.deployBar()"));
    }

    //**** giving a bar (bytes table) without xpdl file (and classes)
    final byte[] barWithoutXpdl = Misc.generateJar(classes);
    try {
      getManagementAPI().deployBar(barWithoutXpdl);
      fail("Check DeploymentException raised if no xpdl into the bar file");
    } catch (final DeploymentException e) {
      // ok
      assertTrue(e.getMessage(), e.getMessage().contains("The given bar archive does not contain any '.xpdl' file"));
    }

    //Check more than 1 xpdl
    final URL url1 = this.getClass().getResource(XPDLFILE);
    final URL url2 = this.getClass().getResource(XPDLFILEV2);
    final byte[] xpdlContent1 = Misc.getAllContentFrom(url1);
    final byte[] xpdlContent2 = Misc.getAllContentFrom(url2);
    try {
      final Map<String, byte[]> resources = new HashMap<String, byte[]>();
      resources.put(XPDLFILE, xpdlContent1);
      resources.put(XPDLFILEV2, xpdlContent2);
      resources.putAll(Misc.getResources(classes));

      final byte[] barWith2Xpdl = Misc.generateJar(resources);
      mapProcessDefinitions = getManagementAPI().deployBar(barWith2Xpdl);
      fail("Check DeploymentException is raised if 2 xpdl files");
    } catch (final DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains("The given bar archive contains many '.xpdl' files"));
    }

    //**** Test deployBar(urlBar) with valid parameter

    //**** Check if the URL == null
    urlBar  = null;
    try {
      getManagementAPI().deployBar(urlBar);
      fail("Check null bar");
    } catch (final IllegalArgumentException e) {
      // ok
      assertTrue(e.getMessage().contains("Some parameters are null in org.ow2.bonita.facade.impl.ManagementAPIImpl.deployBar()"));
    }

    //**** Check if the URL contain an empty archive

    final File urlonBarEmptyFile = createTempFile(XPDLFILEV2, "bar");
    urlonBarEmptyFile.deleteOnExit();
    final URL urlonBarEmpty = urlonBarEmptyFile.toURI().toURL();
    try {
      getManagementAPI().deployBar(urlonBarEmpty);
      fail("Check DeploymentException raised if no xpdl into the bar file");
    } catch (final DeploymentException e) {
      // ok
      assertTrue(e.getMessage(), e.getMessage().contains("The given bar archive does not contain any '.xpdl' file"));
    }

    //**** test no archive at the url
    final File tempFile = createTempFile("toto", "ext");
    tempFile.deleteOnExit();
    tempFile.delete();
    final URL urlonNoArchive = tempFile.toURI().toURL();
    try {
      getManagementAPI().deployBar(urlonNoArchive);
      fail("Check DeploymentException raised if no xpdl into the bar file");
    } catch (final DeploymentException e) {
      // ok
    } finally {
      Misc.deleteDir(tempFile);
    }
  }

  /*
   *  Testing deploy with parameters:
   *     - org.ow2.bonita.deployment.Deployment
   *     - java.net.URL, java.lang.Class<?>[]
   */
  @SuppressWarnings("deprecation")
  public void testDeploy() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource(XPDLFILE);
    final Map<String, ProcessDefinition> mapProcessDefinitions =
      getManagementAPI().deployBar(getBarContent(xpdlUrl, Accept.class));
    //Check deployment is OK
    final PackageDefinitionUUID packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    getManagementAPI().undeploy(packageDefinitionUUID);
    getManagementAPI().deletePackage(packageDefinitionUUID);

    // IllegalArgumentException
    try {
      getManagementAPI().deploy(null);
      fail("Check IllegalArgumentException");
    } catch (final IllegalArgumentException e) {
      // ok
    }
  }

  /*
   *  Testing deployZip with parameters:
   *     - java.net.URL
   *     - byte[]
   */
  public void testDeployZip() throws BonitaException {
    //TODO ??
  }

  /*
   *  Testing deployXpdl with parameters:
   *     - java.net.URL
   */
  @SuppressWarnings("deprecation")
  public void testDeployXpdlURL() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource(XPDLFILE);
    Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployXpdl(xpdlUrl);
    //Check deployment is OK
    final PackageDefinitionUUID packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    getManagementAPI().undeploy(packageDefinitionUUID);
    getManagementAPI().deletePackage(packageDefinitionUUID);

    // IllegalArgumentException
    try {
      getManagementAPI().deployXpdl((URL)null);
      fail("Check IllegalArgumentException");
    } catch (final IllegalArgumentException e) {
      // ok
    }

    //check at least a case of bad parsing preventing to deploy the xpdl
    xpdlUrl = this.getClass().getResource(XPDLCORRUPTED);
    try {
      mapProcessDefinitions = getManagementAPI().deployXpdl(xpdlUrl);
    } catch (final DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("errors during parsing of xpdl file"));
    }

  }

  /*
   *  Testing deployXpdl with parameters:
   *     - byte[]
   */
  @SuppressWarnings("deprecation")
  public void testDeployXpdlByteArray() throws BonitaException {
    byte[] xpdlStream = null;

    try {
      xpdlStream = Misc.getAllContentFrom(this.getClass().getResource(XPDLFILE));
    } catch (final IOException e1) {
      fail("unable to build a byte[] from the given xpdl URL");
    }
    Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployXpdl(xpdlStream);
    //Check deployment is OK
    final PackageDefinitionUUID packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    getManagementAPI().undeploy(packageDefinitionUUID);
    getManagementAPI().deletePackage(packageDefinitionUUID);

    // IllegalArgumentException
    try {
      getManagementAPI().deployXpdl((byte[])null);
      fail("Check IllegalArgumentException");
    } catch (final IllegalArgumentException e) {
      // ok
    }

    //check at least a case of bad parsing preventing to deploy the xpdl
    try {
      xpdlStream = Misc.getAllContentFrom(this.getClass().getResource(XPDLCORRUPTED));
      mapProcessDefinitions = getManagementAPI().deployXpdl(xpdlStream);
    } catch (final DeploymentException e) {
      assertTrue(Misc.getStackTraceFrom(e), e.getMessage().contains("errors during parsing of xpdl file"));
    } catch (final IOException e) {
      fail("unable to build a byte[] from the given xpdl URL");
    }

  }

  /*
   * Testing deployClass
   * covers methods with parameters being:
   *     - byte[]
   */
  public void testDeployClass() throws BonitaException, IOException {
    // IllegalArgumentException
    try {
      getManagementAPI().deployClass(null);
      fail("Check null bar");
    } catch (final IllegalArgumentException e) {
      // ok
      assertTrue(e.getMessage().
          contains("Some parameters are null"));
    }

    //Not correct byte table
    final byte[] byteEmpty = {};
    try {
      getManagementAPI().deployClass(byteEmpty);
      fail("Check empty/not correct bar");
    } catch (final DeploymentException e) {
      // ok
      assertEquals("The byte table is not correct parameter.", e.getMessage());
    }

    //valid param
    final String className = "org.ow2.bonita.facade.Accept";
    URL url = this.getClass().getResource("Accept.class");
    final byte[] bytesClasses = Misc.getAllContentFrom(url);
    getManagementAPI().deployClass(bytesClasses);

    //Invalid test case : impossible to deploy 2 times at common ClassLoader
    try {
      getManagementAPI().deployClass(bytesClasses);
      fail("Check impossible to deploy 2 times same class");
    } catch (final DeploymentException e) {
      assertEquals(Misc.getStackTraceFrom(e), "There is already a deployed class with this name: " + className, e.getMessage());
    }
    getManagementAPI().removeClass(className);
    //Valid test case:
    //Deploy the class into the Globalclassloader
    //Try to deploy class into the packageClassloader (class into the deploy bar file)
    getManagementAPI().deployClass(bytesClasses);
    final java.lang.Class< ? >[] classes = {Accept.class};
    url = this.getClass().getResource(XPDLFILE);
    final byte[] xpdlContent = Misc.getAllContentFrom(url);
    final byte[] barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILE, xpdlContent, classes);
    Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    final ProcessDefinition process2 = mapProcessDefinitions.get("process2");
    final ProcessDefinitionUUID process2UUID = process2.getUUID();
    PackageDefinitionUUID packageUUID = process2.getPackageDefinitionUUID();
    final ProcessInstanceUUID instanceUUID1 = getRuntimeAPI().instantiateProcess(process2UUID);
    //Check it's possible to remove the global class (not used for the instantiated process)
    getManagementAPI().removeClass(className);
    //Check instantiate is always possible
    final ProcessInstanceUUID instanceUUID2 = getRuntimeAPI().instantiateProcess(process2UUID);
    getRuntimeAPI().deleteProcessInstance(instanceUUID1);
    getRuntimeAPI().deleteProcessInstance(instanceUUID2);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);

    //Valid test case:
    //Deploy class into the packageClassloader (class into the deploy bar file)
    //Try to deploy the class into the Globalclassloader
    mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    getManagementAPI().deployClass(bytesClasses);

    // undeploy package and classes before next test
    getManagementAPI().removeClass(className);
    packageUUID = mapProcessDefinitions.get("process2").getPackageDefinitionUUID();


    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Testing deployClassesInJar (parameter: byte[])
   */
  public void testDeployClassesInJar() throws BonitaException, IOException {

    // IllegalArgumentException
    try {
      getManagementAPI().deployClassesInJar(null);
      fail("Check null bar");
    } catch (final IllegalArgumentException e) {
      // ok
      assertTrue(e.getMessage().
          contains("Some parameters are null in org.ow2.bonita.facade.impl.ManagementAPIImpl.deployClassesInJar()"));
    }
    //valid parameter
    final java.lang.Class< ? >[] classes = {Accept.class};
    final byte[] bytesBarClasses = Misc.generateJar(classes);
    getManagementAPI().deployClassesInJar(bytesBarClasses);

    getManagementAPI().removeClass(Accept.class.getName());
  }

  /*
   * Testing deployClasses
   * covers methods with parameters being:
   *     - java.util.Set<byte[]>,
   */
  public void testDeployClasses() throws BonitaException, IOException {
    getManagementAPI().deployClasses(ClassDataTool.getClasses(Accept.class));
    getManagementAPI().removeClass(Accept.class.getName());

    // TODO: test with more than one class.
  }

  /*
   * testing replaceClass
   */
  public void testReplaceClass() throws BonitaException, IOException {
    final java.lang.Class< ? >[] classes = {Accept.class};
    final String className = "org.ow2.bonita.facade.Accept";
    final String newClassName = "org.ow2.bonita.facade.Acceptation";


    final URL url = this.getClass().getResource("Accept.class");
    final byte[] bytesClasses = Misc.getAllContentFrom(url);
    final URL newUrl = this.getClass().getResource("Accept.class");
    final byte[] newBytesClasses = Misc.getAllContentFrom(newUrl);
    getManagementAPI().deployClass(bytesClasses);

    //DeploymentException if trying to replace without having globally deploying the new one.
    try {
      getManagementAPI().replaceClass(newClassName, newBytesClasses);
      fail("Check DeploymentException is raised if the new class is not already globally deployed");
    } catch (final DeploymentException e) {
      assertEquals("There is no class defined in global class repository  with name: "
          + "org.ow2.bonita.facade.Acceptation", e.getMessage());
    }

    //valid parameter for replace
    getManagementAPI().removeClass(className);
    getManagementAPI().deployClass(newBytesClasses);
    getManagementAPI().replaceClass(className, newBytesClasses);

    //Invalid parameter : give  bytes[] of a bar file instead of a simple byte[]
    try {
      final byte[] bytesBarClasses = Misc.generateJar(classes);
      getManagementAPI().deployClass(bytesBarClasses);
    } catch (final DeploymentException e) {
      assertEquals("The byte table is not correct parameter.", e.getMessage());
    }

    //Invalid parameter: give an empty byte table to replace a given class
    final byte[] byteEmpty = {};
    try {
      getManagementAPI().replaceClass(className, byteEmpty);
      fail("check DeploymentException");
    } catch (final DeploymentException e) {
      assertEquals("The byte table is not correct parameter.", e.getMessage());
    }

    // undeploy class
    getManagementAPI().removeClass(className);
  }

  /*
   * testing removeClasses(null)
   */
  public void testRemoveClasses() throws BonitaException, IOException {
    final String className = "org.ow2.bonita.facade.Accept";
    final String notExistClassName = "org.ow2.bonita.facade.AcceptPas";

    final URL url = this.getClass().getResource("Accept.class");
    byte[] bytesClasses = Misc.getAllContentFrom(url);
    getManagementAPI().deployClass(bytesClasses);

    //invalid case: not existing class
    try {
      getManagementAPI().removeClass(notExistClassName);
    } catch (final DeploymentException e) {
      assertEquals("There is no class defined in global class repository  "
          + "with name: org.ow2.bonita.facade.AcceptPas", e.getMessage());
    }

    //invalid case : null className
    try {
      getManagementAPI().removeClass(null);
      fail("must get an error with a null className");
    } catch (final IllegalArgumentException e) {
      //ok
    }

    //**** Valid case
    getManagementAPI().removeClass(className);

    //**** Invalid test case: Deploy globally class A. deploy a package using class A. Try to remove class A
    //global deploy of the class : AcceptGlobal
    final URL classUrl = this.getClass().getResource("AcceptGlobal.class");
    bytesClasses = Misc.getAllContentFrom(classUrl);
    getManagementAPI().deployClass(bytesClasses);

    //Process deploy of the class : Accept
    final java.lang.Class< ? >[] processClasses = {Accept.class};
    final URL xpdlUrl = this.getClass().getResource(XPDLFILEV2);
    final byte[] xpdlContent = Misc.getAllContentFrom(xpdlUrl);
    final byte[] barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILEV2, xpdlContent, processClasses);
    final Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);

    //Check deployment is OK
    final PackageDefinitionUUID packageUUID = checkDeployment(mapProcessDefinitions);


    //Invalid test case: impossible to remove a global class still used by a deployed process
    try {
      getManagementAPI().removeClass(AcceptGlobal.class.getName());
      fail("It must be forbidden to undeploy a class used by an existing process");
    } catch (final DeploymentException e) {
      assertEquals("Package is still using global class. "
          + "Cannot undeploy class." + " className: " + e.getClassName()
          + " packageDefinitionUUID: " + e.getPackageDefinitionUUID(), e.getMessage());
    }
    //Invalid test case: impossible to remove a process class (furthermore still used by a process)
    try {
      getManagementAPI().removeClass("org.ow2.bonita.facade.Accept");
      fail("It must be forbidden to undeploy a class used by an existing process");
    } catch (final DeploymentException e) {
      assertEquals("There is no class defined in global class repository  with name: " + Accept.class.getName(), e.getMessage());
    }

    //valid test case: if the package is undeployed the class can be undeployed.
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().removeClass(AcceptGlobal.class.getName());
  }

  /*
   *  testing undeploy
   */
  public void testUndeploy() throws BonitaException, IOException {
    //**** Valid paramleter
    // already done into previous test.

    //**** null parameter
    try {
      getManagementAPI().undeploy(null);
      fail("Check IllegalArgumentException raised if null parameter");
    } catch (final IllegalArgumentException e) {
      // ok
      assertTrue(e.getMessage().contains("Some parameters are null in org.ow2.bonita.facade.impl.ManagementAPIImpl.undeploy()"));
    }

    //**** Not existing Package
    final PackageDefinitionUUID badPackageDefinitionUUID = IdFactory.getNewPackageUUID();

    try {
      getManagementAPI().undeploy(badPackageDefinitionUUID);
      fail("Check DeploymentException raised if the package does not exists in the journal");
    } catch (final DeploymentException de) {
      final String expected = "package " + badPackageDefinitionUUID + " not found in journal!";
      assertTrue("Expected: " + expected + ", result: " + de.getMessage(), de.getMessage().contains(expected));
    }

    //try to undeploy package with not finished instance
    final java.lang.Class< ? >[] processClasses = {Accept.class, AcceptGlobal.class};
    final URL xpdlUrl = this.getClass().getResource(XPDLFILEV2);
    final byte[] xpdlContent = Misc.getAllContentFrom(xpdlUrl);
    final byte[] barWithXpdl = BusinessArchiveFactory.getBusinessArchiveContent(XPDLFILEV2, xpdlContent, processClasses);
    final Map<String, ProcessDefinition> mapProcessDefinitions = getManagementAPI().deployBar(barWithXpdl);
    final PackageDefinitionUUID packageDefinitionUUID = checkDeployment(mapProcessDefinitions);
    final ProcessDefinitionUUID processDefinitionUUID = mapProcessDefinitions.get("process1").getUUID();
    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processDefinitionUUID);
    try {
      getManagementAPI().undeploy(packageDefinitionUUID);
      fail("Check it's not possible to undeploy process having running instances");
    } catch (final DeploymentException e) {
      final String expected = "Error during undeployment: there are still running instances";
      assertTrue("Expected: " + expected, e.getMessage().contains(expected));
    }
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageDefinitionUUID);
    getManagementAPI().deletePackage(packageDefinitionUUID);
  }

  public void testDeletePackage() throws BonitaException {
    //check IllegalArgumentException with null parameter
    //check undeployed packages are deleted and then we can deploy it again
    //check undeletablepackageException
    //check package, processes and instances are deleted
    //check an exception is thrown if the package doesn't exist

    try {
      getManagementAPI().deletePackage(null);
      fail("IllegalArgumentException must be thrown with null parameter");
    } catch (final IllegalArgumentException e) {
      // Test pass
    }

    URL xpdlUrl = this.getClass().getResource("deletePackage.xpdl");

    //check undeployed packages are deleted
    Map<String, ProcessDefinition> processes = getManagementAPI().deployBar(getBarContent(xpdlUrl));
    PackageDefinitionUUID packageUUID = processes.get("deletePackage1").getPackageDefinitionUUID();
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    try {
      getQueryDefinitionAPI().getPackage(packageUUID);
      fail("Package must not be found as it was deleted.");
    } catch (final PackageNotFoundException e) {
      //ok
    }
    //processes
    try {
      getQueryDefinitionAPI().getProcess(processes.get("deletePackage1").getUUID());
      fail("Process must not be found as it was deleted.");
    } catch (final ProcessNotFoundException e) {
      //ok
    }
    try {
      getQueryDefinitionAPI().getProcess(processes.get("deletePackage2").getUUID());
      fail("Process must not be found as it was deleted.");
    } catch (final ProcessNotFoundException e) {
      //ok
    }

    //check a deleted package can be deployed again
    processes = getManagementAPI().deployBar(getBarContent(xpdlUrl));
    packageUUID = processes.get("deletePackage1").getPackageDefinitionUUID();
    getManagementAPI().deletePackage(packageUUID);

    processes = getManagementAPI().deployBar(getBarContent(xpdlUrl));

    ProcessDefinition process1 = processes.get("deletePackage1");
    final ProcessDefinitionUUID process1UUID = process1.getUUID();

    final ProcessDefinition process2 = processes.get("deletePackage2");
    final ProcessDefinitionUUID process2UUID = process2.getUUID();

    packageUUID = process1.getPackageDefinitionUUID();

    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Set<ProcessInstance> instances = null;
    TaskUUID taskUUID = null;

    //create 2 instances. 1 will not finished (waiting on task), 2 will finished
    ProcessInstanceUUID instance1 = getRuntimeAPI().instantiateProcess(process1UUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    assertNotNull(todolist);
    assertFalse(todolist.isEmpty());
    assertTrue(todolist.size() == 1);

    final ProcessInstanceUUID instance2 = getRuntimeAPI().instantiateProcess(process2UUID);
    assertEquals(InstanceState.FINISHED, getQueryRuntimeAPI().getProcessInstance(instance2).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance2, ActivityState.READY);
    assertNotNull(todolist);
    assertTrue(todolist.isEmpty());

    //finish instance 1
    todolist = getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    taskUUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    assertEquals(InstanceState.FINISHED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());

    instances = getQueryRuntimeAPI().getProcessInstances(process1UUID);
    assertNotNull(instances);
    assertTrue(instances.size() == 1);

    instances = getQueryRuntimeAPI().getProcessInstances(process2UUID);
    assertNotNull(instances);
    assertTrue(instances.size() == 1);

    //check the package can be deleted
    getManagementAPI().deletePackage(packageUUID);

    //check everything was correctly deleted
    //package
    try {
      getQueryDefinitionAPI().getPackage(packageUUID);
      fail("Package must not be found as it was deleted.");
    } catch (final PackageNotFoundException e) {
      //ok
    }
    //processes
    try {
      getQueryDefinitionAPI().getProcess(process1UUID);
      fail("Process must not be found as it was deleted.");
    } catch (final ProcessNotFoundException e) {
      //ok
    }
    try {
      getQueryDefinitionAPI().getProcess(process2UUID);
      fail("Process must not be found as it was deleted.");
    } catch (final ProcessNotFoundException e) {
      //ok
    }
    //instances
    instances = getQueryRuntimeAPI().getProcessInstances(process1UUID);
    assertNotNull(instances);
    assertTrue(instances.isEmpty());
    instances = getQueryRuntimeAPI().getProcessInstances(process2UUID);
    assertNotNull(instances);
    assertTrue(instances.isEmpty());

    //check that if we try to delete an unknown package, packageNotFoundException is thrown
    try {
      getManagementAPI().deletePackage(packageUUID);
      fail("An exception must be thrown when trying to delete instances of an unexisting process");
    } catch (final PackageNotFoundException pnfe) {
      // ok
    }

    //Test cannot delete when instance running
    xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    final ProcessDefinition externProcess = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get("deleteInstance");

    xpdlUrl = this.getClass().getResource("deletePackageSubflow.xpdl");
    processes = getManagementAPI().deployBar(getBarContent(xpdlUrl));

    process1 = processes.get("deletePackage_subflow");
    packageUUID = process1.getPackageDefinitionUUID();
    instance1 = getRuntimeAPI().instantiateProcess(process1.getUUID());

    instances = getQueryRuntimeAPI().getProcessInstances(process1.getUUID());
    assertNotNull(instances);
    assertTrue(instances.size() == 1);
    assertTrue(instances.iterator().next().getInstanceState().equals(InstanceState.STARTED));

    final Set<ProcessInstanceUUID> childInstances = getQueryRuntimeAPI().getProcessInstance(instance1).getChildrenInstanceUUID();
    assertNotNull(childInstances);
    assertTrue(childInstances.size() == 1);
    final ProcessInstance childInstance = getQueryRuntimeAPI().getProcessInstance(childInstances.iterator().next());
    final PackageDefinitionUUID childPackageUUID = childInstance.getPackageDefinitionUUID();

    try {
      getManagementAPI().deletePackage(childPackageUUID);
      fail("UndeletableInstanceException must be thrown when trying to delete package with processes instances running");
    } catch (final UndeletableInstanceException e) {
      // ok
    }
    getRuntimeAPI().deleteAllProcessInstances(process1.getUUID());
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().deletePackage(externProcess.getPackageDefinitionUUID());
  }

  public void testDeployWitoutVersion() throws BonitaException {
    //check if version is not set, an error is thrown
    // version must be set either in package either in processes
    final URL xpdlUrl = this.getClass().getResource("noVersion.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("A package without version can't be deployed!");
    } catch (final DeploymentException e) {
      //the process which doesn't have a version
      assertTrue(e.getMessage().contains("noVersion"));
    }
  }

  public void testProcessVersionsOverridePackageVersion() throws BonitaException {
    // check if a process has no version defined, it take the package one
    // check if a process has a version defined, it overrides the package one

    final URL xpdlUrl = this.getClass().getResource("manyVersions.xpdl");

    //check undeployed packages are deleted
    final Map<String, ProcessDefinition> processes = getManagementAPI().deployBar(getBarContent(xpdlUrl));

    final ProcessDefinition process1 = processes.get("packageVersion");
    final ProcessDefinition process2 = processes.get("processVersion");
    final PackageDefinitionUUID packageUUID = process1.getPackageDefinitionUUID();
    final PackageDefinition package1 = getQueryDefinitionAPI().getPackage(packageUUID);

    assertEquals("packageVersion", process1.getVersion());
    assertEquals(package1.getVersion(), process1.getVersion());
    assertEquals("processVersion", process2.getVersion());

    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeployTwiceSamePackage() throws BonitaException {
    //check two packaes can't be deployed twice
    // 1) in the journal
    // 2) 1 in journal and one in history
    final URL xpdlUrl = this.getClass().getResource("deletePackage.xpdl");
    final PackageDefinitionUUID packageUUID =
      getManagementAPI().deployBar(getBarContent(xpdlUrl)).get("deletePackage1").getPackageDefinitionUUID();
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("The same package can't be deployed twice !");
    } catch (final DeploymentException e) {
      //ok
    }

    getManagementAPI().undeploy(packageUUID);

    //now, the package has moved to history, try to deploy one again in journal
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("The same package can't be deployed twice !");
    } catch (final DeploymentException e) {
      //ok
    }

    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeployCommandAtPackageLevel() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("deletePackage.xpdl");
    final Map<String, ProcessDefinition> processes =
      getManagementAPI().deployBar(getBarContent(xpdlUrl, GetProcessIdCommand.class));
    final PackageDefinitionUUID packageUUID = processes.entrySet().iterator().next().getValue().getPackageDefinitionUUID();
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeployTxHookNonRollback() throws BonitaException {
    //check the case of an impelementation... : TxHookImpl implements TxHook
    URL xpdlUrl = this.getClass().getResource("txHookNonRollback.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, TxHookImpl.class));
      fail("Impossible to deploy a a process declaring a non rollback hook and giving a TxHook class!");
    } catch (final DeploymentException e) {
      //ok
    }
    //check the case of a subclass... : MyTxHook1 implements TxHookSubItf extends TxHook
    xpdlUrl = this.getClass().getResource("subInterfaceTxHook.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, TxHookSubItf.class, MyTxHook1.class));
      fail("Impossible to deploy a a process declaring a non rollback hook and giving a TxHook class!");
    } catch (final DeploymentException e) {
      //ok
    }
    //check the case of a super class : MyTxHook2 extends TxHookImpl implements TxHook
    xpdlUrl = this.getClass().getResource("subClassTxHook.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, TxHookImpl.class, MyTxHook2.class));
      fail("Impossible to deploy a a process declaring a non rollback hook and giving a TxHook class!");
    } catch (final DeploymentException e) {
      //ok
    }
  }
  //TODO
  /*
  public void testSameProcessIdInDifferentPackages() throws BonitaException {
    //check a process id can be used by many packages (same or different versions)
    //assertTrue(false);

  }
  public void testSameProcessIdWithDifferentVersions() throws BonitaException {
    //check a process id can be deployed in many versions
    //assertTrue(false);
  }*/
  //--------------------------------------- private methods ---------------------------------------------------
  private PackageDefinitionUUID checkDeployment(final Map<String,
      ProcessDefinition> mapProcessDefinitions) throws BonitaException {
    assertTrue(mapProcessDefinitions.size() == 2);
    final Collection<ProcessDefinition> lp = mapProcessDefinitions.values();
    final Iterator<ProcessDefinition> it = lp.iterator();
    final ProcessDefinition procDef = it.next();
    final PackageDefinitionUUID packageDefinitionUUID = procDef.getPackageDefinitionUUID();
    assertEquals("testManagementAPI", getQueryDefinitionAPI().getPackage(packageDefinitionUUID).getName());

    // Change
    final Set<String> expectedProcessIds = new HashSet<String>();
    expectedProcessIds.add("process1");
    expectedProcessIds.add("process2");
    assertEquals(expectedProcessIds, mapProcessDefinitions.keySet());
    final ProcessDefinition process = getQueryDefinitionAPI().getLastProcess("testManagementAPI", "process1");
    final String processId = process.getProcessId();
    assertNotNull(processId);
    assertEquals("process1", processId);
    return packageDefinitionUUID;
  }

}
