package org.ow2.bonita.facade;

import java.net.URL;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import junit.framework.Assert;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.activity.instantiation.instantiator.TestOneJoinActivityInstantiator;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.UncancellableInstanceException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.SimpleCallbackHandler;

public class RuntimeAPITest extends APITestCase {


  /*
   *   test instantiateProcess(String processDefinitionUUID) from RuntimeAPI
   *
   *  Test:
   *   -instantiation + start execution
   *   -returned instance has STARTED state
   *  test :
   *     throws ProcessNotFoundException if the process has not been found
   *     (ex. : ""  random processDefinitionUUID) ,
   *     throws BonitaInternalException if an exception occurs (ex. if processDefinitionUUID == null)
   */
  public void testInstantiateProcess() throws BonitaException {
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      this.getRuntimeAPI().instantiateProcess(null);
      Assert.fail("Check null parameter");
    } catch (final IllegalArgumentException be) {
      // OK
    }
    try {
      this.getRuntimeAPI().instantiateProcess(IdFactory.getNewProcessUUID());
      Assert.fail("Check non existent process");
    } catch (final ProcessNotFoundException be) {
      //OK
    }
    try {
      final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
      // instantiateProcess calls start so the instance is started.
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());

      this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
      this.getManagementAPI().undeploy(packageUUID);
      this.getManagementAPI().deletePackage(packageUUID);
    } catch (final BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      Assert.fail("Exception thrown: " + Misc.getStackTraceFrom(be));
    }
  }

  /*
   *  Test instantiateProcess(String processDefinitionUUID, Map<String Object>variables) from RuntimeAPI
   *    check the map of variables has been added to the created instance.
   */
  public void testInstantiateProcessWithMapVariables() throws BonitaException {
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      final Map<String, Object> expectedVarMap = new HashMap<String, Object>();
      expectedVarMap.put("str1", "toto");
      final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID, expectedVarMap);
      // instantiateProcess calls start so the instance is started.
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());

      final Map<String, Object> currentVarMap = this.getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
      //check the instance get the expected variable map.
      Assert.assertEquals(expectedVarMap, currentVarMap);

      this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
      this.getManagementAPI().undeploy(packageUUID);
      this.getManagementAPI().deletePackage(packageUUID);
    } catch (final BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      Assert.fail("Exception thrown: " + Misc.getStackTraceFrom(be));
    }
  }

  /*
   *  test deleteInstance & deleteInstances
   */
  public void testDeleteInstance() throws BonitaException {
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      //test deleteInstance()
      final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      Collection<ActivityInstance< TaskInstance >> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
      Assert.assertTrue("Check task is not empty", tasks.size() == 1);
      this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
      try {
        //getInstance call the journal !!
        this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
        Assert.fail("Check fail getting instance into the journal with getInstance()");
      } catch (final InstanceNotFoundException ie) {
        Assert.assertTrue("Check exception has well been raised !", ie.getInstanceUUID().equals(instanceUUID));
      }

      //test deleteInstances()
      final ProcessInstanceUUID instanceUUID1 = this.getRuntimeAPI().instantiateProcess(processUUID);
      final ProcessInstanceUUID instanceUUID2 = this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID1).getInstanceState());
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID2).getInstanceState());

      tasks = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
      Assert.assertEquals("Check task is not empty", 2, tasks.size());

      this.getRuntimeAPI().deleteAllProcessInstances(processUUID);

      final Set<ProcessInstance> processInstances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
      Assert.assertTrue("No more process instance into the journal", processInstances.size() == 0);
      this.getManagementAPI().undeploy(packageUUID);
      this.getManagementAPI().deletePackage(packageUUID);
    } catch (final BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      Assert.fail("Exception thrown:" + Misc.getStackTraceFrom(be));
    }
  }

  /*
   *  Test startTask(String taskUUID) from the runtimeAPI.
   *
   * test initial state need to be READY.
   * change state to EXECUTING.
   * tests following exceptions are raised:
   *  @throws TaskNotFoundException if the task has not been found.
   *  @throws IllegalTaskStateException if the state of the task has not READY state.
   *  @throws BonitaInternalException if an exception occurs (ex. : if taskUUID == null)
   */
  public void testStartTask() throws BonitaException {
    ProcessInstanceUUID instanceUUID = null;
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    } catch (final BonitaException e) {
      Assert.fail("Ouch! How can it be?! Should never reach this statement!, exception: " + Misc.getStackTraceFrom(e));
    }
    try {
      this.getRuntimeAPI().startTask(null, true);
      Assert.fail("Check null argument");
    } catch (final IllegalArgumentException be) {
      // ok
    }
    try {
      this.getRuntimeAPI().startTask(IdFactory.getNewTaskUUID(), true);
      Assert.fail("Check non existent task");
    } catch (final TaskNotFoundException be) {
      //OK
    }
    this.checkStopped(instanceUUID, new String[]{});
    final Collection<ActivityInstance< TaskInstance >> activitiesToDo =
      this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, activitiesToDo.size());
    final ActivityInstance< TaskInstance > taskActivity = activitiesToDo.iterator().next();
    Assert.assertEquals(ActivityState.READY, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    try {
      this.getRuntimeAPI().startTask(taskActivity.getBody().getUUID(), true);
    } catch (final BonitaException e) {
      Assert.fail("Ouch! How can it be?! Should never reach this statement!, exception: " + Misc.getStackTraceFrom(e));
    }
    Assert.assertEquals(ActivityState.EXECUTING, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test terminateTask(String taskUUID) from the runtimeAPI.
   *
   * Test initial state required is : EXECUTING.
   * Test if finish is successful state become FINISHED.
   * Test
   *    - @throws TaskNotFoundException if the task has not been found.
   *    - @throws IllegalTaskStateException if the state of the task has not EXECUTING state.
   *    - @throws BonitaInternalException if an exception occurs (for ex. with illegal taskUUID parameter).
   *
   */
  public void testFinishTask() throws BonitaException {
    ProcessInstanceUUID instanceUUID = null;
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    } catch (final BonitaException e) {
      Assert.fail("Ouch! How can it be?! Should never reach this statement!");
    }
    try {
      this.getRuntimeAPI().finishTask(null, true);
      Assert.fail("Check null argument");
    } catch (final IllegalArgumentException be) {
      // ok
    }
    try {
      this.getRuntimeAPI().finishTask(IdFactory.getNewTaskUUID(), true);
      Assert.fail("Check non existent instance");
    } catch (final TaskNotFoundException be) {
      //OK
    }
    this.checkStopped(instanceUUID, new String[]{});
    Collection<ActivityInstance< TaskInstance >> activitiesToDo =
      this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, activitiesToDo.size());
    ActivityInstance< TaskInstance > taskActivity = activitiesToDo.iterator().next();
    try {
      this.getRuntimeAPI().startTask(taskActivity.getBody().getUUID(), true);
    } catch (final BonitaException e) {
      Assert.fail("Ouch! How can it be?! Should never reach this statement!, exception: " + Misc.getStackTraceFrom(e));
    }
    activitiesToDo = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.EXECUTING);
    Assert.assertEquals(1, activitiesToDo.size());
    taskActivity = activitiesToDo.iterator().next();
    Assert.assertEquals(ActivityState.EXECUTING, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    //Check IllegalTaskStateException (suspend the task and try to finish)
    try {
      this.getRuntimeAPI().suspendTask(taskActivity.getBody().getUUID(), true);
      this.getRuntimeAPI().finishTask(taskActivity.getBody().getUUID(), true);
    } catch (final IllegalTaskStateException be) {
      //OK
    }

    try {
      this.getRuntimeAPI().resumeTask(taskActivity.getBody().getUUID(), false);
      final Collection<ActivityInstance<TaskInstance>> resumedTasks =
        this.getQueryRuntimeAPI().getTaskList("admin", ActivityState.EXECUTING);
      Assert.assertNotNull(resumedTasks);
      Assert.assertEquals(1, resumedTasks.size());
      this.getRuntimeAPI().finishTask(taskActivity.getBody().getUUID(), true);
    } catch (final BonitaException e) {
      Assert.fail("Ouch! How can it be?! Should never reach this statement!, exception: " + Misc.getStackTraceFrom(e));
    }

    Assert.assertEquals(ActivityState.FINISHED, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());
    final Collection<ActivityInstance<TaskInstance>> finishedTasks = this.getQueryRuntimeAPI().getTaskList("admin", ActivityState.FINISHED);
    Assert.assertNotNull(finishedTasks);
    Assert.assertEquals(1, finishedTasks.size());
    final ActivityInstance<TaskInstance> finishedTask = finishedTasks.iterator().next();
    Assert.assertEquals(ActivityState.FINISHED, finishedTask.getBody().getState());
    Assert.assertEquals(taskActivity.getActivityId(), finishedTask.getActivityId());

    // Try to finish a finished task
    try {
      this.getRuntimeAPI().finishTask(finishedTask.getBody().getUUID(), true);
      Assert.fail("expecting exception");
    } catch (final IllegalTaskStateException e) {
      // ok
    }
    // Try to start a finished task
    try {
      this.getRuntimeAPI().startTask(finishedTask.getBody().getUUID(), true);
      Assert.fail("expecting exception");
    } catch (final IllegalTaskStateException e) {
      // ok
    }

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   *  test exceptions and the documented case from javadoc
   *  (if the instance or activity var does not exist it is created)
   *  for methods (of RuntimeAPI) :
   *     - setInstanceVariable
   *     - setActivityVariable
   *
   *     TODO (not tested):
   *       - test the AmbiguousActivityException raised if multiple iterations exist !!!!!
   *       - test ActivityNotFoundException if setting activity variables when activty
   *          already finished
   *       - test InstanceNotFoundException if setting instance variables when instance
   *          already finished
   *      Note:   need probably to introduce a new exception for that !!!!!
   */
  public void testSetVariables() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    final ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    final ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    final String activity1 = "act1";
    Collection<ActivityInstance<ActivityBody>> acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    Assert.assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    Assert.assertEquals("act1", activityInst.getActivityId());

    //test InstanceNotFoundException
    try {
      this.getRuntimeAPI().setProcessInstanceVariable(badInstanceUUID, "string_process", "titi");
      Assert.fail("Check non existent Instance");
    } catch (final InstanceNotFoundException be) {
      //OK
      Assert.assertEquals(badInstanceUUID, be.getInstanceUUID());
    }

    //test setActivityVariable with activityUUID
    try {
      this.getRuntimeAPI().setActivityInstanceVariable(badActivityUUID, "string_activity", "tutu");
      Assert.fail("Check non existent Activity Instance");
    } catch (final ActivityNotFoundException be) {
      //OK
      Assert.assertEquals(badActivityUUID, be.getActivityUUID());
    }

    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});

    //get info with "act2" activity
    final String activity2 = "act2";
    acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity2);
    Assert.assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    final ActivityInstanceUUID activityUUID = activityInst.getUUID();
    Assert.assertEquals(activity2, activityInst.getActivityId());


    //**** testing setVariable() ****
    //with valid parameter to set process instance variable
    this.getRuntimeAPI().setVariable(activityUUID, "string_process", "new_titi");
    Assert.assertEquals("new_titi", this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_process"));

    //with valid parameter to set activity instance variable
    this.getRuntimeAPI().setVariable(activityUUID, "string_activity", "new_titi");
    Assert.assertEquals("new_titi",this.getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_activity"));


    //with invalid parameters to test exceptions:
    //InstanceNotFoundException
    //TODO
    try {
      this.getRuntimeAPI().setVariable(badActivityUUID, "string_process", "titi");
      Assert.fail("Check non existent Instance");
    } catch (final ActivityNotFoundException be) {
      //OK
      Assert.assertEquals(badActivityUUID, be.getActivityUUID());
    }

    //VariableNotFoundException
    try {
      this.getRuntimeAPI().setVariable(activityUUID, "string_process_bidon", "titi");
      Assert.fail("Check non existent Instance");
    } catch (final VariableNotFoundException be) {
      //OK
      Assert.assertEquals("string_process_bidon", be.getVariableId());
    }

    // start & terminate activity2 task
    this.executeTask(instanceUUID, activity2);

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }



  /*
   * Testing getActivityvariables() without cycle/iteration
   * There's 2 methods for getActivityVariables()
   * 1) Map<String, Object> getActivityVariables(ProcessInstanceUUID instanceUUID,String activityId)
   *  throws: ActivityNotFoundException,
   *           AmbiguousActivityException,
   *           InstanceNotFoundException;
   *  2) Map<String, Object> getActivityVariables(ActivityInstanceUUID activityUUID)
   *   throws ActivityNotFoundException;
   *
   *  Check also if trying with wrong iterationId an ActivityNotFoundException is raised
   *  (specific exception with the iteration id as parameter).
   */
  public void testGetActivityVariables() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";
    final String activity1 = "act1";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    //ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    final ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    Collection<ActivityInstance<ActivityBody>> acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    Assert.assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();

    //in act1 there's only process variables so the method returns an empty list
    Map<String, Object> variables = this.getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    Assert.assertTrue(variables.isEmpty());

    //Check raised exceptions with getActivityVariables(ActivityInstanceUUID activityUUID)
    // 1- ActivityNotFoundException
    try {

      this.getQueryRuntimeAPI().getActivityInstanceVariables(badActivityUUID);
      Assert.fail("Check non existent activity");
    } catch (final ActivityNotFoundException be) {
      //OK
    }

    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});

    //execution is poniting to act2 which contains process and activity variables
    acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act2");
    Assert.assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();

    //when using parameters: activityUUID
    variables = this.getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_activity"));
    Assert.assertTrue(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    this.executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node


    acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act3");
    Assert.assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();
    variables = this.getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    Assert.assertTrue(variables.isEmpty());

    this.checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetProcessVariables() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";
    final String activity1 = "act1";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    final ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //in act1 there's only process variables
    Map<String, Object> variables = this.getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));

    //Check RAISED EXCEPTIONS
    // InstanceNotFoundException
    try {
      this.getQueryRuntimeAPI().getProcessInstanceVariables(badInstanceUUID);
      Assert.fail("Check non existent instance");
    } catch (final InstanceNotFoundException be) {
      //OK
    }

    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});

    // Check variables within execution pointing to act2 node
    variables = this.getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));
    Assert.assertFalse(variables.containsKey("string_activity"));
    Assert.assertFalse(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    this.executeTask(instanceUUID, "act2");

    // check variables within execution pointing to act3 node
    variables = this.getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));
    Assert.assertFalse(variables.containsKey("string_activity"));
    Assert.assertFalse(variables.containsKey("enum_activity"));

    this.checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetVariables() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";
    final String activity1 = "act1";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    final ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    Collection<ActivityInstance<ActivityBody>> acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    Assert.assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();

    //in act1 there's only process variables
    Map<String, Object> variables = this.getQueryRuntimeAPI().getVariables(activityUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));

    //Check RAISED EXCEPTIONS
    // ActivityNotFoundException
    try {
      this.getQueryRuntimeAPI().getVariables(badActivityUUID);
      Assert.fail("Check non existent activity");
    } catch (final ActivityNotFoundException be) {
      //OK
    }


    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});

    acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act2");
    Assert.assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();

    variables = this.getQueryRuntimeAPI().getVariables(activityUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(4, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));
    Assert.assertTrue(variables.containsKey("string_activity"));
    Assert.assertTrue(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    this.executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node
    acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act3");
    Assert.assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();

    variables = this.getQueryRuntimeAPI().getVariables(activityUUID);
    Assert.assertFalse(variables.isEmpty());
    Assert.assertEquals(2, variables.size());
    Assert.assertTrue(variables.containsKey("string_process"));
    Assert.assertTrue(variables.containsKey("enum_process"));
    Assert.assertFalse(variables.containsKey("string_activity"));
    Assert.assertFalse(variables.containsKey("enum_activity"));

    this.checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetActivityInstance() throws BonitaException {
    final String processId = "usertest2";
    final URL xpdlUrl = this.getClass().getResource("usertest2_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      //generating bad UUID to test exceptions
      final ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

      final Collection<ActivityInstance<ActivityBody>> acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a");
      Assert.assertEquals(1, acts.size());
      final ActivityInstanceUUID activityUUID = acts.iterator().next().getUUID();

      ActivityInstance<ActivityBody> activityInst = this.getQueryRuntimeAPI().getActivityInstance(activityUUID);
      Assert.assertNotNull(activityInst);
      Assert.assertEquals("a", activityInst.getActivityId());
      //check ActivityNotFoundException
      try {
        activityInst = this.getQueryRuntimeAPI().getActivityInstance(badActivityUUID);
      } catch (final ActivityNotFoundException ae) {
        Assert.assertEquals(badActivityUUID, ae.getActivityUUID());
      }
    } catch (final ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }
    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Testing getActivityVariable() without cycle/iteration
   *
   *  getActivityVariable(ActivityInstanceUUID activityUUID,String variableId)
   *     throws : ActivityNotFoundException,
   *              VariableNotFoundException;
   */
  public void testGetActivityVariable() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";
    final String activity1 = "act1";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    //ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    final ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    final Collection<ActivityInstance<ActivityBody>> acts = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    Assert.assertEquals(1, acts.size());
    final ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    Assert.assertNotNull(activityInst);
    final ActivityInstanceUUID activityUUID = activityInst.getUUID();

    //Check getActivityVariable(ActivityInstanceUUID activityUUID,String variableId) method
    //process var
    try {
      this.getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_process");
      Assert.fail("string_process is not local to act1 !");
    } catch (final VariableNotFoundException ve) {
      //OK
      Assert.assertEquals("string_process", ve.getVariableId());
    }

    //activity var does not exist
    //VariableNotFoundException
    try {
      this.getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_activity");
      Assert.fail("Check the activity var does not exist in act1");
    } catch (final VariableNotFoundException ve) {
      //OK
      Assert.assertEquals("string_activity", ve.getVariableId());
    }
    //ActivityNotFoundException
    try {
      this.getQueryRuntimeAPI().getActivityInstanceVariable(badActivityUUID, "string_process");
      Assert.fail("Check non existent badActivityUUID");
    } catch (final ActivityNotFoundException ae) {
      //OK
      Assert.assertEquals(badActivityUUID, ae.getActivityUUID());
    }

    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});
    // activityInst is unused
    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act2");

    // start & terminate "act2" task
    this.executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node

    //    activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act3");
    // activityInst is unused
    this.checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetProcessVariable() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    final String processId = "variables";
    final String activity1 = "act1";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    final ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //Check getActivityVariable(ActivityInstanceUUID activityUUID,String variableId) method
    //process var
    String varStringProcess = (String)this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_process");
    Assert.assertTrue(varStringProcess.equals("initial string value"));

    //No activity var exist
    //VariableNotFoundException
    try {
      this.getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_activity");
      Assert.fail("Check the activity var does not exist in act1");
    } catch (final VariableNotFoundException ve) {
      //OK
      Assert.assertEquals("string_activity", ve.getVariableId());
    }
    //InstanceNotFoundException
    try {
      varStringProcess = (String)this.getQueryRuntimeAPI().getProcessInstanceVariable(badInstanceUUID, "string_process");
      Assert.fail("Check non existent badActivityUUID");
    } catch (final InstanceNotFoundException ae) {
      //OK
      Assert.assertEquals(badInstanceUUID, ae.getInstanceUUID());
    }

    // start & terminate "act1" task
    this.executeTask(instanceUUID, activity1);
    this.checkStopped(instanceUUID, new String[]{activity1});
    // activityInst is unused
    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act2");

    // start & terminate "act2" task
    this.executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node

    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act3");
    // activityInst is unused
    this.checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    this.executeTask(instanceUUID, "act3");
    this.checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test Collection<TaskInstance> getTaskList(TaskState taskState);
   * from RuntimeReadOnlyAPI
   */

  //Takes into account
  //    . the states (4 states available)
  //    . the 3 types of methods depending on the parameters:
  //                . TaskState taskState,
  //                . ProcessInstanceUUID instanceUUID,
  //                . java.lang.String userId
  // Note for the state:
  // if state == null  return IllegalArgumentException.
  // case 1: state = null and state = READY
  // - if task assigned to a user check this (authenticated) user can get the task.
  // - check that an other users of the candidates list can't get the task.
  // - if the task is not assigned (role mapper without performer assign)
  // - check that an other user in the candidate list can get the task.
  // - if 2 instances are created check that 2 tasks are available.
  // (check the tasks has READY state)
  // case 2: state = SUSPENDED
  // check task has been assigned to the user performing suspendTask
  // case 3: state = EXECUTING
  // check task has been assigned to the user performing startTask
  // case 4: state = FINISHED
  public void testGetTaskList() throws BonitaException, LoginException {

    final String processId = "testGetTask";
    final String loginAdmin = "admin";
    final String passwordAdmin = "bpm";
    final String loginJohn = "john";
    final String passwordJohn = "bpm";
    final URL xpdlUrl = this.getClass().getResource("testGetTask.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl, AdminsRoleMapper.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    TaskUUID taskUUID = null;
    ProcessInstanceUUID instanceUUID = null;
    final ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //**** case 1 test ***
    try {
      instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      //getToDoList with taskState = null
      try {
        this.getQueryRuntimeAPI().getTaskList(null);
        Assert.fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (final IllegalArgumentException e) {
        //OK
      }
      try {
        this.getQueryRuntimeAPI().getTaskList(instanceUUID, null);
        Assert.fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (final IllegalArgumentException e) {
        //OK
      }
      try {
        this.getQueryRuntimeAPI().getTaskList(instanceUUID, loginAdmin, null);
        Assert.fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (final IllegalArgumentException e) {
        //OK
      }

      //getToDoList to get task for activity = "a" with taskState = READY
      final Collection<ActivityInstance< TaskInstance >> taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
      Assert.assertEquals(1, taskActivities.size());

      final Collection<ActivityInstance< TaskInstance >> taskActivitiesWithInst =
        this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
      Assert.assertEquals(1, taskActivitiesWithInst.size());

      //InstanceNotFoundException for getTaskList with argt: instanceUUID, taskState
      try {
        this.getQueryRuntimeAPI().getTaskList(badInstanceUUID, ActivityState.READY);
        Assert.fail("check InstanceNotFoundException has been raised");
      } catch (final InstanceNotFoundException ie) {
        Assert.assertEquals(ie.getInstanceUUID(), badInstanceUUID);
      }

      //InstanceNotFoundException for getTaskList with argt: instanceUUID, userId, taskState
      try {
        this.getQueryRuntimeAPI().getTaskList(badInstanceUUID, loginAdmin, ActivityState.READY);
        Assert.fail("check InstanceNotFoundException has been raised");
      } catch (final InstanceNotFoundException ie) {
        Assert.assertEquals(ie.getInstanceUUID(), badInstanceUUID);
      }

      final ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();
      Assert.assertNotNull(taskActivity.getBody().getUUID());
      taskUUID = taskActivity.getBody().getUUID();
      Assert.assertEquals("a", taskActivity.getActivityId());
      Assert.assertNotNull(taskActivity.getBody().getCreatedDate());
    } catch (final ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }

    //check the user john can't get the task
    this.loginAs(loginJohn, passwordJohn);
    Collection<ActivityInstance< TaskInstance >> taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(0, taskActivities.size());

    // go on finish task a (create task b)
    // loginAdmin is the user assigned to the task : a
    this.loginAs(loginAdmin, passwordAdmin);

    //case 4: test for state = FINISHED
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertEquals(0, taskActivities.size());

    //start&finish task : a
    this.getRuntimeAPI().startTask(taskUUID, true);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.EXECUTING);
    Assert.assertEquals(1, taskActivities.size());
    this.getRuntimeAPI().finishTask(taskUUID, true);

    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertEquals(1, taskActivities.size());
    Assert.assertEquals("a", taskActivities.iterator().next().getActivityId());

    //check admin user can get the task b
    this.loginAs(loginAdmin, passwordAdmin);
    //Tasks for authenticated user
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(1, taskActivities.size());
    //Tasks for a choosen user (admin)
    final Collection<ActivityInstance< TaskInstance >> taskActivitiesAdminUser =
      this.getQueryRuntimeAPI().getTaskList(loginAdmin,ActivityState.READY);
    final Collection<ActivityInstance< TaskInstance >> taskActivitiesAdminUserWithInstUUID =
      this.getQueryRuntimeAPI().getTaskList(instanceUUID, loginAdmin,ActivityState.READY);

    Assert.assertEquals("Check loginAdmin can get its tasks",1, taskActivities.size());
    Assert.assertEquals("Check tasks can be got for admin user",1, taskActivitiesAdminUserWithInstUUID.size());

    final ActivityInstance< TaskInstance > taskActivityAdminUser = taskActivitiesAdminUser.iterator().next();
    final ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();

    Assert.assertEquals(taskActivityAdminUser.getActivityId() , taskActivity.getActivityId());

    //other checks on got taskInstance
    Assert.assertNotNull(taskActivity.getActivityId());
    taskUUID = taskActivity.getBody().getUUID();

    //check task b is not assigned and candidates list contains : admin, john
    Assert.assertEquals(BonitaConstants.SYSTEM_USER, taskActivity.getBody().getUpdatedBy());
    Assert.assertFalse(taskActivity.getBody().isTaskAssigned());
    try {
      Assert.fail(taskActivity.getBody().getTaskUser());
    } catch (final IllegalStateException ie) {
      //OK
    }


    Assert.assertTrue(taskActivity.getBody().getTaskCandidates().containsAll(Arrays.asList(loginAdmin, loginJohn)));
    Assert.assertEquals("b", taskActivity.getActivityId());

    //check john user can get also the task b
    this.loginAs(loginJohn, passwordJohn);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(1, taskActivities.size());

    //Check that if an other instance of the same process is created 2 tasks are available for admin user.
    this.loginAs(loginAdmin, passwordAdmin);
    instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
    //getToDoList to get task for activity = "a"
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(2, taskActivities.size());

    //delete the 2 nd created instance
    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);

    //*** case 2 test ****
    //**suspend the task b withou assignment !  (only candidates are set)
    this.loginAs(loginAdmin, passwordAdmin);
    this.getRuntimeAPI().suspendTask(taskUUID, false);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.SUSPENDED);
    Assert.assertEquals(1, taskActivities.size());

    //check john user can't get the suspended task (by admin user)
    this.loginAs(loginJohn, passwordJohn);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.SUSPENDED);
    Assert.assertEquals(1, taskActivities.size());

    //check admin user can get the suspended task (by admin user)
    this.loginAs(loginAdmin, passwordAdmin);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.SUSPENDED);
    Assert.assertEquals(1, taskActivities.size());

    //**resumeTask and check that both admin and john can get the task b
    //meaning that the assigned user has been set as it was before the suspend (ie null)
    this.loginAs(loginJohn, passwordJohn);
    this.getRuntimeAPI().resumeTask(taskUUID, false);

    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(1, taskActivities.size());

    this.loginAs(loginAdmin, passwordAdmin);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertEquals(1, taskActivities.size());


    //*** case 3 test ***
    //check that only the user having done the startTask can get the task.
    this.loginAs(loginAdmin, passwordAdmin);
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.loginAs(loginAdmin, passwordAdmin);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.EXECUTING);
    Assert.assertEquals(1, taskActivities.size());
    this.loginAs(loginJohn, passwordJohn);
    taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.EXECUTING);
    Assert.assertEquals(0, taskActivities.size());

    //go on until undeploying
    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testDoneList() throws BonitaException {
    final String processId = "doneList";
    final URL xpdlUrl = this.getClass().getResource("doneList.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        AdminsRoleMapper.class)).get(processId);
    final ProcessDefinitionUUID processUUID = process.getUUID();

    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Collection<ActivityInstance<TaskInstance>> donelist = null;

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    final String otherUser = "anotherUser";

    //all tasks are assign to instanceInitiator by default
    donelist = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(0, donelist.size());

    todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(1, todolist.size());

    //execute the first task and check it appears in the done list
    final TaskUUID task1UUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().startTask(task1UUID, true);
    this.getRuntimeAPI().finishTask(task1UUID, true);

    donelist = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(1, donelist.size());
    Assert.assertEquals(task1UUID, donelist.iterator().next().getBody().getUUID());

    //assign the second task to another user
    todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(1, todolist.size());
    final TaskUUID task2UUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().assignTask(task2UUID, otherUser);

    //check the other user has task2 in its todolist
    todolist = this.getQueryRuntimeAPI().getTaskList(otherUser, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(1, todolist.size());
    Assert.assertEquals(task2UUID, todolist.iterator().next().getBody().getUUID());

    //check the task is not anymore in todoList of the logged user
    todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(0, todolist.size());

    //execute the second task as 'another user"
    this.getRuntimeAPI().startTask(task2UUID, false);
    this.getRuntimeAPI().finishTask(task2UUID, false);

    //check the second task is  in the donelist of the other user
    donelist = this.getQueryRuntimeAPI().getTaskList(otherUser, ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(1, donelist.size());
    Assert.assertEquals(task2UUID, donelist.iterator().next().getBody().getUUID());

    //check the second task is not in the donelist of the current logged user. Only t1 is in donelist
    donelist = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(1, donelist.size());
    Assert.assertEquals(task1UUID, donelist.iterator().next().getBody().getUUID());

    //execute the third task and check it appears in the done list of the logged user
    todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(1, todolist.size());

    final TaskUUID task3UUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().startTask(task3UUID, true);
    this.getRuntimeAPI().finishTask(task3UUID, true);

    donelist = this.getQueryRuntimeAPI().getTaskList(ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(2, donelist.size());

    //check doneList of the logged user contains task1 and task3
    final Set<TaskUUID> expected = this.buildSet(task1UUID, task3UUID);
    final Set<TaskUUID> donelistUUIDs = new HashSet<TaskUUID>();
    final Iterator<ActivityInstance<TaskInstance>> it = donelist.iterator();
    while (it.hasNext()) {
      donelistUUIDs.add(it.next().getBody().getUUID());
    }
    Assert.assertEquals(expected, donelistUUIDs);

    //check doneList of the other user contains task2
    donelist = this.getQueryRuntimeAPI().getTaskList(otherUser, ActivityState.FINISHED);
    Assert.assertNotNull(donelist);
    Assert.assertEquals(1, donelist.size());
    Assert.assertEquals(task2UUID, donelist.iterator().next().getBody().getUUID());

    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  private Set<TaskUUID> buildSet(final TaskUUID... taskUUIDs) {
    final Set<TaskUUID> s = new HashSet<TaskUUID>();
    for (final TaskUUID taskUUID : taskUUIDs) {
      s.add(taskUUID);
    }
    return s;
  }
  public void testGetTask() throws BonitaException {
    final String processId = "usertest2";
    final URL xpdlUrl = this.getClass().getResource("usertest2_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final TaskUUID newTaskUUID = IdFactory.getNewTaskUUID();
    ProcessInstanceUUID instanceUUID = null;
    try {
      instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
      Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      final Collection<ActivityInstance< TaskInstance >> taskActivities = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
      Assert.assertEquals(1, taskActivities.size());
      final TaskUUID taskUUID = taskActivities.iterator().next().getBody().getUUID();

      //getTask for task within activity = a
      final ActivityInstance< TaskInstance > taskActivity = this.getQueryRuntimeAPI().getTask(taskUUID);
      Assert.assertNotNull(taskActivity.getActivityId());
      Assert.assertEquals("a", taskActivity.getActivityId());
      Assert.assertNotNull(taskActivity.getBody().getCreatedDate());
    } catch (final ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }
    try {
      this.getQueryRuntimeAPI().getTask(newTaskUUID);
      Assert.fail("This task does not exist");
    } catch (final TaskNotFoundException e) {
      Assert.assertEquals(newTaskUUID, e.getTaskUUID());
    }

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeleteAllProcessInstances() throws BonitaException {

    try {
      this.getRuntimeAPI().deleteAllProcessInstances(null);
    } catch (final IllegalArgumentException e) {
      // Test pass
    }

    final String processId = "deleteInstance";
    final URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Set<ProcessInstance> instances = null;
    TaskUUID taskUUID = null;

    //check that both running and finished instances are deleted
    ProcessInstanceUUID instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = this.getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertFalse(todolist.isEmpty());
    Assert.assertTrue(todolist.size() == 1);

    ProcessInstanceUUID instance2 = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instance2).getInstanceState());
    todolist = this.getQueryRuntimeAPI().getTaskList(instance2, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertFalse(todolist.isEmpty());
    Assert.assertTrue(todolist.size() == 1);
    taskUUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);
    Assert.assertEquals(InstanceState.FINISHED, this.getQueryRuntimeAPI().getProcessInstance(instance2).getInstanceState());

    instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertNotNull(instances);
    Assert.assertTrue(instances.size() == 2);

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);

    instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertNotNull(instances);
    Assert.assertTrue(instances.isEmpty());

    //check that if we try to delete instances of an unknown process, processNotFoundException is thrown
    final ProcessDefinitionUUID wrongProcessUUID = IdFactory.getNewProcessUUID();
    try {
      this.getRuntimeAPI().deleteAllProcessInstances(wrongProcessUUID);
      Assert.fail("An exception must be thrown when trying to delete instances of an unexisting process");
    } catch (final ProcessNotFoundException pnfe) {
      // ok
    }
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);

    // Check UndeletableInstanceException is thrown when trying to delete process which have parentInstance
    process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get("subflow");
    packageUUID = process.getPackageDefinitionUUID();
    processUUID = process.getUUID();

    instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(instance1);
    final Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    Assert.assertNotNull(instancesUUID);
    Assert.assertTrue(instancesUUID.size() == 1);
    final ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next();
    instance2 = this.getRuntimeAPI().instantiateProcess(processUUID);

    try {
      this.getRuntimeAPI().deleteAllProcessInstances(
        this.getQueryRuntimeAPI().getProcessInstance(childInstanceUUID).getProcessDefinitionUUID());
      Assert.fail("Cannot delete a process instance having a parent instance");
    } catch (final UndeletableInstanceException e) {
      // Ok
      instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
      Assert.assertNotNull(instances);
      Assert.assertTrue(instances.size() == 2);
    }

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeleteProcessInstance() throws BonitaException {
    // many cases to test :
    // 0) check IllegalArgumentException with null parameter
    // 1) delete a running instance : check instance not available in querier
    // 2) delete an archived instance : check available in querier, delete, check not available
    // 3) check instanceNotFound thrown if instance can't be found in repository neither in journal, neither in history
    // 4) check UndeletableInstanceException

    // case 0
    try {
      this.getRuntimeAPI().deleteProcessInstance(null);
      Assert.fail("IllegalArgumentException must be thrown");
    } catch (final IllegalArgumentException e) {
      // ok
    }

    final String processId = "deleteInstance";
    final URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instance1 = null;
    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Set<ProcessInstance> instances = null;
    TaskUUID taskUUID = null;

    // case 1
    instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = this.getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertFalse(todolist.isEmpty());
    Assert.assertTrue(todolist.size() == 1);
    this.getRuntimeAPI().deleteProcessInstance(instance1);
    try {
      this.getQueryRuntimeAPI().getProcessInstance(instance1);
      Assert.fail("Instance is just deleted! We must a an exception here!");
    } catch (final InstanceNotFoundException infe) {
      // ok
    }
    instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertNotNull(instances);
    Assert.assertTrue(instances.isEmpty());

    // case 2
    instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = this.getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertFalse(todolist.isEmpty());
    Assert.assertTrue(todolist.size() == 1);
    taskUUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);
    Assert.assertEquals(InstanceState.FINISHED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instance1);
    try {
      this.getQueryRuntimeAPI().getProcessInstance(instance1);
      Assert.fail("Instance is just deleted! We must a an exception here!");
    } catch (final InstanceNotFoundException infe) {
      // ok
    }
    instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertNotNull(instances);
    Assert.assertTrue(instances.isEmpty());

    // case 3
    try {
      this.getRuntimeAPI().deleteProcessInstance(instance1);
      Assert.fail("Unable to delete an instance which can't be found in journal neither in repository");
    } catch (final InstanceNotFoundException infe) {
      // ok
    }

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);

    // case 4 : Check UndeletableInstanceException is thrown when trying to delete process which have parentInstance
    process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get("subflow");
    packageUUID = process.getPackageDefinitionUUID();
    processUUID = process.getUUID();

    instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(instance1);
    final Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    Assert.assertNotNull(instancesUUID);
    Assert.assertTrue(instancesUUID.size() == 1);
    final ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next();
    try {
      this.getRuntimeAPI().deleteProcessInstance(childInstanceUUID);
      Assert.fail("Cannot delete a process instance having a parent instance");
    } catch (final UndeletableInstanceException e) {
      // Ok
      instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
      Assert.assertNotNull(instances);
      Assert.assertTrue(instances.size() == 1);

      todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
      Assert.assertNotNull(todolist);
      Assert.assertTrue(todolist.size() == 1);
      taskUUID = todolist.iterator().next().getBody().getUUID();

      this.getRuntimeAPI().startTask(taskUUID, true);
      this.getRuntimeAPI().finishTask(taskUUID, true);
    }

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   *  Test cancelProcessInstance(ProcessInstanceUUID uuid) from the runtimeAPI.
   * Cancels a process instance which has a ready task. Check the ask is canceled.
   */
  public void testCancelProcessInstance() throws BonitaException {
    ProcessInstanceUUID instanceUUID = null;
    final String processId = "usertest1";
    final URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance< TaskInstance >> activitiesToDo =
      this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, activitiesToDo.size());
    final ActivityInstance< TaskInstance > taskActivity = activitiesToDo.iterator().next();
    Assert.assertEquals(ActivityState.READY, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    this.getRuntimeAPI().cancelProcessInstance(instanceUUID);

    activitiesToDo = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(0, activitiesToDo.size());

    Assert.assertEquals(ActivityState.CANCELLED, this.getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    Assert.assertEquals(InstanceState.CANCELLED, processInstance.getInstanceState());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }



  public void testCancelProcessInstanceBadArgs() throws BonitaException {
    try {
      this.getRuntimeAPI().cancelProcessInstance(new ProcessInstanceUUID(UUID.randomUUID().toString()));
      Assert.fail(InstanceNotFoundException.class.getSimpleName() + " expected");
    } catch (final InstanceNotFoundException e) {
      // ok
    }

    try {
      this.getRuntimeAPI().cancelProcessInstance(null);
      Assert.fail(IllegalArgumentException.class.getSimpleName() + " expected");
    } catch (final IllegalArgumentException e) {
      // ok
    }
  }

  public void testCancelWithSubflow() throws BonitaException {
    // case to test :
    // cancel a process with running subflow activities

    final URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");

    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get("subflow");
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(instance1);
    final Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    Assert.assertNotNull(instancesUUID);
    Assert.assertTrue(instancesUUID.size() == 1);
    final ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next();

    Collection<ActivityInstance<TaskInstance>> todolist = this.getQueryRuntimeAPI().getTaskList(childInstanceUUID, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertTrue(todolist.size() == 1);

    this.getRuntimeAPI().cancelProcessInstance(instance1);

    final Set<ProcessInstance> instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
    Assert.assertNotNull(instances);
    Assert.assertEquals(1, instances.size());

    Assert.assertEquals(InstanceState.CANCELLED, instances.iterator().next().getInstanceState());

    todolist = this.getQueryRuntimeAPI().getTaskList(childInstanceUUID, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(0, todolist.size());

    todolist = this.getQueryRuntimeAPI().getTaskList(childInstanceUUID, ActivityState.CANCELLED);
    Assert.assertNotNull(todolist);
    Assert.assertEquals(1, todolist.size());

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  // Test cancel with multiInstantiation.
  public void testCancelMultiInstantiation() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("multiInstantiation.xpdl");
    final String processId = "multiInstantiation";

    final Map<String, ProcessDefinition> processMap = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
      TestOneJoinActivityInstantiator.class));
    final ProcessDefinition process = processMap.get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    final Set<ActivityInstance<ActivityBody>> actIntances = this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
    Assert.assertEquals(2, actIntances.size());

    final Set<Object> varValues = new HashSet<Object>();

    // check two instance of activity r1 have been created
    for (final ActivityInstance<ActivityBody> activityInstance : actIntances) {
      final Map<String, Object> variables = activityInstance.getLastKnownVariableValues();
      Assert.assertTrue(variables.containsKey("testVar"));
      varValues.add(variables.get("testVar"));
    }

    final Set<Object> expectedValues = new HashSet<Object>();
    expectedValues.add("val1");
    expectedValues.add("val2");

    Assert.assertEquals(expectedValues, varValues);

    // cancel instance
    this.getRuntimeAPI().cancelProcessInstance(instanceUUID);

    // check the two tasks have been cancelled.
    for (final ActivityInstance<ActivityBody> activityInstance : actIntances) {
      Assert.assertEquals(ActivityState.CANCELLED,
          this.getQueryRuntimeAPI()
            .getActivityInstance(activityInstance.getUUID())
              .getBody().getState());
    }

    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testCancelSubflowProcessInstance() throws BonitaException {
    // case to test :
    // cancel a child process

    final URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");

    // Check UncancellableInstanceException is thrown when trying to cancel process which have parentInstance
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get("subflow");
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    final ProcessInstance processInstance = this.getQueryRuntimeAPI().getProcessInstance(instance1);
    final Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    Assert.assertNotNull(instancesUUID);
    Assert.assertTrue(instancesUUID.size() == 1);
    final ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next();
    try {
      this.getRuntimeAPI().cancelProcessInstance(childInstanceUUID);
      Assert.fail("Cannot cancel a process instance having a parent instance");
    } catch (final UncancellableInstanceException e) {
      // Ok
      final Set<ProcessInstance> instances = this.getQueryRuntimeAPI().getProcessInstances(processUUID);
      Assert.assertNotNull(instances);
      Assert.assertTrue(instances.size() == 1);

      final Collection<ActivityInstance<TaskInstance>> todolist = this.getQueryRuntimeAPI().getTaskList(ActivityState.READY);
      Assert.assertNotNull(todolist);
      Assert.assertTrue(todolist.size() == 1);
      final TaskUUID taskUUID = todolist.iterator().next().getBody().getUUID();

      this.getRuntimeAPI().startTask(taskUUID, true);
      this.getRuntimeAPI().finishTask(taskUUID, true);
    }

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testCancelFinishedProcessInstance() throws BonitaException {
    // case to test :
    // cancel an archived instance
    final String processId = "deleteInstance";
    final URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instance1 = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(InstanceState.STARTED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    final Collection<ActivityInstance<TaskInstance>>  todolist = this.getQueryRuntimeAPI().getTaskList(instance1, ActivityState.READY);
    Assert.assertNotNull(todolist);
    Assert.assertFalse(todolist.isEmpty());
    Assert.assertTrue(todolist.size() == 1);
    final TaskUUID taskUUID = todolist.iterator().next().getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);
    Assert.assertEquals(InstanceState.FINISHED, this.getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());

    try {
      this.getRuntimeAPI().cancelProcessInstance(instance1);
      Assert.fail("Instance is finished! We must a an exception here!");
    } catch (final UncancellableInstanceException e) {
      // ok
    }

    this.getRuntimeAPI().deleteAllProcessInstances(processUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  private void loginAs(final String loginName, final String password) throws LoginException {
    this.loginContext.logout();
    try {
      this.loginContext = new LoginContext("Bonita", new SimpleCallbackHandler(loginName, password));
      this.loginContext.login();
    } catch (final LoginException e) {
      throw new RuntimeException("Please, configure a JAAS test user with login: "
          + loginName + " and password: " + password, e);
    }
  }


}
