/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.imposterizer.CustomImposterizer;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskFullInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.impl.TaskFullInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;


/**
 * @author Pierre Vigneras
 * @date 16 mai 08
 */
public abstract class WorkflowElementFactory {

  private WorkflowElementFactory() {

  }

  public static PackageFullDefinition getNewPackage(final PackageDefinitionUUID uuid,
      final Date deployedDate, final Date undeployDate, final String packageId) {
    final Mockery mockery = new Mockery();
    final PackageFullDefinition packageDef = mockery.mock(PackageFullDefinition.class);
    mockery.checking(new Expectations() {
      {
        allowing(packageDef).getPackageDefinitionUUID(); will(returnValue(uuid));
        allowing(packageDef).getDeployedDate(); will(returnValue(deployedDate));
        allowing(packageDef).getUndeployedDate(); will(returnValue(undeployDate));
        allowing(packageDef).getUUID(); will(returnValue(uuid));
        allowing(packageDef).getPackageId(); will(returnValue(packageId));
        ignoring(packageDef);
      }
    });

    return packageDef;
  }

  public static ProcessFullInstance getNewInstance(final ProcessInstanceUUID uuid,
      final PackageDefinitionUUID packageUUID,
      final ProcessDefinitionUUID processUUID,
      final ActivityInstance< ? extends ActivityBody>... activities) {
    final CustomImposterizer imposterizer = new CustomImposterizer();
    imposterizer.setClassesToImpost(new Class< ? >[]{ProcessFullInstance.class});
    
    final Mockery mockery = new Mockery() { { setImposteriser(imposterizer); } };
    final ProcessFullInstance instance = mockery.mock(ProcessFullInstance.class);
    final Set<ActivityInstance< ? extends ActivityBody>> activitySet = new HashSet<ActivityInstance< ? extends ActivityBody>>();
    if (activities != null) {
      activitySet.addAll(Arrays.asList(activities));
    }

    mockery.checking(new Expectations() {
      {
        allowing(instance).getPackageDefinitionUUID(); will(returnValue(packageUUID));
        allowing(instance).getProcessDefinitionUUID(); will(returnValue(processUUID));
        allowing(instance).getProcessInstanceUUID(); will(returnValue(uuid));
        allowing(instance).getUUID(); will(returnValue(uuid));
        allowing(instance).getActivities(); will(returnValue(activitySet));
        if (activities != null) {
          for (ActivityInstance< ? extends ActivityBody> activity : activities) {
            final String id = activity.getActivityId();
            allowing(instance).getActivity(with(same(id)), with(any(String.class)), with(any(String.class)));
            will(returnValue(activity));
          }
        }
        ignoring(instance);
      }
    });

    return instance;
  }

  @SuppressWarnings("unchecked")
  public static ActivityFullInstance<TaskInstance> getNewActivity(final ActivityInstanceUUID uuid,
      final String activityId,
      final PackageDefinitionUUID packageUUID,
      final ProcessDefinitionUUID processUUID,
      final ProcessInstanceUUID instanceUUID,
      final String iterationId,
      final String activityInstanceId) {

    final Mockery mockery = new Mockery();
    final ActivityFullInstance<TaskInstance> activity = mockery.mock(ActivityFullInstance.class);
    mockery.checking(new Expectations() {
      {
        allowing(activity).getPackageDefinitionUUID(); will(returnValue(packageUUID));
        allowing(activity).getProcessDefinitionUUID(); will(returnValue(processUUID));
        allowing(activity).getProcessInstanceUUID(); will(returnValue(instanceUUID));
        allowing(activity).getUUID(); will(returnValue(uuid));
        allowing(activity).getActivityId(); will(returnValue(activityId));
        allowing(activity).getBody(); will(returnValue(getNewTask()));
        allowing(activity).getIterationId(); will(returnValue(iterationId));
        allowing(activity).getActivityInstanceId(); will(returnValue(activityInstanceId));
        ignoring(activity);
      }
    });

    return activity;
  }

  public static TaskInstance getNewTask() {
    final Mockery mockery = new Mockery();
    final TaskInstance task = mockery.mock(TaskFullInstance.class);
    final TaskUUID taskUUID = IdFactory.getNewTaskUUID();
    final Date d = new Date();
    mockery.checking(new Expectations() {
      {
        allowing(task).getUUID();will(returnValue(taskUUID));
        allowing(task).copy();will(
          returnValue(new TaskFullInstanceImpl(taskUUID, ActivityState.INITIAL, d, "system")));
        ignoring(task);
      }
    });
    return task;
  }
}
