/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.hook.misc;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.ClassDataTool;
import org.ow2.bonita.util.Misc;

/**
 * @author "Charles Souillard"
 */
public class MiscHookTest extends APITestCase {

  public void testIncoherentHook() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "incoherentHook";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        IncrementHook.class, FailingHook.class, IncoherentHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testFailingHook() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "failingHook";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        IncrementHook.class, FailingHook.class, IncoherentHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      getRuntimeAPI().instantiateProcess(processUUID);
      fail("This test must have fail because the hook was in transaction and throwing an exception");
    } catch (final Exception e) {
      getRuntimeAPI().deleteAllProcessInstances(processUUID);
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
      assertTrue(causeContains("FailingHookException", e));
    }
  }

  public void testManyHooks() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "manyHooks";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        IncrementHook.class, FailingHook.class, IncoherentHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    final String counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    assertEquals("3", counter);
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testGlobalHook() throws BonitaException, IOException {
    // deploy a global class used in a process and check it is found
    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "manyHooks";
    getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class, FailingHook.class, IncoherentHook.class));

    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    final String counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    assertEquals("3", counter);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().removeClasses(
        new String[]{
            IncrementHook.class.getName(),
            FailingHook.class.getName(),
            IncoherentHook.class.getName()
        });
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeployGlobalTwiceKO() throws BonitaException, IOException {
    //deploy two times the same class (className) and check that an exception is thrown
    getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class));
    try {
      getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class));
      fail("This test should have fail !!");
    } catch (final DeploymentException e) {
      getManagementAPI().removeClass(IncrementHook.class.getName());
      assertEquals("There is already a deployed class with this name: " + IncrementHook.class.getName(), e.getMessage());
    }
  }

  public void testDeployGlobalTwiceOK() throws BonitaException, IOException {
    //deploy two times the same class (className) and check that an exception is thrown
    getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class));
    getManagementAPI().removeClass(IncrementHook.class.getName());
    getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class));
    getManagementAPI().removeClass(IncrementHook.class.getName());
  }

  public void testDeployProcessAndGlobalHook() throws BonitaException, IOException {
    //deploy the same class in process and global, check that the process one is used
    final InputStream in = this.getClass().getResourceAsStream("IncrementHook2.bytecode");
    if (in == null) {
      throw new IllegalArgumentException("Can't find ressource: IncrementHook2.bytecode");
    }
    final byte[] incrementHook = Misc.getAllContentFrom(in);
    getManagementAPI().deployClass(incrementHook);

    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "manyHooks";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        IncrementHook.class, FailingHook.class, IncoherentHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    String counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    assertEquals("3", counter);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);

    //do not deploy IncrementHook in package and check the global one is called
    process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        FailingHook.class, IncoherentHook.class)).get(processId);
    packageUUID = process.getPackageDefinitionUUID();
    processUUID = process.getUUID();
    instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    assertEquals("6", counter);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().removeClass(IncrementHook.class.getName());
  }

  public void testUndeploy() throws BonitaException, IOException {
    //check undeploy forbidden : try to undeploy a global used by a process : must fail
    getManagementAPI().deployClasses(ClassDataTool.getClasses(IncrementHook.class, FailingHook.class));
    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "manyHooks";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        FailingHook.class, IncoherentHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    //this undeploy must be OK as the process defines its own FailingHook
    getManagementAPI().removeClass(FailingHook.class.getName());
    try {
      getManagementAPI().removeClass(IncrementHook.class.getName());
      fail("It must be forbidden to undeploy a class used by an existing process");
    } catch (final DeploymentException e) {
      assertEquals("Package is still using global class. "
          + "Cannot undeploy class." + " className: " + e.getClassName()
          + " packageDefinitionUUID: " + e.getPackageDefinitionUUID(), e.getMessage());
    }
    //here undeploy the referenced process
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    //check undeploy OK : try to undeploy a global not used anymore : must succeed
    getManagementAPI().removeClass(IncrementHook.class.getName());
    //check impossible to undeploy an unexisting class...
    try {
      getManagementAPI().removeClass(IncrementHook.class.getName());
      fail("This test must have failed as it is impossible to remove an unexisting class...");
    } catch (final DeploymentException e) {
      //assertTrue(causeContains("There is no class with name " + IncrementHook.class.getName()
      //    + " defined in global class repository.", e));
      assertEquals("There is no class defined in global class repository  with name: " + e.getClassName(), e.getMessage());

    }
  }

  public void testRedeploy() throws BonitaException, IOException {
    //undeploy a class from global, deploy a new one with the same name and check the new one is used
    final InputStream in = this.getClass().getResourceAsStream("IncrementHook2.bytecode");
    if (in == null) {
      throw new IllegalArgumentException("Can't find ressource: IncrementHook2.bytecode");
    }
    final byte[] incrementHook = Misc.getAllContentFrom(in);


    getManagementAPI().deployClasses(getClasses(IncrementHook.class));
    getManagementAPI().removeClass(IncrementHook.class.getName());
    getManagementAPI().deployClass(incrementHook);

    final URL xpdlUrl = this.getClass().getResource("miscHook.xpdl");
    final String processId = "manyHooks";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        FailingHook.class, IncoherentHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"uniqueActivity"});
    final String counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "counter");
    assertEquals("6", counter);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().removeClass(IncrementHook.class.getName());
  }

  public void testZip() {

  }

}
