/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.integration.cycle;

import java.net.URL;
import java.util.Collection;
import java.util.Iterator;

import junit.framework.Assert;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;

/**
 * Test. Deploys a xpdl process.
 *
 * @author Miguel Valdes
 */
public class CycleTest extends APITestCase {


  public void testBadIteration() {
    final URL xpdlUrl = this.getClass().getResource("badIteration.xpdl");
    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl));
      Assert.fail("Expected DeploymentException");
    } catch (final DeploymentException e) {
      // ok
      Assert.assertTrue(e.getMessage().contains("iteration points to an unknown node: cannot find a node with id 'a2'"));
    }

  }

  /*
   * Test a simple cycle.
   * Correct execution is: init a1 a2 a4 a1 a2 end
   */
  public void testSimpleCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("simpleCycle.xpdl");
    final String processId = "simpleCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a cycle with a split/and and a Join/and
   * Correct execution is init -> a1 -> a3 -> a2 -> a4 -> a1 -> a3 -> a2 -> end
   *                               |--> a5 --|              |--> a5 --|
   */
  public void testSplitAndJoinCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("splitAndJoinCycle.xpdl");
    final String processId = "splitAndJoinCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());
    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  public void testSplitXorJoinCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("splitXorJoinCycle.xpdl");
    final String processId = "splitXorJoinCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    int nb = 0;
    try {
      nb += this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size();
    } catch (final ActivityNotFoundException e) {
      // TODO: handle exception
    }
    try {
      nb += this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size();
    } catch (final ActivityNotFoundException e) {
      // TODO: handle exception
    }
    Assert.assertTrue(nb >= 2);
    Assert.assertTrue(nb <= 4);

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a simple proEd cycle.
   * Correct execution is: init a1 a2 a4 a1 a2 end
   */
  public void testSimpleCycleProEd() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("simpleCycleProEd.xpdl");
    final String processId = "simpleCycleProEd";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkStopped(instanceUUID, new String[]{"init", "a1"});

    Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());

    TaskInstance task = tasks.iterator().next().getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkStopped(instanceUUID, new String[]{"init", "a2"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());

    task = tasks.iterator().next().getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a nested cycle.
   * Correct execution is: init a1 a3 a2 a4 a3 a2 a6 a5 a1 a3 a2 a6 end
   */
  public void testNestedCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("nestedCycle.xpdl");
    final String processId = "nestedCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class, HookBeforeTerminateUpdateVariable2.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();


    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "a5", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(3, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(3, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a6").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }


  /*
   * Test an arbitrary cycle.
   * Correct execution is: init a1 a3 a2 a4 a1 a3 a2 a6 a5 a3 a2 a6 end
   */
  public void testArbitraryCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("arbitraryCycle.xpdl");
    final String processId = "arbitraryCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class, HookBeforeTerminateUpdateVariable2.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "a5", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(3, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(3, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a6").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a simple proEd cycle with an extra entry point.
   * Entry point with AND join is not allowed
   * Correct execution is: init a3  a1 a2 a4 a1 a2 end
   */
  public void testEntryAndJoinProEd() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("entryAndCycle.xpdl");
    final String processId = "entryAndCycle";
    try {
      this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
      Assert.fail("iteration with AND join is not allowed");
    } catch (final DeploymentException e) {
      Assert.assertTrue(e.getMessage().contains("AND join"));
    }
/*    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"init", "a1", "a3"});

    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    assertEquals(1, tasks.size());

    TaskInstance task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkStopped(instanceUUID, new String[]{"init", "a2"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    assertEquals(1, tasks.size());

    task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
*/
  }

  /*
   * Test a simple proEd cycle with an extra entry point.
   * Correct execution is: init a0/a3  a2 a1 a4 a2 a1 a4 a2 end
   */
  public void testEntryXorJoinProEd() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("entryXorCycle.xpdl");
    final String processId = "entryXorCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkStopped(instanceUUID, new String[]{"init"});

    // Check there are two tasks. Execute task from a3
    Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());
    Iterator<ActivityInstance<TaskInstance>> it = tasks.iterator();
    ActivityInstance<TaskInstance> act = it.next();
    if ("a0".equals(act.getActivityId())) {
      act = it.next();
    }
    Assert.assertEquals("a3",act.getActivityId());
    TaskInstance task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkStopped(instanceUUID, new String[]{"init", "a3", "a2", "a1"});

    // Tasks from a4 and from a0 has been created, task from a0 has been removed
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    act = tasks.iterator().next();
    Assert.assertEquals("a4",act.getActivityId());
    task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());

    act = tasks.iterator().next();
    Assert.assertEquals("a4",act.getActivityId());
    task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    Assert.assertEquals(3, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);

    ///////////////////////////////////////////
    // Check same test when entering from a0 //
    ///////////////////////////////////////////
    instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);
    this.checkStopped(instanceUUID, new String[]{"init"});

    // Check there are two tasks. Execute task from a3
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());
    it = tasks.iterator();
    act = it.next();
    if ("a3".equals(act.getActivityId())) {
      act = it.next();
    }
    Assert.assertEquals("a0",act.getActivityId());
    task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkStopped(instanceUUID, new String[]{"init", "a0", "a1"});

    // Tasks from a4 has been created
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    act = tasks.iterator().next();
    Assert.assertEquals("a4",act.getActivityId());
    task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    Assert.assertEquals(1, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());

    act = tasks.iterator().next();
    Assert.assertEquals("a4",act.getActivityId());
    task = act.getBody();
    this.getRuntimeAPI().startTask(task.getUUID(), true);
    this.getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    this.getRuntimeAPI().finishTask(task.getUUID(), true);

    this.checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);


    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * Checks that execution of a task in a previous iteration
   * (left there because of a Join XOR)
   * doesn't propagate to the new iteration.
   * @throws BonitaException
   */
  public void testJoinXorTaskCycle() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("joinXorTaskCycle.xpdl");
    final String processId = "joinXorTaskCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    // task a3 and a5 are available
    String iterationId = null;
    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a3")) {
        this.getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        this.getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
        iterationId = activity.getIterationId();
      }
    }
    Assert.assertNotNull(iterationId);

    // task a3 has been done, a5 has been aborted by join XOR a2
    // new cycle -> 2 tasks a3 and a5
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a5")) {
        // check task does not belong to previous iteration
        Assert.assertFalse(iterationId.equals(activity.getIterationId()));
      }
    }

    this.checkStopped(instanceUUID, new String[]{"init", "a2"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a5")) {
        this.getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        this.getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
      }
    }

    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * Checks that JoinXOR cancel works in cycles with exit points
   * doesn't propagate to the new iteration.
   * @throws BonitaException
   */
  public void testJoinXorTaskCycleExit() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("joinXorTaskCycleExit.xpdl");
    final String processId = "joinXorTaskCycle";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    // task a3 and a5 are available
    String iterationId = null;
    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a5")) {
        this.getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        this.getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
        iterationId = activity.getIterationId();
      }
    }
    Assert.assertNotNull(iterationId);

    // task a5 has been done, a3 has been aborted by join XOR a2
    // new cycle -> 2 tasks a3 and a5
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a3")) {
        // check task does not belong to previous iteration
        Assert.assertFalse(iterationId.equals(activity.getIterationId()));
      }
    }

    this.checkStopped(instanceUUID, new String[]{"init", "a2"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(2, tasks.size());

    // execute task a3
    for (final ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a3")) {
        this.getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        this.getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
      }
    }

    // check task a5 has been aborted
    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(0, tasks.size());

    // check execution is finished
    this.checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "BonitaEnd"});
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    Assert.assertEquals(1, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    Assert.assertEquals(2, this.getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }


  /*
   * Test a nested cycle with all nodes in the super cycle.
   * Correct execution is: init step0 step1 step2 step1 step2 step3 step0 step1 step2 step3 end
   */
  public void testNestedCycleNoExtraNodes() throws BonitaException {
    final URL xpdlUrl = this.getClass().getResource("testNestedCycleNoExtraNodes.xpdl");
    final String processId = "testNestedCycleNoExtraNodes";
    final ProcessDefinition process = this.getManagementAPI().deployBar(this.getBarContent(xpdlUrl)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();


    final ProcessInstanceUUID instanceUUID = this.getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance<TaskInstance>> tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    ActivityInstance<TaskInstance> task = tasks.iterator().next();
    Assert.assertEquals("init", task.getActivityId());
    TaskUUID taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);


    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step0", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "loop1", "yes");
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step3", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step0", task.getActivityId());
    this.getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "loop2", "yes");
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    tasks = this.getQueryRuntimeAPI().getTaskList(instanceUUID, ActivityState.READY);
    Assert.assertEquals(1, tasks.size());
    task = tasks.iterator().next();
    Assert.assertEquals("step3", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    this.getRuntimeAPI().startTask(taskUUID, true);
    this.getRuntimeAPI().finishTask(taskUUID, true);

    Assert.assertNotNull(this.getQueryRuntimeAPI().getProcessInstance(instanceUUID).getEndedDate());
    this.getRuntimeAPI().deleteProcessInstance(instanceUUID);
    this.getManagementAPI().undeploy(packageUUID);
    this.getManagementAPI().deletePackage(packageUUID);
  }
}
