/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.services.record;

import java.util.Date;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.WorkflowElementFactory;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.PackageFullDefinitionImpl;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.pvm.env.Environment;
import org.ow2.bonita.pvm.internal.cmd.Command;
import org.ow2.bonita.pvm.internal.cmd.CommandService;
import org.ow2.bonita.services.Archivable;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Misc;

/**
 * @author Pierre Vigneras
 * @date Feb 27, 2008
 */
public class HistoryTest extends APITestCase {

  private static final int INSERTION_NB_MIN = 5;
  private static final int INSERTION_NB_MAX = 10;

  protected CommandService commandService;

  @Override
  public void setUp() throws Exception {
    super.setUp();
    this.commandService = GlobalEnvironmentFactory.getEnvironmentFactory().get(CommandService.class);
  }

  @Override
  public void tearDown() throws Exception {
    this.commandService = null;
    super.tearDown();
  }

  public void testArchiveSeveralTimeSamePackage() throws BonitaException {
    final String packageId = IdFactory.getNewPackageId();
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
      final long date = new Date().getTime();
      final PackageFullDefinition pack = WorkflowElementFactory.getNewPackage(packageUUID,
          new Date(date),
          new Date(date + 100),
          packageId);
      archivePackage(pack);
    }

    Set<PackageFullDefinition> records = getPackages();
    assertTrue(records.toString(), records.size() == n);
    records = getPackages(packageId);
    assertTrue(records.toString(), records.size() == n);
    for (final PackageFullDefinition pack : records) {
      getManagementAPI().deletePackage(pack.getUUID());
    }
  }

  public void testArchiveDifferentPackages() throws BonitaException {
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      final long date = new Date().getTime();
      final String packageId = IdFactory.getNewPackageId();
      final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
      final PackageFullDefinition pack = WorkflowElementFactory.getNewPackage(packageUUID,
          new Date(date),
          new Date(date + 100),
          packageId);
      archivePackage(pack);
    }
    final Set<PackageFullDefinition> records = getPackages();
    assertTrue(records.toString(), records.size() == n);
    for (final PackageFullDefinition pack : records) {
      getManagementAPI().deletePackage(pack.getUUID());
    }
  }

  public void testLastArchivedPackage() throws BonitaException {
    final String packageId = IdFactory.getNewPackageId();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final long date = new Date().getTime();
    final PackageFullDefinition record = WorkflowElementFactory.getNewPackage(packageUUID,
        new Date(date + 1000),
        new Date(date + 2000),
        packageId);
    archivePackage(record);
    final PackageDefinition last = getLastDeployedPackage(packageId, PackageDefinition.PackageState.UNDEPLOYED);
    assertNotNull(last);
    assertEquals(last.getPackageDefinitionUUID(), record.getPackageDefinitionUUID());

    final PackageDefinition pd = getPackage(record.getPackageDefinitionUUID());
    assertNotNull(Misc.deepToString(record), pd);
    assertEquals(Misc.deepToString(pd), record.getPackageDefinitionUUID(), pd.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(record.getUUID());
  }

  public void testGetNonExistentPackage() {
    final PackageDefinition record = getLastDeployedPackage(IdFactory.getNewPackageId(),
        PackageDefinition.PackageState.UNDEPLOYED);
    assertNull(record);
    final Set<PackageFullDefinition> records = getPackages(IdFactory.getNewPackageId());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetPackageNull() {
    try {
      getLastDeployedPackage(null, null);
      fail("Should not accept a null packageId!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }

    try {
      getPackages((String) null);
      fail("Should not accept a null packageId!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
  }

  public void testArchiveNullPackage() {
    try {
      archivePackage((PackageFullDefinition) null);
      fail("Should not accept a null packageFullDef!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
  }

  public void testArchiveSeveralInstances() throws BonitaException {
    ProcessInstanceUUID instanceUUID;
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();

    // Archive multiple instances of the same process
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      instanceUUID = IdFactory.getNewInstanceUUID();
      final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
    }
    final Set<ProcessFullInstance> records = getProcessInstances(processUUID);
    assertTrue(records.size() + " == " + n + "?", records.size() == n);
    for (final ProcessFullInstance instance : records) {
      getRuntimeAPI().deleteProcessInstance(instance.getUUID());
    }

  }

  public void testArchiveOneInstanceAndCheckSomeFields() throws BonitaException {
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();
    final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID, packageUUID, processUUID);
    archiveProcess(instance);
    final ProcessFullInstance newRecord = getProcessInstance(instanceUUID);
    assertNotNull(newRecord);
    assertEquals(newRecord.getUUID(), instance.getUUID());
    assertEquals(newRecord.getProcessInstanceUUID(), instance.getProcessInstanceUUID());
    assertEquals(newRecord.getProcessDefinitionUUID(), instance.getProcessDefinitionUUID());
    assertEquals(newRecord.getParentInstanceUUID(), instance.getParentInstanceUUID());
    assertEquals(newRecord.getInitialVariableValues(), instance.getInitialVariableValues());
    getRuntimeAPI().deleteProcessInstance(instance.getUUID());
  }

  public void testArchiveSameInstance() throws BonitaException {
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();

    ProcessFullInstance instance = null;
    try {
      instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
      instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
      fail("It should not be possible to archive twice the same instance!");
    } catch (final IllegalArgumentException e) {
      // ok
      getRuntimeAPI().deleteProcessInstance(instance.getUUID());
    }
  }

  public void testArchiveNullProcessInstance() {
    try {
      archiveProcess((ProcessFullInstance) null);
      fail("Should not accept a null packageFullDef!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
  }

  public void testGetNonExistentProcessInstance() {
    final ProcessFullInstance record = getProcessInstance(IdFactory.getNewInstanceUUID());
    assertNull(record);

    final Set<ProcessFullInstance> records = getProcessInstances(IdFactory.getNewProcessUUID());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetProcessInstanceNull() {
    try {
      getProcessInstance(null);
      fail("Should not accept a null processInstanceUUID!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
    try {
      getProcessInstances(null);
      fail("Should not accept a null processInstanceUUID!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
  }

  @SuppressWarnings("unchecked")
  public void testActivity() throws BonitaException {
    final String activityId1 = IdFactory.getNewActivityId();
    final String activityId2 = IdFactory.getNewActivityId();
    final ActivityInstanceUUID activity1UUID = IdFactory.getNewActivityUUID();
    final ActivityInstanceUUID activity2UUID = IdFactory.getNewActivityUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final String iterationId = IdFactory.getNewIterationId();
    final String activityInstanceId = IdFactory.getNewActivityInstanceId();

    final ActivityFullInstance<TaskInstance> activity1 = WorkflowElementFactory.getNewActivity(activity1UUID,
        activityId1,
        packageUUID,
        processUUID,
        instanceUUID,
        iterationId,
        activityInstanceId);
    final ActivityFullInstance<TaskInstance> activity2 = WorkflowElementFactory.getNewActivity(activity2UUID,
        activityId2,
        packageUUID,
        processUUID,
        instanceUUID,
        iterationId,
        activityInstanceId);

    final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID,
        packageUUID,
        processUUID,
        activity1,
        activity2);

    archiveProcess(instance);

    final ProcessFullInstance historyInstance = getProcessInstance(instanceUUID);
    assertNotNull(historyInstance);
    final Set<ActivityFullInstance<ActivityBody>> instanceActivities = historyInstance.getActivities();
    assertNotNull(instanceActivities);
    assertEquals(2, instanceActivities.size());

    for (final ActivityFullInstance<ActivityBody> historyActivity : instanceActivities) {
      final String activityId = historyActivity.getActivityId();
      if (activityId.equals(activityId1)) {
        assertActivityEquals(activity1, historyActivity);
      } else if (activityId.equals(activityId2)) {
        assertActivityEquals(activity2, historyActivity);
      } else {
        fail("Wrong Id: " + activityId);
      }
    }

    assertActivityEquals(activity1, getActivityInstance(instanceUUID, activityId1, iterationId, activityInstanceId));
    assertActivityEquals(activity2, getActivityInstance(instanceUUID, activityId2, iterationId, activityInstanceId));

    getRuntimeAPI().deleteProcessInstance(instance.getUUID());
  }

  public void testGetNonExistentActivityInstance() {
    Set<ActivityFullInstance<ActivityBody>> records;
    ActivityFullInstance<ActivityBody> record;
    record = getActivityInstance(IdFactory.getNewActivityUUID());
    assertNull(record);
    records = getActivityInstances(IdFactory.getNewInstanceUUID());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetActivityInstanceNull() {
    try {
      getActivityInstance(null);
      fail("Should not accept a null activityInstanceUUID!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
    try {
      getActivityInstances(null);
      fail("Should not accept a null activityInstanceUUID!");
    } catch (final IllegalArgumentException iae) {
      // ok
    }
  }

  private void assertActivityEquals(final ActivityFullInstance< ? extends ActivityBody> expected,
      final ActivityFullInstance< ? extends ActivityBody> actual) {
    assertNotNull(actual);
    assertNotNull(expected);
    assertEquals(expected.getUUID(), actual.getUUID());
    assertEquals(expected.getActivityId(), actual.getActivityId());
    assertEquals(expected.getProcessDefinitionUUID(), actual.getProcessDefinitionUUID());
    assertEquals(expected.getProcessInstanceUUID(), actual.getProcessInstanceUUID());
    assertEquals(expected.getIterationId(), actual.getIterationId());
    assertTrue(expected.getBody() instanceof TaskInstance);
    assertEquals(expected.getBody().getUUID(), actual.getBody().getUUID());
  }

  private void archiveProcess(final ProcessFullInstance processInstance) {
    this.commandService.execute(new ArchiveCommand(processInstance));
  }

  private void archivePackage(final PackageFullDefinition packageDef) {
    this.commandService.execute(new ArchiveCommand(packageDef));
  }

  private Set<PackageFullDefinition> getPackages() {
    return this.commandService.execute(new GetPackagesCommand());
  }

  private Set<PackageFullDefinition> getPackages(final String packageId) {
    return this.commandService.execute(new GetPackagesCommandWithId(packageId));
  }

  private PackageFullDefinition getLastDeployedPackage(final String packageId, final PackageDefinition.PackageState packageState) {
    return this.commandService.execute(new GetLastDeployedPackage(packageId, packageState));
  }

  private ActivityFullInstance<ActivityBody> getActivityInstance(final ActivityInstanceUUID instanceUUID) {
    return this.commandService.execute(new GetActivityInstanceOnlyUUID(instanceUUID));
  }

  private ActivityFullInstance<ActivityBody> getActivityInstance(final ProcessInstanceUUID instanceUUID,
      final String activityId, final String iterationId, final String activityInstanceId) {
    return this.commandService.execute(new GetActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId));
  }

  private Set<ActivityFullInstance<ActivityBody>> getActivityInstances(final ProcessInstanceUUID processInstanceUUID) {
    return this.commandService.execute(new GetActivityInstances(processInstanceUUID));
  }

  private ProcessFullInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID) {
    return this.commandService.execute(new GetProcessInstance(processInstanceUUID));
  }

  private Set<ProcessFullInstance> getProcessInstances(final ProcessDefinitionUUID processDefUUID) {
    return this.commandService.execute(new GetProcessInstances(processDefUUID));
  }

  private PackageFullDefinition getPackage(final PackageDefinitionUUID packDefUUID) {
    return this.commandService.execute(new GetPackageWithUUID(packDefUUID));
  }
  private Querier getHistory() {
    return EnvTool.getHistoryQueriers();
  }
  /* List of commands */

  class GetPackageWithUUID implements Command<PackageFullDefinition> {
    /**
     * 
     */
    private static final long serialVersionUID = -3798444175939857662L;
    private final PackageDefinitionUUID packDefUUID;

    public GetPackageWithUUID(final PackageDefinitionUUID packDefUUID) {
      this.packDefUUID = packDefUUID;
    }
    public PackageFullDefinition execute(final Environment environment) throws Exception {
      return new PackageFullDefinitionImpl(getHistory().getPackage(this.packDefUUID));
    }
  }

  class GetProcessInstances implements Command<Set<ProcessFullInstance>> {
    /**
     * 
     */
    private static final long serialVersionUID = 8568074371790934617L;
    private final ProcessDefinitionUUID processDefUUID;

    public GetProcessInstances(final ProcessDefinitionUUID processDefUUID) {
      this.processDefUUID = processDefUUID;
    }
    public Set<ProcessFullInstance> execute(final Environment environment) throws Exception {
      return getHistory().getProcessInstances(this.processDefUUID);
    }
  }

  class GetProcessInstance implements Command<ProcessFullInstance> {
    /**
     * 
     */
    private static final long serialVersionUID = -1233442971215657292L;
    private final ProcessInstanceUUID processInstanceUUID;

    public GetProcessInstance(final ProcessInstanceUUID instanceUUID) {
      this.processInstanceUUID = instanceUUID;
    }
    public ProcessFullInstance execute(final Environment environment) throws Exception {
      final ProcessFullInstance processInstance = getHistory().getProcessInstance(this.processInstanceUUID);
      if (processInstance != null) {
        // Load process instance attributes to prevent LazyInitializationException
        for (final Object o : processInstance.getInitialVariableValues().values()) {
          o.toString();
        }
        for (final Object o : processInstance.getLastKnownVariableValues().values()) {
          o.toString();
        }
        for (final VariableUpdate o : processInstance.getVariableUpdates()) {
          o.toString();
        }
        for (final ActivityFullInstance<ActivityBody> activity : processInstance.getActivities()) {
          activity.toString();
        }
      }
      return processInstance;
    }

  }

  class GetActivityInstances implements Command<Set<ActivityFullInstance<ActivityBody>>> {
    /**
     * 
     */
    private static final long serialVersionUID = -2133035331171368188L;
    private final ProcessInstanceUUID processInstanceUUID;

    public GetActivityInstances(final ProcessInstanceUUID instanceUUID) {
      this.processInstanceUUID = instanceUUID;
    }
    public Set<ActivityFullInstance<ActivityBody>> execute(final Environment environment) throws Exception {
      return getHistory().getActivityInstances(this.processInstanceUUID);
    }

  }

  class GetActivityInstance implements Command<ActivityFullInstance<ActivityBody>> {

    /**
     * 
     */
    private static final long serialVersionUID = -7665263370176516500L;
    private final ProcessInstanceUUID processInstanceUUID;
    private final String activityId;
    private final String iterationId;
    private final String activityInstanceId;

    public GetActivityInstance(final ProcessInstanceUUID instanceUUID,
        final String activityId, final String iterationId, final String activityInstanceId) {
      this.processInstanceUUID = instanceUUID;
      this.activityId = activityId;
      this.iterationId = iterationId;
      this.activityInstanceId = activityInstanceId;
    }
    public ActivityFullInstance<ActivityBody> execute(final Environment environment) throws Exception {
      return getHistory().getActivityInstance(this.processInstanceUUID, this.activityId, this.iterationId, this.activityInstanceId);
    }

  }

  class GetActivityInstanceOnlyUUID implements Command<ActivityFullInstance<ActivityBody>> {
    /**
     * 
     */
    private static final long serialVersionUID = -7788725939578050007L;
    private final ActivityInstanceUUID actInstanceUUID;

    public GetActivityInstanceOnlyUUID(final ActivityInstanceUUID instanceUUID) {
      this.actInstanceUUID = instanceUUID;
    }
    public ActivityFullInstance<ActivityBody> execute(final Environment environment) throws Exception {
      return getHistory().getActivityInstance(this.actInstanceUUID);
    }

  }

  class GetLastDeployedPackage implements Command<PackageFullDefinition> {

    /**
     * 
     */
    private static final long serialVersionUID = -1125399157847476928L;
    private final String packageId;
    private final PackageDefinition.PackageState packageState;

    public GetLastDeployedPackage(final String packageId, final PackageDefinition.PackageState packageState) {
      this.packageId = packageId;
      this.packageState = packageState;
    }
    public PackageFullDefinition execute(final Environment environment) throws Exception {
      final PackageFullDefinition pack = getHistory().getLastDeployedPackage(this.packageId, this.packageState);
      return pack;
    }
  }

  class GetPackagesCommand implements Command<Set<PackageFullDefinition>> {

    /**
     * 
     */
    private static final long serialVersionUID = 5874326128608389947L;

    public Set<PackageFullDefinition> execute(final Environment environment) throws Exception {
      final Set<PackageFullDefinition> packages = getHistory().getPackages();
      if (packages != null) {
        //Load packages to prevent LazyInitializationException
        packages.toString();
      }
      return packages;
    }
  }

  class GetPackagesCommandWithId implements Command<Set<PackageFullDefinition>> {

    /**
     * 
     */
    private static final long serialVersionUID = 6559499951223096884L;
    private String packageId = null;

    public GetPackagesCommandWithId(final String packageId) {
      this.packageId = packageId;
    }

    public Set<PackageFullDefinition> execute(final Environment environment) throws Exception {
      final Set<PackageFullDefinition> packages = getHistory().getPackages(this.packageId);
      if (packages != null) {
        //Load packages to prevent LazyInitializationException
        packages.toString();
      }
      return packages;
    }
  }

  class ArchiveCommand implements Command<Archivable> {

    /**
     * 
     */
    private static final long serialVersionUID = -285059260009674418L;
    private PackageFullDefinition packageDef = null;
    private ProcessFullInstance processInstance = null;

    public ArchiveCommand(final PackageFullDefinition packageDef) {
      this.packageDef = packageDef;
    }
    public ArchiveCommand(final ProcessFullInstance processInstance) {
      this.processInstance = processInstance;
    }

    public Archivable execute(final Environment environment) throws Exception {
      final Archiver archiver = EnvTool.getArchiver();
      if (this.packageDef != null) {
        archiver.archive(this.packageDef);
        return this.packageDef;
      }
      archiver.archive(this.processInstance);
      return this.processInstance;
    }
  }
}
