/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.variable;

import java.net.URL;
import java.util.Date;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.TypeMismatchException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.DateUtil;


/**
 * @author Guillaume Porcher
 *
 */
public class VariableBasicTypeTest extends APITestCase {
  
  public void testStringVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    String str = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string");
    assertNotNull(str);
    assertEquals("initial value", str);
    
    // set  with valid value
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "string", "new value");
    
    str = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string");
    assertNotNull(str);
    assertEquals("new value", str);
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "string", new Date());
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    str = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string");
    assertNotNull(str);
    assertEquals("new value", str);
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testBooleanVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    Boolean bool = (Boolean) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "bool");
    assertNotNull(bool);
    assertTrue(bool);
    
    // set  with valid value
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "bool", Boolean.FALSE);
    
    bool = (Boolean) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "bool");
    assertNotNull(bool);
    assertFalse(bool);
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "bool", "test");
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    bool = (Boolean) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "bool");
    assertNotNull(bool);
    assertFalse(bool);
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testFloatVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    Double floatVar = (Double) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "float_var");
    assertNotNull(floatVar);
    assertEquals(42.42, floatVar);
    
    // set with valid value
    // don't use Math.PI because some db have a lower precision than java) 
    double testValue = 3.14159265358979;
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "float_var", testValue);
    
    floatVar = (Double) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "float_var");

    assertNotNull(floatVar);
    assertEquals(testValue, floatVar);
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "float_var", "test");
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    floatVar = (Double) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "float_var");
    assertNotNull(floatVar);
    assertEquals(testValue, floatVar);
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testIntegerVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    Long intVar = (Long) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "integer");
    assertNotNull(intVar);
    assertEquals(42L, intVar.intValue());
    
    // set  with valid value
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "integer", 28L);
    
    intVar = (Long) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "integer");
    assertNotNull(intVar);
    assertEquals(28L, intVar.intValue());
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "integer", "test");
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    intVar = (Long) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "integer");
    assertNotNull(intVar);
    assertEquals(28L, intVar.intValue());
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testDatetimeVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    Date dateVar = (Date) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "datetime");
    assertNotNull(dateVar);
    assertEquals(DateUtil.parseDate("2008/07/31/14/00/00"), dateVar);
    
    // set  with valid value
    Date newDate = new Date();
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "datetime", newDate);
    
    dateVar = (Date) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "datetime");
    assertNotNull(dateVar);
    assertEquals(newDate, dateVar);
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "datetime", "test");
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    dateVar = (Date) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "datetime");
    assertNotNull(dateVar);
    assertEquals(newDate, dateVar);
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testPerformerVariable() throws Exception {
    URL xpdlUrl = ActivityVariableTest.class.getResource("basicTypeVariables.xpdl");
    String processId = "basicTypeVariables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    // initial value
    String performer = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "performer");
    assertNotNull(performer);
    assertEquals("user", performer);
    
    // set  with valid value
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "performer", "user2");
    
    performer = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "performer");
    assertNotNull(performer);
    assertEquals("user2", performer);
    
    // set with invalid value
    try {
      getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "performer", "badUser");
      fail("expecting exception");
    } catch (TypeMismatchException e) {
      // ok
    }
    
    // check that the value has not changed 
    performer = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "performer");
    assertNotNull(performer);
    assertEquals("user2", performer);
    
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
}
