package org.ow2.bonita.versioning;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.hook.misc.IncrementHook;
import org.ow2.bonita.hook.misc.MiscHookTest;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BusinessArchiveFactory;
import org.ow2.bonita.util.Misc;

public class VersioningTest extends APITestCase {

  private final String packageId = "workflowPackage";
  private final String processId = "workflowProcess"; 
  
  /* ************************
     TEST DEPLOYMENT FAILURE
  ****************************/
  
  public void testTwiceSameBothVersionDeployment() throws BonitaException {
    testFirstDeployOkSecondFail("WorkflowPack1.0Proc1.0.xpdl", "WorkflowPack1.0Proc1.0.xpdl");
  }
  
  public void testTwiceSamePackageVersionDeployment() throws BonitaException {
    testFirstDeployOkSecondFail("WorkflowPack1.0Proc1.0.xpdl", "WorkflowPack1.0Proc1.1.xpdl");
  }

  public void testDiffPackageVersionSameProcessVersion() throws BonitaException {
    testFirstDeployOkSecondFail("WorkflowPack1.0Proc1.0.xpdl", "WorkflowPack1.1Proc1.0.xpdl");
  }
  
  public void testWrongPackageVersion() throws BonitaException {
    //Deploy a package version, then try to deploy the same package with an older version --> should fail
    testFirstDeployOkSecondFail("WorkflowPack1.1Proc1.0.xpdl", "WorkflowPack1.0Proc1.0.xpdl");
  }

  public void testWrongProcessVersion() throws BonitaException {
    //Deploy a process version, then try to deploy the same process with an older version --> should fail
    testFirstDeployOkSecondFail("WorkflowPack1.0Proc1.1.xpdl", "WorkflowPack1.1Proc1.0.xpdl");
  }  

  private void testFirstDeployOkSecondFail(String url1, String url2) throws BonitaException {
    URL xpdlURL1 = VersioningTest.class.getResource(url1);
    URL xpdlURL2 = VersioningTest.class.getResource(url2);
    PackageDefinitionUUID packageUUID = null;
    ProcessDefinition processDef = null;
    try {
      processDef = getManagementAPI().deployBar(getBarContent(xpdlURL1)).get(processId);
      packageUUID = processDef.getPackageDefinitionUUID();
    } catch (DeploymentException e) {
      fail("First deployment should pass");
    }
    try {
      getManagementAPI().deployBar(getBarContent(xpdlURL2));
      fail("Second deployment should fail");
    } catch (DeploymentException e) {
      // Test pass
    }
    Set<ProcessDefinition> processes = getQueryDefinitionAPI().getProcesses(ProcessState.DEPLOYED);
    assertEquals(1, processes.size());
    assertTrue(processes.contains(processDef));
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    
  }

  /* ***************************
    TEST DEPLOY OK
    Test right processes belongs to right package, right instances belongs to right processes
    Test right processes undeployment
   ****************************/
  
  public void testTwiceDifferentPackageProcessVersion() throws BonitaException {
    deployOkTestGoodBelongings("WorkflowPack1.0Proc1.0.xpdl", packageId, "1.0","1.0",
                               "WorkflowPack1.1Proc1.1.xpdl", packageId, "1.1","1.1");
  }
  
  public void testDiffPackageSameProcessVersion() throws BonitaException {
    deployOkTestGoodBelongings("WorkflowPack1.0Proc1.0.xpdl", packageId, "1.0","1.0",
                               "WorkflowPack2_1.0Proc1.0.xpdl", "workflowPackage2", "1.0","1.0");
  }  
  
  // Deploy two same process in different version --> must be successful
  // Test if the first process belong to the first package, process 2 to the package 2
  // Same test with instances and processes
  // Test of undeployment : undeploy second package, check process 1 and package 1 are still in the engine / Then do the inverse
  private void deployOkTestGoodBelongings(String url1, String packageId1, String pack1Version, String process1Version, 
    String url2, String packageId2, String pack2Version, String process2Version) throws BonitaException {

    URL xpdlURL1 = VersioningTest.class.getResource(url1);
    URL xpdlURL2 = VersioningTest.class.getResource(url2);
    ProcessDefinition process1 = null, process2 = null;
    ProcessInstanceUUID instance1 = null, instance2 = null;
    try {
      process1 = getManagementAPI().deployBar(getBarContent(xpdlURL1)).get(processId);
    } catch (DeploymentException e) {
      fail("First deployment should pass");
    }
    try {
      process2 = getManagementAPI().deployBar(getBarContent(xpdlURL2)).get(processId);
    } catch (DeploymentException e) {
      fail("Second deployment should pass");
    }
    Set<ProcessDefinition> processes = getQueryDefinitionAPI().getProcesses(processId, ProcessState.DEPLOYED);
    assertEquals(2, processes.size());
    assertEquals(process1.getPackageDefinitionUUID().toString(), packageId1 + "-" + pack1Version);
    assertEquals(process2.getPackageDefinitionUUID().toString(), packageId2 + "-" + pack2Version);
    
    try {
      instance1 = getRuntimeAPI().instantiateProcess(process1.getProcessDefinitionUUID());
      instance2 = getRuntimeAPI().instantiateProcess(process2.getProcessDefinitionUUID());
      assertTrue(instance1.toString().startsWith(
          packageId1 + "-" + pack1Version + "-" + processId + "-" + process1Version + "$"));
      assertTrue(instance2.toString().startsWith(
          packageId2 + "-" + pack2Version + "-" + processId + "-" + process2Version + "$"));
      assertNotSame(instance1, instance2);
      assertEquals(process1.getProcessDefinitionUUID(),
        getQueryRuntimeAPI().getProcessInstance(instance1).getProcessDefinitionUUID());
      assertEquals(process2.getProcessDefinitionUUID(), 
          getQueryRuntimeAPI().getProcessInstance(instance2).getProcessDefinitionUUID());
    } catch (ProcessNotFoundException e) {
      fail("Instantiation should pass");
    }
    
    // Test undeploy package 2 then package 1
    checkUndeployment(process2, process1);

    process1 = getManagementAPI().deployBar(getBarContent(xpdlURL1)).get(processId);
    process2 = getManagementAPI().deployBar(getBarContent(xpdlURL2)).get(processId);

    instance1 = getRuntimeAPI().instantiateProcess(process1.getProcessDefinitionUUID());
    instance2 = getRuntimeAPI().instantiateProcess(process2.getProcessDefinitionUUID());
    
    // Test undeploy package 1 then package 2
    checkUndeployment(process1, process2);
  }

  //Check if the right process version is undeployed
  private void checkUndeployment(ProcessDefinition process1, ProcessDefinition process2) throws BonitaException {
    getManagementAPI().undeploy(process1.getPackageDefinitionUUID());
    assertEquals(1, getQueryDefinitionAPI().getProcesses(processId, ProcessState.DEPLOYED).size());
    assertTrue(getQueryDefinitionAPI().getProcesses(processId, ProcessState.DEPLOYED).contains(process2));
    assertEquals(1, getQueryDefinitionAPI().getProcesses(processId, ProcessState.UNDEPLOYED).size());
    assertTrue(getQueryDefinitionAPI().getProcesses(processId, ProcessState.UNDEPLOYED).contains(process1));

    getManagementAPI().undeploy(process2.getPackageDefinitionUUID());
    assertEquals(0, getQueryDefinitionAPI().getProcesses(processId, ProcessState.DEPLOYED).size());
    assertEquals(2, getQueryDefinitionAPI().getProcesses(processId, ProcessState.UNDEPLOYED).size());
    assertTrue(getQueryDefinitionAPI().getProcesses(processId, ProcessState.UNDEPLOYED).contains(process1));
    assertTrue(getQueryDefinitionAPI().getProcesses(processId, ProcessState.UNDEPLOYED).contains(process2));
    
    getManagementAPI().deletePackage(process1.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(process2.getPackageDefinitionUUID());
  }

  
  /* ***************************
     HOOK TEST
  ****************************/  
    
  public void testHookDiffPackAndProcessVersions() throws BonitaException, IOException {
    checkHooks("hook/HookPack1.0Proc1.0.xpdl", "hook/HookPack1.1Proc1.1.xpdl");
  }
  
  private void checkHooks(String hookFile1, String hookFile2) throws BonitaException, IOException {
    //check rights hooks are used with different process versions
    final InputStream in = MiscHookTest.class.getResourceAsStream("IncrementHook2.bytecode");
    if (in == null) {
      throw new IllegalArgumentException("Can't find ressource: IncrementHook2.bytecode");
    }
    byte[] incrementHook2 = Misc.getAllContentFrom(in);

    URL xpdlUrl1 = this.getClass().getResource(hookFile1);
    URL xpdlUrl2 = this.getClass().getResource(hookFile2);
    String processId = "Hook";

    ProcessDefinition process1 = getManagementAPI().deployBar(getBarContent(xpdlUrl1,
        IncrementHook.class)).get(processId);
    
    
    Set<byte[]> hookBytes = new HashSet<byte[]>();
    hookBytes.add(incrementHook2);
    
    Map<String, byte[]> resources = new HashMap<String, byte[]>();
    resources.put(xpdlUrl2.toString(), Misc.getAllContentFrom(xpdlUrl2));
    resources.put(IncrementHook.class.getName() + ".class", incrementHook2);
    byte[] barContent = BusinessArchiveFactory.getBusinessArchiveContent(resources);
    ProcessDefinition process2 = getManagementAPI().deployBar(barContent).get(processId);
    
    
    ProcessDefinitionUUID processUUID1 = process1.getUUID();
    ProcessDefinitionUUID processUUID2 = process2.getUUID();

    ProcessInstanceUUID instanceUUID1 = getRuntimeAPI().instantiateProcess(processUUID1);
    ProcessInstanceUUID instanceUUID2 = getRuntimeAPI().instantiateProcess(processUUID2);
    
    checkExecutedOnce(instanceUUID1, new String[]{"uniqueActivity"});
    String counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID1, "counter");
    assertEquals("1", counter);
    
    checkExecutedOnce(instanceUUID2, new String[]{"uniqueActivity"});
    counter = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID2, "counter");
    assertEquals("2", counter);
    
    getManagementAPI().deletePackage(process1.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(process2.getPackageDefinitionUUID());
  }
  
  
  
  /* *************************************
     TEST UNDEPLOY WITH RUNNING INSTANCES
  ****************************************/  
  public void testRunInstDiffPackAndProcVersion() throws BonitaException, IOException {
    checkUndeployRunningInstances("WorkflowRuntimePack1.0Proc1.0.xpdl", "WorkflowRuntimePack1.1Proc1.1.xpdl");
  }
  
  private void checkUndeployRunningInstances(String url1, String url2) throws BonitaException {
    // Check instances belong to the right process definition
    URL xpdlURL1 = VersioningTest.class.getResource(url1);
    URL xpdlURL2 = VersioningTest.class.getResource(url2);
    ProcessDefinition process1, process2;
    ProcessInstanceUUID instanceUUID1, instanceUUID2;
    
    String processId = "Workflow";
    
    byte[] xpdlContent1 = null, xpdlContent2 = null, bar1 = null, bar2 = null;
    xpdlURL1 = VersioningTest.class.getResource(url1);
    xpdlURL2 = VersioningTest.class.getResource(url2);
    try {
      xpdlContent1 = Misc.getAllContentFrom(xpdlURL1);
      bar1 = BusinessArchiveFactory.getBusinessArchiveContent(url1, xpdlContent1, WorkflowException.class);
      
      xpdlContent2 = Misc.getAllContentFrom(xpdlURL2);
      bar2 = BusinessArchiveFactory.getBusinessArchiveContent(url2, xpdlContent2, WorkflowException.class);
    } catch (IOException e) {
      fail();
    }
    
    process1 = getManagementAPI().deployBar(bar1).get(processId);
    process2 = getManagementAPI().deployBar(bar2).get(processId);
    
    instanceUUID1 = getRuntimeAPI().instantiateProcess(process1.getProcessDefinitionUUID());
    try {
      getManagementAPI().undeploy(process1.getPackageDefinitionUUID());
      fail("Undeployment shouldn't pass when instances are running");
    } catch (DeploymentException e) {
      // Test pass
    }
    try {
      getManagementAPI().undeploy(process2.getPackageDefinitionUUID());
      getManagementAPI().deletePackage(process2.getPackageDefinitionUUID());
    } catch (DeploymentException e) {
      fail("Undeployment should pass");
    }
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID1);
    getManagementAPI().undeploy(process1.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(process1.getPackageDefinitionUUID());
    

    process1 = getManagementAPI().deployBar(bar1).get(processId);
    process2 = getManagementAPI().deployBar(bar2).get(processId);
    
    instanceUUID2 = getRuntimeAPI().instantiateProcess(process2.getProcessDefinitionUUID());
    
    try {
      getManagementAPI().undeploy(process1.getPackageDefinitionUUID());
    } catch (DeploymentException e) {
      fail("Undeployment should pass");
    }
    try {
      getManagementAPI().undeploy(process2.getPackageDefinitionUUID());
      fail("Undeployment shouldn't pass when instances are running");
    } catch (DeploymentException e) {
      // Test pass
    }
    getRuntimeAPI().deleteProcessInstance(instanceUUID2);
    getManagementAPI().deletePackage(process1.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(process2.getPackageDefinitionUUID());
  }
}
